/*
 * Copyright (C) 2002-2003 Ardis Technolgies <roman@ardistech.com>
 *
 * Released under the terms of the GNU GPL v2.0.
 */

#include <linux/module.h>
#include <linux/hash.h>
#include <net/tcp.h>
#include <scsi/scsi.h>

#include "iscsi.h"
#include "iscsi_dbg.h"
#include "iotype.h"

#define STATUS_SHIFT		1

unsigned long debug_enable_flags;

static kmem_cache_t *iscsi_cmnd_cache;
static char dummy_data[1024];

struct iscsi_param default_iscsi_param = {
	.flags = (SESSION_FLG_INITIAL_RTT | SESSION_FLG_IMMEDIATEDATA
		  | SESSION_FLG_DATAPDUINORDER | SESSION_FLG_DATASEQUENCEINORDER),
	.max_connections = 1,
	.max_data_pdu_length = 8192,
	.max_burst_length = 262144,
	.first_burst_length = 65536,
	.default_wait_time = 2,
	.default_retain_time = 20,
	.max_outstanding_r2t = 1,
	.error_recovery_level = 0,
};

static int ctr_major;
static char ctr_name[] = "ietctl";
extern struct file_operations ctr_fops;

static u32 iscsi_cmnd_write_size(struct iscsi_cmnd *cmnd)
{
	struct iscsi_scsi_cmd_hdr *hdr = cmnd_hdr(cmnd);

	if (hdr->flags & ISCSI_CMD_WRITE)
		return be32_to_cpu(hdr->data_length);
	return 0;
}

static u32 iscsi_cmnd_read_size(struct iscsi_cmnd *cmnd)
{
	struct iscsi_scsi_cmd_hdr *hdr = cmnd_hdr(cmnd);

	if (hdr->flags & ISCSI_CMD_READ) {
		if (!(hdr->flags & ISCSI_CMD_WRITE))
			return be32_to_cpu(hdr->data_length);
		if (hdr->flags & ISCSI_CMD_READ) {
			struct iscsi_rlength_ahdr *ahdr =
				(struct iscsi_rlength_ahdr *)cmnd->pdu.ahs;
			if (ahdr && ahdr->ahstype == ISCSI_AHSTYPE_RLENGTH)
				return be32_to_cpu(ahdr->read_length);
		}
	}
	return 0;
}

static void iscsi_device_queue_cmnd(struct iscsi_cmnd *cmnd)
{
	wthread_queue(cmnd);
}

static void iscsi_scsi_queuecmnd(struct iscsi_cmnd *cmnd)
{
	struct iscsi_queue *queue = &cmnd->lun->queue;

	dprintk(D_GENERIC, "%p\n", cmnd);

	if ((cmnd->pdu.bhs.flags & ISCSI_CMD_ATTR_MASK) != ISCSI_CMD_UNTAGGED &&
	    (cmnd->pdu.bhs.flags & ISCSI_CMD_ATTR_MASK) != ISCSI_CMD_SIMPLE) {
		cmnd->pdu.bhs.flags &= ~ISCSI_CMD_ATTR_MASK;
		cmnd->pdu.bhs.flags |= ISCSI_CMD_UNTAGGED;
	}

	spin_lock(&queue->queue_lock);

	set_cmnd_queued(cmnd);

	switch (cmnd->pdu.bhs.flags & ISCSI_CMD_ATTR_MASK) {
	case ISCSI_CMD_UNTAGGED:
	case ISCSI_CMD_SIMPLE:
		if (!list_empty(&queue->wait_list) || queue->ordered_cmnd)
			goto pending;
		queue->active_cnt++;
		break;

	default:
		BUG();
	}
	spin_unlock(&queue->queue_lock);

	iscsi_device_queue_cmnd(cmnd);
	return;
 pending:
	assert(list_empty(&cmnd->list));

	list_add_tail(&cmnd->list, &queue->wait_list);
	cmnd->state = ISCSI_STATE_PENDING;
	spin_unlock(&queue->queue_lock);
	return;
}

static void iscsi_scsi_dequeuecmnd(struct iscsi_cmnd *cmnd)
{
	struct iscsi_queue *queue;

	if (!cmnd->lun)
		return;
	queue = &cmnd->lun->queue;
	spin_lock(&queue->queue_lock);
	switch (cmnd->pdu.bhs.flags & ISCSI_CMD_ATTR_MASK) {
	case ISCSI_CMD_UNTAGGED:
	case ISCSI_CMD_SIMPLE:
		--queue->active_cnt;
		break;
	case ISCSI_CMD_ORDERED:
	case ISCSI_CMD_HEAD_OF_QUEUE:
	case ISCSI_CMD_ACA:
		BUG();
	default:
		/* Should the iscsi_scsi_queuecmnd func reject this ? */
		break;
	}

	while (!list_empty(&queue->wait_list)) {
		cmnd = list_entry(queue->wait_list.next, struct iscsi_cmnd, list);
		switch ((cmnd->pdu.bhs.flags & ISCSI_CMD_ATTR_MASK)) {
		case ISCSI_CMD_UNTAGGED:
		case ISCSI_CMD_SIMPLE:
			list_del_init(&cmnd->list);
			queue->active_cnt++;
			iscsi_device_queue_cmnd(cmnd);
			break;
		case ISCSI_CMD_ORDERED:
		case ISCSI_CMD_HEAD_OF_QUEUE:
		case ISCSI_CMD_ACA:
			BUG();
		}
	}

	spin_unlock(&queue->queue_lock);

	return;
}

static void dump_cmnd(unsigned long data)
{
/* 	struct iscsi_cmnd *cmnd = (struct iscsi_cmnd *) data; */
/* 	eprintk("%p %x %x %x %x\n", */
/* 		cmnd, cmnd->state, cmnd_opcode(cmnd), cmnd_itt(cmnd), cmnd_ttt(cmnd)); */
}

/**
 * create a new command.
 *
 * iscsi_cmnd_create - 
 * @conn: ptr to connection (for i/o)
 *
 * @return    ptr to command or NULL
 */

struct iscsi_cmnd *cmnd_alloc(struct iscsi_conn *conn, int req)
{
	struct iscsi_cmnd *cmnd;

	/* TODO: async interface is necessary ? */
	cmnd = kmem_cache_alloc(iscsi_cmnd_cache, GFP_KERNEL|__GFP_NOFAIL);

	memset(cmnd, 0, sizeof(*cmnd));
	INIT_LIST_HEAD(&cmnd->list);
	INIT_LIST_HEAD(&cmnd->pdu_list);
	INIT_LIST_HEAD(&cmnd->conn_list);
	INIT_LIST_HEAD(&cmnd->hash_list);
	cmnd->conn = conn;
	cmnd->state = ISCSI_STATE_NEW;
	spin_lock(&conn->list_lock);
	atomic_inc(&conn->nr_cmnds);
	if (req)
		list_add_tail(&cmnd->conn_list, &conn->pdu_list);
	spin_unlock(&conn->list_lock);
	cmnd->tio = NULL;

	dprintk(D_GENERIC, "%p:%p\n", conn, cmnd);

	init_timer(&cmnd->timer);
	cmnd->timer.data = 0;

	return cmnd;
}

/**
 * create a new command used as response.
 *
 * iscsi_cmnd_create_rsp_cmnd - 
 * @cmnd: ptr to request command
 *
 * @return    ptr to response command or NULL
 */

static struct iscsi_cmnd *iscsi_cmnd_create_rsp_cmnd(struct iscsi_cmnd *cmnd, int final)
{
	struct iscsi_cmnd *rsp = cmnd_alloc(cmnd->conn, 0);

	if (final)
		set_cmnd_final(rsp);
	list_add_tail(&rsp->pdu_list, &cmnd->pdu_list);
	rsp->req = cmnd;
	return rsp;
}

/**
 * get ptr to response command, which belongs to this request.
 * NOTE: no error checking.
 *
 * iscsi_cmnd_get_rsp_cmnd - 
 * @req_cmnd: ptr to request command
 *
 * @return    ptr to response comman
 */

static struct iscsi_cmnd *iscsi_cmnd_get_rsp_cmnd(struct iscsi_cmnd *req_cmnd)
{
	return list_entry(req_cmnd->pdu_list.prev, struct iscsi_cmnd, pdu_list);
}

static void iscsi_cmnds_init_write(struct list_head *send)
{
	struct iscsi_cmnd *cmnd = list_entry(send->next, struct iscsi_cmnd, list);
	struct iscsi_conn *conn = cmnd->conn;
	struct list_head *pos, *next;

	spin_lock(&conn->list_lock);

	list_for_each_safe(pos, next, send) {
		cmnd = list_entry(pos, struct iscsi_cmnd, list);

		dprintk(D_GENERIC, "%p:%x\n", cmnd, cmnd_opcode(cmnd));

		list_del_init(&cmnd->list);

		assert(conn == cmnd->conn);

		cmnd->state = ISCSI_STATE_WRITE;
		list_add_tail(&cmnd->list, &conn->write_list);
	}

	spin_unlock(&conn->list_lock);

	nthread_wakeup(conn->session->target);
}

static void iscsi_cmnd_init_write(struct iscsi_cmnd *cmnd)
{
	LIST_HEAD(head);

	assert(list_empty(&cmnd->list));
	list_add(&cmnd->list, &head);
	iscsi_cmnds_init_write(&head);
}

static void do_send_data_rsp(struct iscsi_cmnd *cmnd)
{
	struct iscsi_conn *conn = cmnd->conn;
	struct iscsi_cmnd *data_cmnd;
	struct tio *tio = cmnd->tio;
	struct iscsi_scsi_cmd_hdr *req = cmnd_hdr(cmnd);
	struct iscsi_data_in_hdr *rsp;
	u32 pdusize, expsize, scsisize, size, offset, sn;
	LIST_HEAD(send);

	dprintk(D_GENERIC, "%p\n", cmnd);
	pdusize = conn->session->param.max_data_pdu_length;
	expsize = iscsi_cmnd_read_size(cmnd);
	size = min(expsize, tio->size);
	offset = 0;
	sn = 0;

	while (1) {
		data_cmnd = iscsi_cmnd_create_rsp_cmnd(cmnd, size <= pdusize);
		tio_get(tio);
		data_cmnd->tio = tio;
		rsp = (struct iscsi_data_in_hdr *)&data_cmnd->pdu.bhs;

		rsp->opcode = ISCSI_OP_SCSI_DATA_IN;
		rsp->itt = req->itt;
		rsp->ttt = cpu_to_be32(ISCSI_RESERVED_TAG);
		rsp->buffer_offset = offset;
		rsp->data_sn = cpu_to_be32(sn);

		if (size <= pdusize) {
			data_cmnd->pdu.datasize = size;
			rsp->flags = ISCSI_FLG_FINAL | ISCSI_FLG_STATUS;

			scsisize = tio->size;
			if (scsisize < expsize) {
				rsp->flags |= ISCSI_FLG_RESIDUAL_UNDERFLOW;
				size = expsize - scsisize;
			} else if (scsisize > expsize) {
				rsp->flags |= ISCSI_FLG_RESIDUAL_OVERFLOW;
				size = scsisize - expsize;
			} else
				size = 0;
			rsp->residual_count = cpu_to_be32(size);
			list_add_tail(&data_cmnd->list, &send);

			break;
		}

		data_cmnd->pdu.datasize = pdusize;

		size -= pdusize;
		offset += pdusize;
		sn++;

		list_add_tail(&data_cmnd->list, &send);
	}

	iscsi_cmnds_init_write(&send);
}

static struct iscsi_cmnd *create_scsi_rsp(struct iscsi_cmnd *scsi_cmnd)
{
	struct iscsi_cmnd *cmnd;
	struct iscsi_scsi_cmd_hdr *req = cmnd_hdr(scsi_cmnd);
	struct iscsi_scsi_rsp_hdr *rsp;

	cmnd = iscsi_cmnd_create_rsp_cmnd(scsi_cmnd, 1);

	rsp = (struct iscsi_scsi_rsp_hdr *)&cmnd->pdu.bhs;
	rsp->opcode = ISCSI_OP_SCSI_RSP;
	rsp->flags = ISCSI_FLG_FINAL;
	rsp->response = ISCSI_RESPONSE_COMMAND_COMPLETED;
	rsp->cmd_status = GOOD << STATUS_SHIFT;
	rsp->itt = req->itt;

	return cmnd;
}

void send_scsi_rsp(struct iscsi_cmnd *cmnd, int (*func)(struct iscsi_cmnd *))
{
	struct iscsi_cmnd *rsp;
	struct iscsi_scsi_rsp_hdr *rsp_hdr;
	u32 size;

	rsp = create_scsi_rsp(cmnd);
	rsp_hdr = (struct iscsi_scsi_rsp_hdr *) &rsp->pdu.bhs;
	if ((size = iscsi_cmnd_read_size(cmnd)) != 0) {
		rsp_hdr->flags |= ISCSI_FLG_RESIDUAL_UNDERFLOW;
		rsp_hdr->residual_count = cpu_to_be32(size);
	}

	if (func(cmnd) < 0)
		eprintk("%x\n", cmnd_opcode(cmnd));

	iscsi_cmnd_init_write(rsp);
}

static struct iscsi_cmnd *create_sense_rsp(struct iscsi_cmnd *scsi_cmnd,
					   u8 sense_key, u8 asc, u8 ascq)
{
	struct iscsi_cmnd *cmnd;
	struct iscsi_scsi_cmd_hdr *req = cmnd_hdr(scsi_cmnd);
	struct iscsi_scsi_rsp_hdr *rsp;
	struct tio *tio;
	struct iscsi_sense_data *sense;

	cmnd = iscsi_cmnd_create_rsp_cmnd(scsi_cmnd, 1);

	rsp = (struct iscsi_scsi_rsp_hdr *)&cmnd->pdu.bhs;
	rsp->opcode = ISCSI_OP_SCSI_RSP;
	rsp->flags = ISCSI_FLG_FINAL;
	rsp->response = ISCSI_RESPONSE_COMMAND_COMPLETED;
	rsp->cmd_status = CHECK_CONDITION << STATUS_SHIFT;
	rsp->itt = req->itt;

	tio = cmnd->tio = tio_alloc(1);
	sense = (struct iscsi_sense_data *) page_address(tio->pvec[0]);
	assert(sense);
	clear_page(sense);
	sense->length = cpu_to_be16(14);
	sense->data[0] = 0xf0;
	sense->data[2] = sense_key;
	sense->data[7] = 6;	// Additional sense length
	sense->data[12] = asc;
	sense->data[13] = ascq;

	cmnd->pdu.datasize = sizeof(struct iscsi_sense_data) + 14;
	tio->size = (cmnd->pdu.datasize + 3) & -4;
	tio->offset = 0;

	return cmnd;
}

void send_data_rsp(struct iscsi_cmnd *cmnd, int (*func)(struct iscsi_cmnd *))
{
	struct iscsi_cmnd *rsp;

	if (func(cmnd) < 0) {
		rsp = create_sense_rsp(cmnd, ILLEGAL_REQUEST, 0x24, 0x0);
		iscsi_cmnd_init_write(rsp);
	} else
		do_send_data_rsp(cmnd);
}

/**
 * Free a command.
 * Also frees the additional header.
 *
 * iscsi_cmnd_remove - 
 * @cmnd: ptr to command
 */

void iscsi_cmnd_remove(struct iscsi_cmnd *cmnd)
{
	struct iscsi_conn *conn;

	if (!cmnd)
		return;
	dprintk(D_GENERIC, "%p\n", cmnd);
	conn = cmnd->conn;
	kfree(cmnd->pdu.ahs);

	if (!list_empty(&cmnd->list)) {
		struct iscsi_scsi_cmd_hdr *req = cmnd_hdr(cmnd);

		eprintk("cmnd %p still on some list?, %x, %x, %x, %x, %x, %x, %x %lx %lx\n",
			cmnd, req->opcode, req->scb[0], req->flags, req->itt,
			be32_to_cpu(req->data_length),
			req->cmd_sn, be32_to_cpu(cmnd->pdu.datasize), cmnd->state, conn->state);

		if (cmnd->req) {
			struct iscsi_scsi_cmd_hdr *req = cmnd_hdr(cmnd->req);
			eprintk("%p %x %u\n", req, req->opcode, req->scb[0]);
		}
		dump_stack();
		BUG();
	}
	list_del(&cmnd->list);
	spin_lock(&conn->list_lock);
	atomic_dec(&conn->nr_cmnds);
	list_del(&cmnd->conn_list);
	spin_unlock(&conn->list_lock);

	if (cmnd->timer.data)
		del_timer_sync(&cmnd->timer);

	if (cmnd->tio)
		tio_put(cmnd->tio);

	kmem_cache_free(iscsi_cmnd_cache, cmnd);
}

static void iscsi_cmnd_skip_pdu(struct iscsi_cmnd *cmnd)
{
	struct iscsi_conn *conn = cmnd->conn;
	struct tio *tio = cmnd->tio;
	struct iscsi_scsi_cmd_hdr *req = cmnd_hdr(cmnd);
	char *addr;
	u32 size;
	int i;

	eprintk("%x %x %u\n", cmnd_opcode(cmnd), req->scb[0], cmnd->pdu.datasize);

	if (!(size = cmnd->pdu.datasize))
		return;

	if (tio)
		assert(tio->pg_cnt > 0);
	else
		tio = cmnd->tio = tio_alloc(1);

	addr = page_address(tio->pvec[0]);
	assert(addr);
	size = (size + 3) & -4;
	conn->read_size = size;
	for (i = 0; size > PAGE_CACHE_SIZE; i++, size -= PAGE_CACHE_SIZE) {
		assert(i < ISCSI_CONN_IOV_MAX);
		conn->read_iov[i].iov_base = addr;
		conn->read_iov[i].iov_len = PAGE_CACHE_SIZE;
	}
	conn->read_iov[i].iov_base = addr;
	conn->read_iov[i].iov_len = size;
	conn->read_msg.msg_iov = conn->read_iov;
	conn->read_msg.msg_iovlen = ++i;
}

static void iscsi_cmnd_reject(struct iscsi_cmnd *cmnd, int reason)
{
	struct iscsi_cmnd *rej_cmnd;
	struct iscsi_reject_hdr *rsp;
	struct tio *tio;
	char *addr;

	rej_cmnd = iscsi_cmnd_create_rsp_cmnd(cmnd, 1);
	rsp = (struct iscsi_reject_hdr *)&rej_cmnd->pdu.bhs;

	rsp->opcode = ISCSI_OP_REJECT;
	rsp->ffffffff = ISCSI_RESERVED_TAG;
	rsp->reason = reason;

	rej_cmnd->tio = tio = tio_alloc(1);
	addr = page_address(tio->pvec[0]);
	clear_page(addr);
	memcpy(addr, &cmnd->pdu.bhs, sizeof(struct iscsi_hdr));
	tio->size = rej_cmnd->pdu.datasize = sizeof(struct iscsi_hdr);
	iscsi_cmnd_skip_pdu(cmnd);

	cmnd->pdu.bhs.opcode = ISCSI_OP_PDU_REJECT;
}

static void cmnd_set_sn(struct iscsi_cmnd *cmnd, int set_stat_sn)
{
	struct iscsi_conn *conn = cmnd->conn;

	if (set_stat_sn)
		cmnd->pdu.bhs.sn = cpu_to_be32(conn->stat_sn++);
	cmnd->pdu.bhs.exp_sn = cpu_to_be32(conn->session->exp_cmd_sn);
	cmnd->pdu.bhs.max_sn = cpu_to_be32(conn->session->exp_cmd_sn + MAX_QUEUED_CMD_NR);
}

static void update_stat_sn(struct iscsi_cmnd *cmnd)
{
	struct iscsi_conn *conn = cmnd->conn;
	u32 exp_stat_sn;

	cmnd->pdu.bhs.exp_sn = exp_stat_sn = be32_to_cpu(cmnd->pdu.bhs.exp_sn);
	dprintk(D_GENERIC, "%x,%x\n", cmnd_opcode(cmnd), exp_stat_sn);
	if ((int)(exp_stat_sn - conn->exp_stat_sn) > 0 &&
	    (int)(exp_stat_sn - conn->stat_sn) <= 0) {
		// free pdu resources
		cmnd->conn->exp_stat_sn = exp_stat_sn;
	} else if ((int)(exp_stat_sn - conn->exp_stat_sn) != 0) {
		eprintk("stat sequence number error (%x,%x)\n", exp_stat_sn, conn->exp_stat_sn);
	}
}

static int check_cmd_sn(struct iscsi_cmnd *cmnd)
{
	struct iscsi_session *session = cmnd->conn->session;
	u32 cmd_sn;

	cmnd->pdu.bhs.sn = cmd_sn = be32_to_cpu(cmnd->pdu.bhs.sn);
	dprintk(D_GENERIC, "%d(%d)\n", cmd_sn, session->exp_cmd_sn);
	if ((s32)(cmd_sn - session->exp_cmd_sn) >= 0)
		return 0;
	eprintk("sequence error (%x,%x)\n", cmd_sn, session->exp_cmd_sn);
	return -ISCSI_REASON_PROTOCOL_ERROR;
}

static struct iscsi_cmnd *__cmnd_find_hash(struct iscsi_session *session, u32 itt, u32 ttt)
{
	struct list_head *head;
	struct iscsi_cmnd *cmnd;

	head = &session->cmnd_hash[cmnd_hashfn(itt)];

	list_for_each_entry(cmnd, head, hash_list) {
		if (cmnd->pdu.bhs.itt == itt) {
			if ((ttt != ISCSI_RESERVED_TAG) && (ttt != cmnd->target_task_tag))
				continue;
			return cmnd;
		}
	}

	return NULL;
}

static struct iscsi_cmnd *cmnd_find_hash(struct iscsi_session *session, u32 itt, u32 ttt)
{
	struct iscsi_cmnd *cmnd;

	spin_lock(&session->cmnd_hash_lock);

	cmnd = __cmnd_find_hash(session, itt, ttt);

	spin_unlock(&session->cmnd_hash_lock);

	return cmnd;
}

static int cmnd_insert_hash(struct iscsi_cmnd *cmnd)
{
	struct iscsi_session *session = cmnd->conn->session;
	struct iscsi_cmnd *hash_cmnd;
	struct list_head *head;
	int err = 0;
	u32 itt = cmnd->pdu.bhs.itt;

	dprintk(D_GENERIC, "%p:%x\n", cmnd, itt);
	if (itt == ISCSI_RESERVED_TAG) {
		err = -ISCSI_REASON_PROTOCOL_ERROR;
		goto out;
	}

	head = &session->cmnd_hash[cmnd_hashfn(cmnd->pdu.bhs.itt)];

	spin_lock(&session->cmnd_hash_lock);

	hash_cmnd = __cmnd_find_hash(session, itt, ISCSI_RESERVED_TAG);
	if (!hash_cmnd) {
		list_add_tail(&cmnd->hash_list, head);
		set_cmnd_hashed(cmnd);
	} else
		err = -ISCSI_REASON_TASK_IN_PROGRESS;

	spin_unlock(&session->cmnd_hash_lock);

	if (!err) {
		update_stat_sn(cmnd);
		err = check_cmd_sn(cmnd);
	}

out:
	return err;
}

static void cmnd_remove_hash(struct iscsi_cmnd *cmnd)
{
	struct iscsi_session *session = cmnd->conn->session;
	struct iscsi_cmnd *hash_cmnd;

	spin_lock(&session->cmnd_hash_lock);

	hash_cmnd = __cmnd_find_hash(session, cmnd->pdu.bhs.itt, ISCSI_RESERVED_TAG);

	if (hash_cmnd && hash_cmnd == cmnd)
		list_del(&hash_cmnd->hash_list);
	else
		eprintk("%p:%x not found\n", cmnd, cmnd_itt(cmnd));

	spin_unlock(&session->cmnd_hash_lock);
}

static void iscsi_cmnd_ignore_data(struct iscsi_cmnd *scsi_cmnd)
{
	struct iscsi_cmnd *rsp_cmnd;
	struct iscsi_scsi_cmd_hdr *req = cmnd_hdr(scsi_cmnd);
	struct iscsi_scsi_rsp_hdr *rsp;
	u32 size;

	rsp_cmnd = iscsi_cmnd_get_rsp_cmnd(scsi_cmnd);
	rsp = (struct iscsi_scsi_rsp_hdr *)&rsp_cmnd->pdu.bhs;
	if (cmnd_opcode(rsp_cmnd) != ISCSI_OP_SCSI_RSP) {
		eprintk("unexpected response command %u\n", cmnd_opcode(rsp_cmnd));
		return;
	}

	size = iscsi_cmnd_write_size(scsi_cmnd);
	if (size) {
		rsp->flags |= ISCSI_FLG_RESIDUAL_UNDERFLOW;
		rsp->residual_count = cpu_to_be32(size);
	}
	size = iscsi_cmnd_read_size(scsi_cmnd);
	if (size) {
		if (req->flags & ISCSI_CMD_WRITE) {
			rsp->flags |= ISCSI_FLG_BIRESIDUAL_UNDERFLOW;
			rsp->bi_residual_count = cpu_to_be32(size);
		} else {
			rsp->flags |= ISCSI_FLG_RESIDUAL_UNDERFLOW;
			rsp->residual_count = cpu_to_be32(size);
		}
	}
	scsi_cmnd->pdu.bhs.opcode = (scsi_cmnd->pdu.bhs.opcode & ~ISCSI_OPCODE_MASK) | ISCSI_OP_SCSI_REJECT;

	iscsi_cmnd_skip_pdu(scsi_cmnd);
}

/**
 * Initialize conn->read_* from a target command to start reading the pdu data.
 * Called from the read thread.
 *
 * iscsi_cmnd_receive_pdu - 
 * @conn: ptr to connection
 * @tio: ptr to target command
 * @offset: offset in target data
 * @size: size of pdu
 */

static void iscsi_cmnd_receive_pdu(struct iscsi_conn *conn, struct tio *tio, u32 offset, u32 size)
{
	int idx, i;
	char *addr;

	dprintk(D_GENERIC, "%p %u,%u\n", tio, offset, size);
	offset += tio->offset;

	assert(offset < tio->offset + tio->size);
	assert(offset + size <= tio->offset + tio->size);

	idx = offset >> PAGE_CACHE_SHIFT;
	offset &= ~PAGE_CACHE_MASK;

	conn->read_msg.msg_iov = conn->read_iov;
	conn->read_size = (size + 3) & -4;
	conn->read_overflow = 0;

	i = 0;
	while (1) {
		assert(tio->pvec[idx]);
		addr = page_address(tio->pvec[idx]);
		assert(addr);
		conn->read_iov[i].iov_base =  addr + offset;
		if (offset + size <= PAGE_CACHE_SIZE) {
			conn->read_iov[i].iov_len = size;
			conn->read_msg.msg_iovlen = ++i;
			break;
		}
		conn->read_iov[i].iov_len = PAGE_CACHE_SIZE - offset;
		size -= conn->read_iov[i].iov_len;
		offset = 0;
		if (++i >= ISCSI_CONN_IOV_MAX) {
			conn->read_msg.msg_iovlen = i;
			conn->read_overflow = size;
			conn->read_size -= size;
			break;
		}

		idx++;
	}
}

static void iscsi_get_offset_and_length(struct iet_volume *lu, u8 *cmd, loff_t *off, u32 *len)
{
	assert(lu);

	switch (cmd[0]) {
	case READ_6:
	case WRITE_6:
		*off = ((cmd[1] & 0x1f) << 16) + (cmd[2] << 8) + cmd[3];
		*len = cmd[4];
		if (!*len)
			*len = 256;
		break;
	case READ_10:
	case WRITE_10:
	case WRITE_VERIFY:
		*off = be32_to_cpu(*(u32 *)&cmd[2]);
		*len = (cmd[7] << 8) + cmd[8];
		break;
	case READ_16:
	case WRITE_16:
		*off = be64_to_cpu(*(u64 *)&cmd[2]);
		*len = be32_to_cpu(*(u32 *)&cmd[10]);
		break;
	default:
		BUG();
	}

	*off <<= lu->blk_shift;
	*len <<= lu->blk_shift;
}

static u32 translate_lun(u16 * data)
{
	u8 *lun = (u8 *) data;
	u32 res;

	switch (*lun >> 6) {
	case 0:
		res = lun[1];
		break;
	case 1:
		res = (0x3f & lun[0]) << 8 | lun[1];
		break;
	case 2:
	case 3:
	default:
		eprintk("%u %u %u %u\n", data[0], data[1], data[2], data[3]);
		res = ~0U;
		break;
	}

	return res;
}

static void send_r2t(struct iscsi_cmnd *cmnd)
{
	struct iscsi_scsi_cmd_hdr *req = cmnd_hdr(cmnd);
	struct iscsi_cmnd *rsp_cmnd;
	struct iscsi_r2t_hdr *rsp;
	u32 length, offset, burst;
	LIST_HEAD(send);

	length = cmnd->r2t_length;
	burst = cmnd->conn->session->param.max_burst_length;
	offset = be32_to_cpu(req->data_length) - length;

	do {
		rsp_cmnd = iscsi_cmnd_create_rsp_cmnd(cmnd, 0);
		rsp_cmnd->pdu.bhs.ttt = cmnd->target_task_tag;

		rsp = (struct iscsi_r2t_hdr *)&rsp_cmnd->pdu.bhs;
		rsp->opcode = ISCSI_OP_R2T;
		rsp->flags = ISCSI_FLG_FINAL;
		rsp->itt = req->itt;
		rsp->r2t_sn = cpu_to_be32(cmnd->r2t_sn++);
		rsp->buffer_offset = cpu_to_be32(offset);
		if (length > burst) {
			rsp->data_length = cpu_to_be32(burst);
			length -= burst;
			offset += burst;
		} else {
			rsp->data_length = cpu_to_be32(length);
			length = 0;
		}

		dprintk(D_WRITE, "%x %u %u %u %u\n", cmnd_itt(cmnd),
			 be32_to_cpu(rsp->data_length), be32_to_cpu(rsp->buffer_offset),
			 be32_to_cpu(rsp->r2t_sn), cmnd->outstanding_r2t);

		list_add_tail(&rsp_cmnd->list, &send);

		if (++cmnd->outstanding_r2t >= cmnd->conn->session->param.max_outstanding_r2t)
			break;

	} while (length);

	cmnd->state = ISCSI_STATE_SEND_R2T;
	iscsi_cmnds_init_write(&send);
}

static void prepare_execute_scsi(struct iscsi_cmnd *cmnd)
{
	if (cmnd->r2t_length) {
		if (!cmnd->is_unsolicited_data)
			send_r2t(cmnd);
	} else {
		if (cmnd->lun) {
			iscsi_scsi_queuecmnd(cmnd);
		} else {
			cmnd->state = ISCSI_STATE_QUEUED;
			iscsi_device_queue_cmnd(cmnd);
		}
	}
}

static int handle_noop_out(struct iscsi_conn *conn, struct iscsi_cmnd *cmnd)
{
	u32 size, tmp;
	int i, err = 0;

	if (cmnd_ttt(cmnd) != cpu_to_be32(ISCSI_RESERVED_TAG)) {
		/*
		 * We don't request a NOP-Out by sending a NOP-In.
		 * See 10.18.2 in the draft 20.
		 */
		eprintk("initiator bug %x\n", cmnd_itt(cmnd));
		err = -ISCSI_REASON_PROTOCOL_ERROR;
		goto out;
	}

	if (cmnd_itt(cmnd) == cpu_to_be32(ISCSI_RESERVED_TAG)) {
		if (!(cmnd->pdu.bhs.opcode & ISCSI_OP_IMMEDIATE))
			eprintk("%s\n","initiator bug!");
		update_stat_sn(cmnd);
		err = check_cmd_sn(cmnd);
		goto out;
	} else if ((err = cmnd_insert_hash(cmnd)) < 0) {
		eprintk("ignore this request %x\n", cmnd_itt(cmnd));
		goto out;
	}

	if ((size = cmnd->pdu.datasize)) {
		size = (size + 3) & -4;
		conn->read_msg.msg_iov = conn->read_iov;
		if (cmnd->pdu.bhs.itt != cpu_to_be32(ISCSI_RESERVED_TAG)) {
			struct tio *tio;
			int pg_cnt = get_pgcnt(size, 0);

			assert(pg_cnt < ISCSI_CONN_IOV_MAX);
			cmnd->tio = tio = tio_alloc(pg_cnt);
			tio_set(tio, size, 0);

			for (i = 0; i < pg_cnt; i++) {
				conn->read_iov[i].iov_base
					= page_address(tio->pvec[i]);
				tmp = min_t(u32, size, PAGE_CACHE_SIZE);
				conn->read_iov[i].iov_len = tmp;
				conn->read_size += tmp;
				size -= tmp;
			}
		} else {
			for (i = 0; i < ISCSI_CONN_IOV_MAX; i++) {
				conn->read_iov[i].iov_base = dummy_data;
				tmp = min_t(u32, size, sizeof(dummy_data));
				conn->read_iov[i].iov_len = tmp;
				conn->read_size += tmp;
				size -= tmp;
			}
		}
		assert(!size);
		conn->read_overflow = size;
		conn->read_msg.msg_iovlen = i;
	}

out:
	return err;
}

static u32 get_next_ttt(struct iscsi_session *session)
{
	u32 ttt;

	if (session->next_ttt == ISCSI_RESERVED_TAG)
		session->next_ttt++;
	ttt = session->next_ttt++;

	return cpu_to_be32(ttt);
}

static void handle_scsi_cmnd(struct iscsi_conn *conn, struct iscsi_cmnd *cmnd)
{
	struct iscsi_scsi_cmd_hdr *req = cmnd_hdr(cmnd);

	cmnd->lun = volume_lookup(conn->session->target, translate_lun(req->lun));

	dprintk(D_GENERIC, "scsi command: %02x\n", req->scb[0]);
	if (!cmnd->lun) {
		switch (req->scb[0]) {
		case INQUIRY:
		case REPORT_LUNS:
			break;
		default:
			eprintk("%x %x\n", cmnd_itt(cmnd), req->scb[0]);
			create_sense_rsp(cmnd, ILLEGAL_REQUEST, 0x25, 0x0);
			iscsi_cmnd_ignore_data(cmnd);
			goto out;
		}
	}

	switch (req->scb[0]) {
	case SERVICE_ACTION_IN:
		if ((req->scb[1] & 0x1f) != 0x10)
			goto error;
	case INQUIRY:
	case REPORT_LUNS:
	case TEST_UNIT_READY:
	case SYNCHRONIZE_CACHE:
	case VERIFY:
	case VERIFY_16:
	case START_STOP:
	case READ_CAPACITY:
	case MODE_SENSE:
	case REQUEST_SENSE:
	case RESERVE:
	case RELEASE:
	case RESERVE_10:
	case RELEASE_10:
	{
		if (!(req->flags & ISCSI_CMD_FINAL) || cmnd->pdu.datasize) {
			/* unexpected unsolicited data */
			eprintk("%x %x\n", cmnd_itt(cmnd), req->scb[0]);
			create_sense_rsp(cmnd, ABORTED_COMMAND, 0xc, 0xc);
			iscsi_cmnd_ignore_data(cmnd);
		}
		break;
	}
	case READ_6:
	case READ_10:
	case READ_16:
	{
		loff_t offset;
		u32 length;

		if (!(req->flags & ISCSI_CMD_FINAL) || cmnd->pdu.datasize) {
			/* unexpected unsolicited data */
			eprintk("%x %x\n", cmnd_itt(cmnd), req->scb[0]);
			create_sense_rsp(cmnd, ABORTED_COMMAND, 0xc, 0xc);
			iscsi_cmnd_ignore_data(cmnd);
		}

		iscsi_get_offset_and_length(cmnd->lun, req->scb, &offset, &length);
		cmnd->tio = tio_alloc(get_pgcnt(length, offset));
		tio_set(cmnd->tio, length, offset);
		break;
	}
	case WRITE_6:
	case WRITE_10:
	case WRITE_16:
	case WRITE_VERIFY:
	{
		loff_t offset;
		u32 length;

		cmnd->r2t_length = be32_to_cpu(req->data_length) - cmnd->pdu.datasize;
		cmnd->is_unsolicited_data = !(req->flags & ISCSI_CMD_FINAL);
		cmnd->target_task_tag = get_next_ttt(conn->session);

		if (!(conn->session->param.flags & SESSION_FLG_IMMEDIATEDATA) &&
		    cmnd->pdu.datasize) {
			eprintk("%x %x\n", cmnd_itt(cmnd), req->scb[0]);
		}
		if ((conn->session->param.flags & SESSION_FLG_INITIAL_RTT) &&
		    !(req->flags & ISCSI_CMD_FINAL)) {
			eprintk("%x %x\n", cmnd_itt(cmnd), req->scb[0]);
		}

		if (req->scb[0] == WRITE_VERIFY && req->scb[1] & 0x02)
			eprintk("Verification is ignored %x\n", cmnd_itt(cmnd));

		iscsi_get_offset_and_length(cmnd->lun, req->scb, &offset, &length);
		cmnd->tio = tio_alloc(get_pgcnt(length, offset));
		tio_set(cmnd->tio, length, offset);

		if (cmnd->pdu.datasize) {
			iscsi_cmnd_receive_pdu(conn, cmnd->tio, 0, cmnd->pdu.datasize);
		}
		break;
	}
	error:
	default:
		eprintk("Unsupported %x\n", req->scb[0]);
		create_sense_rsp(cmnd, ILLEGAL_REQUEST, 0x20, 0x0);
		iscsi_cmnd_ignore_data(cmnd);
		break;
	}

out:
	return;
}

static void handle_data_out(struct iscsi_conn *conn, struct iscsi_cmnd *cmnd)
{
	struct iscsi_data_out_hdr *req = (struct iscsi_data_out_hdr *)&cmnd->pdu.bhs;
	struct iscsi_cmnd *scsi_cmnd = NULL;
	u32 offset = be32_to_cpu(req->buffer_offset);

	update_stat_sn(cmnd);

/* 	eprintk("%x %x %u %u\n", cmnd_itt(cmnd), req->flags, cmnd->pdu.datasize, offset); */

	cmnd->req = scsi_cmnd = cmnd_find_hash(conn->session, req->itt, req->ttt);
	if (!scsi_cmnd) {
		eprintk("unable to find scsi task %x %x\n",
			cmnd_itt(cmnd), cmnd_ttt(cmnd));
		goto skip_data;
	}

	if (scsi_cmnd->r2t_length < cmnd->pdu.datasize) {
		eprintk("invalid data len %u %u\n", cmnd->pdu.datasize, scsi_cmnd->r2t_length);
		goto skip_data;
	}

	{
		struct iscsi_scsi_cmd_hdr *scsi_req = cmnd_hdr(scsi_cmnd);
		if (scsi_cmnd->r2t_length + offset != be32_to_cpu(scsi_req->data_length))
			eprintk("%u %u %u\n", scsi_cmnd->r2t_length, offset,
				be32_to_cpu(scsi_req->data_length));
	}

	scsi_cmnd->r2t_length -= cmnd->pdu.datasize;

	if (req->ttt == cpu_to_be32(ISCSI_RESERVED_TAG)) {
		/* unsolicited burst data */
		if (scsi_cmnd->pdu.bhs.flags & ISCSI_FLG_FINAL) {
			eprintk("unexpected data from %x %x\n",
				cmnd_itt(cmnd), cmnd_ttt(cmnd));
			goto skip_data;
		}
	}

	dprintk(D_WRITE, "%u %p %p %p %u %u\n", req->ttt, cmnd, scsi_cmnd,
		scsi_cmnd->tio, offset, cmnd->pdu.datasize);

	// CHECKME!
	iscsi_cmnd_receive_pdu(conn, scsi_cmnd->tio, offset, cmnd->pdu.datasize);
	return;

skip_data:
	// reject? FIXME
	cmnd->pdu.bhs.opcode = ISCSI_OP_DATA_REJECT;
	iscsi_cmnd_skip_pdu(cmnd);
	return;
}

static void finish_data_out(struct iscsi_conn *conn, struct iscsi_cmnd *cmnd)
{
	struct iscsi_data_out_hdr *req = (struct iscsi_data_out_hdr *) &cmnd->pdu.bhs;
	struct iscsi_cmnd *scsi_cmnd;
	u32 offset;

	assert(cmnd);
	scsi_cmnd = cmnd->req;

	if (conn->read_overflow) {
		eprintk("%x %u\n", cmnd_itt(cmnd), conn->read_overflow);
		assert(scsi_cmnd->tio);
		offset = be32_to_cpu(req->buffer_offset);
		offset += cmnd->pdu.datasize - conn->read_overflow;
		iscsi_cmnd_receive_pdu(conn, scsi_cmnd->tio, offset, conn->read_overflow);
		return;
	}

	if (req->ttt == cpu_to_be32(ISCSI_RESERVED_TAG)) {
		if (req->flags & ISCSI_FLG_FINAL) {
			scsi_cmnd->is_unsolicited_data = 0;
			prepare_execute_scsi(scsi_cmnd);
		}
	} else {
		/* TODO : proper error handling */
		if (!(req->flags & ISCSI_FLG_FINAL) && scsi_cmnd->r2t_length == 0)
			eprintk("initiator error %x\n", cmnd_itt(scsi_cmnd));

		if (!(req->flags & ISCSI_FLG_FINAL))
			goto out;

		scsi_cmnd->outstanding_r2t--;

		if (scsi_cmnd->r2t_length == 0)
			assert(list_empty(&scsi_cmnd->pdu_list));

		prepare_execute_scsi(scsi_cmnd);
	}

out:
	iscsi_cmnd_remove(cmnd);
	return;
}

static void handle_management_functions(struct iscsi_cmnd *cmnd)
{
	struct iscsi_conn *conn = cmnd->conn;
	struct iscsi_cmnd *rsp_cmnd;
	struct iscsi_task_mgt_hdr *req = (struct iscsi_task_mgt_hdr *)&cmnd->pdu.bhs;
	struct iscsi_task_rsp_hdr *rsp;
	int function;

	eprintk("%p: %x %x %x\n", cmnd, cmnd_opcode(cmnd),
		cmnd->pdu.bhs.sn, req->function & ISCSI_FUNCTION_MASK);
	{
		struct iscsi_cmnd *cmnd;
		dprintk(D_TASK_MGT, "task management command: %x\n", req->function);
		list_for_each_entry(cmnd, &conn->session->pending_list, list) {
			dprintk(D_TASK_MGT, "%p: %u %u\n", cmnd, cmnd_opcode(cmnd),
					cmnd->pdu.bhs.sn);
		}
	}

	rsp_cmnd = iscsi_cmnd_create_rsp_cmnd(cmnd, 1);
	rsp = (struct iscsi_task_rsp_hdr *)&rsp_cmnd->pdu.bhs;

	rsp->opcode = ISCSI_OP_SCSI_TASK_MGT_RSP;
	rsp->flags = ISCSI_FLG_FINAL;
	rsp->itt = req->itt;

	function = req->function & ISCSI_FUNCTION_MASK;

	/* Not implemented properly yet. */
	rsp->response = ISCSI_RESPONSE_FUNCTION_UNSUPPORTED;

#if 0
	switch (function) {
	case ISCSI_FUNCTION_ABORT_TASK:
	case ISCSI_FUNCTION_ABORT_TASK_SET:
	case ISCSI_FUNCTION_CLEAR_ACA:
	case ISCSI_FUNCTION_CLEAR_TASK_SET:
	case ISCSI_FUNCTION_LOGICAL_UNIT_RESET:
		lun = target_lookup_lun(conn->session->target, be16_to_cpu(req->lun[0]));
		if (!lun)
			rsp->response = ISCSI_RESPONSE_UNKNOWN_LUN;
	}

	if (!rsp->response) switch (function) {
	case ISCSI_FUNCTION_ABORT_TASK:
		//rsp->response = ISCSI_RESPONSE_UNKNOWN_TASK;
		rsp->response = ISCSI_RESPONSE_FUNCTION_COMPLETE;
		break;
	case ISCSI_FUNCTION_ABORT_TASK_SET:
		rsp->response = ISCSI_RESPONSE_FUNCTION_REJECTED;
		break;
	case ISCSI_FUNCTION_CLEAR_ACA:
		rsp->response = ISCSI_RESPONSE_FUNCTION_REJECTED;
		break;
	case ISCSI_FUNCTION_CLEAR_TASK_SET:
		rsp->response = ISCSI_RESPONSE_FUNCTION_REJECTED;
		break;
	case ISCSI_FUNCTION_LOGICAL_UNIT_RESET:
		// TODO
		rsp->response = ISCSI_RESPONSE_FUNCTION_COMPLETE;
		//rsp->response = ISCSI_RESPONSE_FUNCTION_REJECTED;
		break;
	case ISCSI_FUNCTION_TARGET_WARM_RESET:
		// TODO
		rsp->response = ISCSI_RESPONSE_FUNCTION_COMPLETE;
		//iscsi_conn_closefd_nolock(conn);
		break;
	case ISCSI_FUNCTION_TARGET_COLD_RESET:
		// TODO
		eprintk("%s\n", "cold reset");
		rsp->response = ISCSI_RESPONSE_FUNCTION_COMPLETE;
		conn_close(conn);
		break;
	case ISCSI_FUNCTION_TASK_REASSIGN:
		rsp->response = ISCSI_RESPONSE_FUNCTION_REJECTED;
		break;
	default:
		rsp->response = ISCSI_RESPONSE_FUNCTION_REJECTED;
		break;
	}
#endif

	iscsi_cmnd_init_write(rsp_cmnd);
}

static void iscsi_cmnd_execute(struct iscsi_cmnd *cmnd)
{
	struct iscsi_cmnd *rsp_cmnd;

	dprintk(D_GENERIC, "%p,%x,%u\n", cmnd, cmnd_opcode(cmnd), cmnd->pdu.bhs.sn);

	switch (cmnd_opcode(cmnd)) {
	case ISCSI_OP_NOOP_OUT:
	{
		struct iscsi_nop_in_hdr *rsp;
		if (cmnd->pdu.bhs.itt != cpu_to_be32(ISCSI_RESERVED_TAG)) {
			rsp_cmnd = iscsi_cmnd_create_rsp_cmnd(cmnd, 1);

			rsp = (struct iscsi_nop_in_hdr *)&rsp_cmnd->pdu.bhs;
			rsp->opcode = ISCSI_OP_NOOP_IN;
			rsp->flags = ISCSI_FLG_FINAL;
			rsp->itt = cmnd->pdu.bhs.itt;
			rsp->ttt = cpu_to_be32(ISCSI_RESERVED_TAG);

			if (cmnd->pdu.datasize)
				assert(cmnd->tio);
			else
				assert(!cmnd->tio);

			if (cmnd->tio) {
				tio_get(cmnd->tio);
				rsp_cmnd->tio = cmnd->tio;
			}

			assert(get_pgcnt(cmnd->pdu.datasize, 0) < ISCSI_CONN_IOV_MAX);
			rsp_cmnd->pdu.datasize = cmnd->pdu.datasize;
			iscsi_cmnd_init_write(rsp_cmnd);
		} else
			iscsi_cmnd_remove(cmnd);
		break;
	}
	case ISCSI_OP_SCSI_CMD:
	{
		if (cmnd->state != ISCSI_STATE_START && cmnd->state != ISCSI_STATE_PUSHED)
			eprintk("unexpected state %lx of cmnd %p\n", cmnd->state, cmnd);

		prepare_execute_scsi(cmnd);
		break;
	}
	case ISCSI_OP_SCSI_TASK_MGT_MSG:
		handle_management_functions(cmnd);
		break;
	case ISCSI_OP_TEXT_CMD:
		break;
	case ISCSI_OP_LOGOUT_CMD:
	{
		struct iscsi_logout_req_hdr *req = (struct iscsi_logout_req_hdr *)&cmnd->pdu.bhs;
		struct iscsi_logout_rsp_hdr *rsp;

		rsp_cmnd = iscsi_cmnd_create_rsp_cmnd(cmnd, 1);
		rsp = (struct iscsi_logout_rsp_hdr *)&rsp_cmnd->pdu.bhs;
		rsp->opcode = ISCSI_OP_LOGOUT_RSP;
		rsp->flags = ISCSI_FLG_FINAL;
		rsp->itt = req->itt;
		iscsi_cmnd_init_write(rsp_cmnd);
		break;
	}
	case ISCSI_OP_SNACK_CMD:
		break;
	case ISCSI_OP_SCSI_REJECT:
		rsp_cmnd = iscsi_cmnd_get_rsp_cmnd(cmnd);
		iscsi_cmnd_init_write(rsp_cmnd);
		break;
	default:
		eprintk("unexpected cmnd op %x\n", cmnd_opcode(cmnd));
		break;
	}
}

static void iscsi_cmnd_send_pdu(struct iscsi_conn *conn, struct tio *tio, u32 offset, u32 size)
{
	dprintk(D_GENERIC, "%p %u,%u\n", tio, offset, size);
	offset += tio->offset;

	assert(offset <= tio->offset + tio->size);
	assert(offset + size <= tio->offset + tio->size);

	conn->write_tcmnd = tio;
	conn->write_offset = offset;
	conn->write_size += size;
}

static void cmnd_send_pdu(struct iscsi_conn *conn, struct iscsi_cmnd *cmnd)
{
	u32 size;
	struct tio *tio;

	if (!cmnd->pdu.datasize)
		return;

	size = (cmnd->pdu.datasize + 3) & -4;
	tio = cmnd->tio;
	assert(tio);
	assert(tio->size == size);
	iscsi_cmnd_send_pdu(conn, tio, 0, size);
}

static void set_cork(struct socket *sock, int on)
{
	int opt = on;
	mm_segment_t oldfs;

	oldfs = get_fs();
	set_fs(get_ds());
	sock->ops->setsockopt(sock, SOL_TCP, TCP_CORK, (void *)&opt, sizeof(opt));
	set_fs(oldfs);
}

void cmnd_release(struct iscsi_cmnd *cmnd, int force)
{
	struct iscsi_cmnd *req, *rsp;
	int is_last = 0;

	if (!cmnd)
		return;

/* 	eprintk("%x %lx %d\n", cmnd_opcode(cmnd), cmnd->flags, force); */

	req = cmnd->req;
	is_last = cmnd_final(cmnd);

	if (force) {
		while (!list_empty(&cmnd->pdu_list)) {
			rsp = list_entry(cmnd->pdu_list.next, struct iscsi_cmnd, pdu_list);
			list_del_init(&rsp->list);
			list_del(&rsp->pdu_list);
			iscsi_cmnd_remove(rsp);
		}
	} else
		if (cmnd_queued(cmnd))
			iscsi_scsi_dequeuecmnd(cmnd);

	if (cmnd_hashed(cmnd))
		cmnd_remove_hash(cmnd);

	list_del_init(&cmnd->pdu_list);
	iscsi_cmnd_remove(cmnd);

	if (is_last) {
		assert(!force);
		assert(req);
		cmnd_release(req, 0);
	}

	return;
}

void cmnd_tx_start(struct iscsi_cmnd *cmnd)
{
	struct iscsi_conn *conn = cmnd->conn;
	struct iovec *iop;

	dprintk(D_GENERIC, "%p:%x\n", cmnd, cmnd_opcode(cmnd));
	assert(cmnd);
	iscsi_cmnd_set_length(&cmnd->pdu);

	set_cork(conn->sock, 1);

	conn->write_iop = iop = conn->write_iov;
	iop->iov_base = &cmnd->pdu.bhs;
	iop->iov_len = sizeof(cmnd->pdu.bhs);
	iop++;
	conn->write_size = sizeof(cmnd->pdu.bhs);

	switch (cmnd_opcode(cmnd)) {
	case ISCSI_OP_NOOP_IN: {
		cmnd_set_sn(cmnd, 1);
		cmnd_send_pdu(conn, cmnd);
		break;
	}
	case ISCSI_OP_SCSI_RSP:
		cmnd_set_sn(cmnd, 1);
		cmnd_send_pdu(conn, cmnd);
		break;
	case ISCSI_OP_SCSI_TASK_MGT_RSP:
		cmnd_set_sn(cmnd, 1);
		break;
	case ISCSI_OP_TEXT_RSP:
		cmnd_set_sn(cmnd, 1);
		break;
	case ISCSI_OP_SCSI_DATA_IN:
	{
		struct iscsi_data_in_hdr *rsp = (struct iscsi_data_in_hdr *)&cmnd->pdu.bhs;
		u32 offset;

		cmnd_set_sn(cmnd, (rsp->flags & ISCSI_FLG_FINAL) ? 1 : 0);
		offset = rsp->buffer_offset;
		rsp->buffer_offset = cpu_to_be32(offset);
		iscsi_cmnd_send_pdu(conn, cmnd->tio, offset, cmnd->pdu.datasize);
		break;
	}
	case ISCSI_OP_LOGOUT_RSP:
		cmnd_set_sn(cmnd, 1);
		break;
	case ISCSI_OP_R2T:
		cmnd_set_sn(cmnd, 0);
		cmnd->pdu.bhs.sn = cpu_to_be32(conn->stat_sn);
		break;
	case ISCSI_OP_ASYNC_MSG:
		cmnd_set_sn(cmnd, 1);
		break;
	case ISCSI_OP_REJECT:
		cmnd_set_sn(cmnd, 1);
		cmnd_send_pdu(conn, cmnd);
		break;
	default:
		eprintk("unexpected cmnd op %x\n", cmnd_opcode(cmnd));
		break;
	}

	iop->iov_len = 0;
	// move this?
	conn->write_size = (conn->write_size + 3) & -4;
	iscsi_dump_pdu(&cmnd->pdu);
}

void cmnd_tx_end(struct iscsi_cmnd *cmnd)
{
	struct iscsi_conn *conn = cmnd->conn;

	dprintk(D_GENERIC, "%p:%x\n", cmnd, cmnd_opcode(cmnd));
	switch (cmnd->pdu.bhs.opcode) {
	case ISCSI_OP_NOOP_IN:
	case ISCSI_OP_SCSI_RSP:
	case ISCSI_OP_SCSI_TASK_MGT_RSP:
	case ISCSI_OP_TEXT_RSP:
	case ISCSI_OP_R2T:
	case ISCSI_OP_ASYNC_MSG:
	case ISCSI_OP_REJECT:
	case ISCSI_OP_SCSI_DATA_IN:
		break;
	case ISCSI_OP_LOGOUT_RSP:
		dprintk(D_GENERIC, "%u\n", conn->cid);
		conn_close(conn);
		break;
	default:
		eprintk("unexpected cmnd op %x\n", cmnd_opcode(cmnd));
		break;
	}

	cmnd->state = ISCSI_STATE_WRITE_DONE;
	list_del_init(&cmnd->list);
	set_cork(conn->sock, 0);
}

/**
 * Push the command for execution.
 * This functions reorders the commands.
 * Called from the read thread.
 *
 * iscsi_session_push_cmnd - 
 * @cmnd: ptr to command
 */

static void iscsi_session_push_cmnd(struct iscsi_cmnd *cmnd)
{
	struct iscsi_session *session = cmnd->conn->session;
	struct list_head *entry;
	u32 cmd_sn;

	dprintk(D_GENERIC, "%p:%x %u,%u\n",
		cmnd, cmnd_opcode(cmnd), cmnd->pdu.bhs.sn, session->exp_cmd_sn);
	if (cmnd->state != ISCSI_STATE_START)
		eprintk("unexpected state %lx of cmnd %p\n", cmnd->state, cmnd);

	if (cmnd->pdu.bhs.opcode & ISCSI_OP_IMMEDIATE) {
		iscsi_cmnd_execute(cmnd);
		return;
	}

	cmd_sn = cmnd->pdu.bhs.sn;
	if (cmd_sn == session->exp_cmd_sn) {
		while (1) {
			session->exp_cmd_sn = ++cmd_sn;
			iscsi_cmnd_execute(cmnd);

			if (list_empty(&session->pending_list))
				break;
			cmnd = list_entry(session->pending_list.next, struct iscsi_cmnd, list);
			if (cmnd->pdu.bhs.sn != cmd_sn)
				break;
			list_del_init(&cmnd->list);
		}
	} else {
		if (before(cmd_sn, session->exp_cmd_sn)) /* close the conn */
			eprintk("unexpected cmd_sn (%u,%u)\n", cmd_sn, session->exp_cmd_sn);

		if (after(cmd_sn, session->exp_cmd_sn + MAX_QUEUED_CMD_NR))
			eprintk("too large cmd_sn (%u,%u)\n", cmd_sn, session->exp_cmd_sn);

		list_for_each(entry, &session->pending_list) {
			struct iscsi_cmnd *tmp = list_entry(entry, struct iscsi_cmnd, list);
			if (before(cmd_sn, tmp->pdu.bhs.sn))
				break;
		}
		cmnd->state = ISCSI_STATE_PUSHED;
		assert(list_empty(&cmnd->list));

		list_add_tail(&cmnd->list, entry);
	}
}

static int check_segment_length(struct iscsi_cmnd *cmnd)
{
	struct iscsi_conn *conn = cmnd->conn;
	struct iscsi_param *param = &conn->session->param;

	if (cmnd->pdu.datasize > param->max_data_pdu_length) {
		eprintk("too lond data %x %u %u\n", cmnd_itt(cmnd),
			cmnd->pdu.datasize, param->max_data_pdu_length);

		if (get_pgcnt(cmnd->pdu.datasize, 0) > ISCSI_CONN_IOV_MAX) {
			conn_close(conn);
			return -EINVAL;
		}
	}

	return 0;
}

void cmnd_rx_start(struct iscsi_cmnd *cmnd)
{
	struct iscsi_conn *conn = cmnd->conn;
	int err = 0;

	iscsi_dump_pdu(&cmnd->pdu);

	if (check_segment_length(cmnd) < 0)
		return;

	switch (cmnd_opcode(cmnd)) {
	case ISCSI_OP_NOOP_OUT:
		err = handle_noop_out(conn, cmnd);
		break;
	case ISCSI_OP_SCSI_CMD:
		if (!(err = cmnd_insert_hash(cmnd)))
			handle_scsi_cmnd(conn, cmnd);
		break;
	case ISCSI_OP_SCSI_TASK_MGT_MSG:
		err = cmnd_insert_hash(cmnd);
		break;
	case ISCSI_OP_SCSI_DATA_OUT:
		handle_data_out(conn, cmnd);
		break;
	case ISCSI_OP_LOGOUT_CMD:
		err = cmnd_insert_hash(cmnd);
		break;
	case ISCSI_OP_TEXT_CMD:
	case ISCSI_OP_SNACK_CMD:
		err = -ISCSI_REASON_UNSUPPORTED_COMMAND;
		break;
	default:
		err = -ISCSI_REASON_UNSUPPORTED_COMMAND;
		break;
	}

	if (err < 0) {
		eprintk("%x %x %d\n", cmnd_opcode(cmnd), cmnd_itt(cmnd), err);
		iscsi_cmnd_reject(cmnd, -err);
	}
}

void cmnd_rx_end(struct iscsi_cmnd *cmnd)
{
	struct iscsi_conn *conn = cmnd->conn;
	struct timer_list *timer = &cmnd->timer;

	timer->data = (unsigned long) cmnd;
	timer->function = dump_cmnd;
	timer->expires = jiffies + 3 * HZ;
	add_timer(timer);

	dprintk(D_GENERIC, "%p:%x\n", cmnd, cmnd_opcode(cmnd));
	switch (cmnd_opcode(cmnd)) {
	case ISCSI_OP_SCSI_REJECT:
	case ISCSI_OP_NOOP_OUT:
	case ISCSI_OP_SCSI_CMD:
	case ISCSI_OP_SCSI_TASK_MGT_MSG:
	case ISCSI_OP_TEXT_CMD:
	case ISCSI_OP_LOGOUT_CMD:
		cmnd->state = ISCSI_STATE_START;
		iscsi_session_push_cmnd(cmnd);
		break;
	case ISCSI_OP_SCSI_DATA_OUT:
		finish_data_out(conn, cmnd);
		break;
	case ISCSI_OP_SNACK_CMD:
		break;
	case ISCSI_OP_PDU_REJECT: {
		/* rejected cmnd must not be pushed, so exp_cmd_sn doesn't increase */
		struct iscsi_cmnd *rsp_cmnd = iscsi_cmnd_get_rsp_cmnd(cmnd);
		iscsi_cmnd_init_write(rsp_cmnd);
		break;
	}
	case ISCSI_OP_DATA_REJECT:
		cmnd_release(cmnd, 0);
		break;
	default:
		eprintk("unexpected cmnd op %x\n", cmnd_opcode(cmnd));
		BUG();
		break;
	}
}

static int iscsi_init(void)
{
	printk(KERN_INFO "iSCSI Enterprise Target Software - version %s\n", IET_VERSION_STRING);

	ctr_major = register_chrdev(0, ctr_name, &ctr_fops);
	if (ctr_major < 0) {
		eprintk("failed to register the control device %d\n", ctr_major);
		return ctr_major;
	}

	iet_procfs_init();

	iscsi_cmnd_cache = kmem_cache_create("iscsi_cmnd", sizeof(struct iscsi_cmnd),
					     0, 0, NULL, NULL);
	tio_init();

	iotype_init();

	return 0;
}

static void iscsi_exit(void)
{
	unregister_chrdev(ctr_major, ctr_name);

	iet_procfs_exit();

	tio_exit();

	iotype_exit();

	kmem_cache_destroy(iscsi_cmnd_cache);
}

module_init(iscsi_init);
module_exit(iscsi_exit);

MODULE_LICENSE("GPL");
