/* Interface to the i2c tuner, audio, and encoder on a matrox capture card.
 *
 * (C) Copyright 2004 Kevin O'Connor <kevin@koconnor.net>
 *
 * This file may be distributed under the terms of the GNU GPL license.
 *
 * Numerous people have contributed to this work; see the README file.
 */

#include <linux/i2c.h>
// Must include i2c.h before media/*.h files or you get warnings in media/id.h
#include <media/tuner.h>
#include <media/audiochip.h>

#include "ks0127.h"
#include "mgatv.h"

// ubiquitous dprintk macro
#define dprintk(fmt, args...) printk(fmt "\n" , ##args)
//#define dprintk(fmt, args...) do { } while (0)

const static struct v4l2_input VideoInputs[] = {
	{
		.index    = 0,
		.name     = "Composite",
		.type     = V4L2_INPUT_TYPE_CAMERA,
		.audioset = 1,
		.std      = V4L2_STD_PAL|V4L2_STD_NTSC,
	}, {
		.index    = 1,
		.name     = "S-Video",
		.type     = V4L2_INPUT_TYPE_CAMERA,
		.audioset = 1,
		.std      = V4L2_STD_PAL|V4L2_STD_NTSC,
	}, {
		.index    = 2,
		.name     = "Television",
		.type     = V4L2_INPUT_TYPE_TUNER,
		.audioset = 1,
		.tuner    = 0,
		.std      = V4L2_STD_PAL|V4L2_STD_NTSC,
	}
};

// Default cropping settings.
static const struct v4l2_rect DefCrop[] = {
	{
		// NTSC
		.left   = 15,
		.top    = 0,
		.width  = 690,
		.height = 476,
	}, {
		// PAL
		.left   = 0,
		.top    = 0,
		.width  = 720,
		.height = 584,
	}
};

static const struct v4l2_queryctrl QueryCtls[] = {
	{
		.id            = V4L2_CID_BRIGHTNESS,
		.name          = "Brightness",
		.minimum       = -127,
		.maximum       = 128,
		.step          = 1,
		.default_value = 0,
		.type          = V4L2_CTRL_TYPE_INTEGER,
	}, {
		.id            = V4L2_CID_CONTRAST,
		.name          = "Contrast",
		.minimum       = -127,
		.maximum       = 128,
		.step          = 1,
		.default_value = 0,
		.type          = V4L2_CTRL_TYPE_INTEGER,
	}, {
		.id            = V4L2_CID_SATURATION,
		.name          = "Saturation",
		.minimum       = -127,
		.maximum       = 128,
		.step          = 1,
		.default_value = 0,
		.type          = V4L2_CTRL_TYPE_INTEGER,
	}, {
		.id            = V4L2_CID_HUE,
		.name          = "Hue",
		.minimum       = -127,
		.maximum       = 128,
		.step          = 1,
		.default_value = 0,
		.type          = V4L2_CTRL_TYPE_INTEGER,
	}, {
		.id            = V4L2_CID_AUDIO_MUTE,
		.name          = "Mute",
		.minimum       = 0,
		.maximum       = 1,
		.type          = V4L2_CTRL_TYPE_BOOLEAN,
	}
};

static void
push_cropping(struct mgatv *card, const struct v4l2_rect *crop)
{
	int x;

	memcpy(&card->crop, crop, sizeof(card->crop));
	x = card->crop.left*2;
	encoder_cmd(card, KS0127_SET_HAVB, &x);
	x = (card->crop.width - CAPWIDTH) * 2 + x;
	encoder_cmd(card, KS0127_SET_HAVE, &x);
}

static int
setinput(struct mgatv *card, int input, int standard)
{
	// XXX - disable/reenable video on input change?
//	if (card->videoin_enabled)
//		return -EBUSY;

	if (input != -1) {
		int soundinput = AUDIO_EXTERN_2;
		card->input = input;
		// XXX - translate to ks0127 ids?
		encoder_cmd(card, KS0127_SET_INPUT, &input);
		if (input == 2)
			soundinput = AUDIO_TUNER;
		audio_cmd(card, AUDC_SET_INPUT, &soundinput);
	}
	if (standard != -1) {
		struct video_channel vc;
		int std = (standard == 0) ? KS_STD_NTSC : KS_STD_PAL;
		encoder_cmd(card, KS0127_SET_STANDARD, &std);
		if (card->standard != standard) {
			card->standard = standard;
			push_cropping(card, &DefCrop[standard]);
		}
		vc.norm = (standard == 0) ? VIDEO_MODE_NTSC : VIDEO_MODE_PAL;
		audio_cmd(card, VIDIOCSCHAN, &vc);
		tuner_cmd(card, VIDIOCSCHAN, &vc);
	}

	return 0;
}

static int
get_mute(struct mgatv *card)
{
	struct video_audio v;
	audio_cmd(card, VIDIOCGAUDIO, &v);
	return (v.flags & VIDEO_AUDIO_MUTE) != 0;
}

static void
set_mute(struct mgatv *card, int mute)
{
	struct video_audio v;
	audio_cmd(card, VIDIOCGAUDIO, &v);
	if (mute)
		v.flags |= VIDEO_AUDIO_MUTE;
	else
		v.flags &= ~VIDEO_AUDIO_MUTE;
	audio_cmd(card, VIDIOCSAUDIO, &v);
}

int
mgatv_controls_ioctl(struct mgatv *card, unsigned int cmd, void *arg)
{
	int ret;

	switch (cmd) {
	case VIDIOC_ENUMINPUT: {
		struct v4l2_input *input = arg;
		if (input->index >= ARRAY_SIZE(VideoInputs))
			return -EINVAL;
		memcpy(input, &VideoInputs[input->index], sizeof(*input));
		// XXX - fill in status info?
		return 0;
	}
	case VIDIOC_G_INPUT:
	{
		int *input = arg;
		*input = card->input;
		dprintk("vidioc_g_input input=%d", *input);
		return 0;
	}
	case VIDIOC_S_INPUT:
	{
		int *i = arg;
		dprintk("vidioc_s_input input=%d", *i);
		if (*i < 0 || *i >= ARRAY_SIZE(VideoInputs))
			return -EINVAL;
		ret = setinput(card, *i, -1);
		return ret;
	}
	case VIDIOC_ENUMSTD:
	{
		struct v4l2_standard *std = arg;
		int index = std->index;
		if (index == 0)
			v4l2_video_std_construct(arg, V4L2_STD_NTSC_M, "NTSC");
		else if (index == 1)
			v4l2_video_std_construct(arg, V4L2_STD_PAL, "PAL");
		else
			return -EINVAL;
		return 0;
	}
	case VIDIOC_G_STD:
	{
		v4l2_std_id *std = arg;
		if (card->standard == 0)
			*std = V4L2_STD_NTSC_M;
		else
			*std = V4L2_STD_PAL;
		dprintk("vidioc_g_std std=%d", (int)*std);
		return 0;
	}
	case VIDIOC_S_STD:
	{
		v4l2_std_id *std = arg;
		int newstd;
		dprintk("vidioc_s_std std=%d", (int)*std);
		if (*std == V4L2_STD_NTSC_M)
			newstd = 0;
		else if (*std == V4L2_STD_PAL)
			newstd = 1;
		else
			return -EINVAL;
		ret = setinput(card, -1, newstd);
		return ret;
	}
	case VIDIOC_QUERYSTD:
	{
		// Try to autodetect the input's picture format.
		v4l2_std_id *std = arg;
		int s = KS_STD_AUTO;
		encoder_cmd(card, KS0127_SET_STANDARD, &s);
		encoder_cmd(card, KS0127_GET_STATUS, &s);
		if (s >> 4)
			*std = V4L2_STD_NTSC_M;
		else
			*std = V4L2_STD_PAL;
		return 0;
	}
	case VIDIOC_G_TUNER: {
                struct v4l2_tuner *t = arg;
		if (t->index != 0)
			return -EINVAL;
		memset(t, 0, sizeof(*t));
                strcpy(t->name, "Television");
                t->type       = V4L2_TUNER_ANALOG_TV;
                t->rangehigh  = 0xffffffffUL;
                t->capability = V4L2_TUNER_CAP_NORM;
                t->rxsubchans = V4L2_TUNER_SUB_MONO | V4L2_TUNER_SUB_STEREO;
		t->audmode    = V4L2_TUNER_MODE_STEREO;
		// XXX - autodetect stereo/mono?
		return 0;
	}
	case VIDIOC_S_TUNER: {
                struct v4l2_tuner *t = arg;
		if (t->index != 0)
			return -EINVAL;
		// XXX - Change audio stereo/mono
		return 0;
	}
	case VIDIOC_G_FREQUENCY: {
		struct v4l2_frequency *f = arg;
		if (f->tuner != 0)
			return -EINVAL;
                f->type       = V4L2_TUNER_ANALOG_TV;
		f->frequency  = card->freq;
		return 0;
	}
	case VIDIOC_S_FREQUENCY: {
		struct v4l2_frequency *f = arg;
		if (f->tuner != 0)
			return -EINVAL;
		card->freq = f->frequency;
		tuner_cmd(card, VIDIOCSFREQ, &f->frequency);
		audio_cmd(card, VIDIOCSFREQ, 0);
		return 0;
	}

	case VIDIOC_CROPCAP: {
		struct v4l2_cropcap *cap = arg;
		cap->bounds.left = 0;
		cap->bounds.top = 0;
		cap->bounds.width = CAPWIDTH;
		cap->bounds.height = CAPHEIGHT;
		memcpy(&cap->defrect, &DefCrop[card->standard]
		       , sizeof(cap->defrect));
		cap->pixelaspect.numerator = 11;
		cap->pixelaspect.denominator = 10;
		return 0;
	}
	case VIDIOC_G_CROP: {
		struct v4l2_crop *crop = arg;
		memcpy(&crop->c, &card->crop, sizeof(crop->c));
		return 0;
	}
	case VIDIOC_S_CROP: {
		struct v4l2_crop *crop = arg;
		dprintk("vidioc_s_crop %dx%d+%d+%d"
			, crop->c.width, crop->c.height
			, crop->c.left, crop->c.top);
		// XXX
//		if (card->videoin_enabled)
//			return -EBUSY;
		crop->c.left = max(min(crop->c.left, CAPWIDTH), 0);
		crop->c.top = max(min(crop->c.top, CAPHEIGHT), 0);
		crop->c.width = max(min(crop->c.width, CAPWIDTH), 0);
		crop->c.height = max(min(crop->c.height, CAPHEIGHT), 0);
		push_cropping(card, &crop->c);
		return 0;
	}

	case VIDIOC_QUERYCTRL: {
		struct v4l2_queryctrl *q = arg;
		int i;

		for (i=0; i<ARRAY_SIZE(QueryCtls); i++) {
			const struct v4l2_queryctrl *item = &QueryCtls[i];
			if (item->id != q->id)
				continue;
			memcpy(q, item, sizeof(*q));
			return 0;
		}
		return -EINVAL;
	}
	case VIDIOC_G_CTRL: {
		struct v4l2_control *c = arg;
		switch (c->id) {
		case V4L2_CID_BRIGHTNESS:
			c->value = card->brightness;
			break;
		case V4L2_CID_CONTRAST:
			c->value = card->contrast;
			break;
		case V4L2_CID_SATURATION:
			c->value = card->saturation;
			break;
		case V4L2_CID_HUE:
			c->value = card->hue;
			break;
		case V4L2_CID_AUDIO_MUTE:
			c->value = get_mute(card);
			break;
		default:
			return -EINVAL;
		}
		return 0;
	}
	case VIDIOC_S_CTRL: {
		struct v4l2_control *c = arg;
		switch (c->id) {
		case V4L2_CID_BRIGHTNESS:
			card->brightness = c->value;
			encoder_cmd(card, KS0127_SET_BRIGHTNESS, &c->value);
			break;
		case V4L2_CID_CONTRAST:
			card->contrast = c->value;
			encoder_cmd(card, KS0127_SET_CONTRAST, &c->value);
			break;
		case V4L2_CID_SATURATION:
			card->saturation = c->value;
			encoder_cmd(card, KS0127_SET_SATURATION, &c->value);
			break;
		case V4L2_CID_HUE:
			card->hue = c->value;
			encoder_cmd(card, KS0127_SET_HUE, &c->value);
			break;
		case V4L2_CID_AUDIO_MUTE:
			set_mute(card, c->value);
			break;
		default:
			return -EINVAL;
		}
		return 0;
	}

	}
	return -ENOIOCTLCMD;
}

int
mgatv_controls_attach(struct mgatv *card)
{
	int tuner;

	// Setup the tuner type.
	switch (card->tuner->addr) {
	case 0x60:
	case 0x63:
		tuner = TUNER_PHILIPS_SECAM;
		card->standard = 1;
		break;
	case 0x61:
		tuner = TUNER_PHILIPS_NTSC;
		break;
	case 0x62:
		tuner = TUNER_PHILIPS_PAL_I;
		card->standard = 1;
		break;
	default:
		printk(KERN_WARNING "mgatv: Unknown tuner type\n");
		return -ENOSYS;
	}
	tuner_cmd(card, TUNER_SET_TYPE, &tuner);

	// Set default cropping.
	push_cropping(card, &DefCrop[card->standard]);

	// XXX - Set the audio to default to on?!
	{
		int i = AUDIO_EXTERN_2;
		audio_cmd(card, AUDC_SET_INPUT, &i);
	}

	return 0;
}

void
mgatv_controls_detach(struct mgatv *card)
{
}
