/* V4L2 interface for Matrox video capture cards.
 *
 * (C) Copyright 2004 Kevin O'Connor <kevin@koconnor.net>
 *
 * This file may be distributed under the terms of the GNU GPL license.
 *
 * Numerous people have contributed to this work; see the README file.
 */

#include <linux/moduleparam.h>

#include "../drivers/video/matrox/matroxfb_base.h"
#include "../drivers/video/matrox/matroxfb_maven.h"
#include "mgatv.h"
#include "ks0127.h"

// Driver version (as reported in VIDIOC_QUERYCAP ioctl)
#define MGATV_VERSION KERNEL_VERSION(0, 1, 0)

// Module parameters
static int video_nr = -1;
module_param(video_nr, int, 0);
MODULE_PARM_DESC(video_nr, "Force a specific v4l minor number");

static int fbmemuse = 2 * 1024 * 1024;
module_param(fbmemuse, int, 0);
MODULE_PARM_DESC(fbmemuse, "The amount of framebuffer memory that may be used");

// ubiquitous dprintk macro
#undef dprintk // drivers/video/matrox/matroxfb_base.h defines it..
#define dprintk(fmt, args...) printk(fmt "\n" , ##args)
//#define dprintk(fmt, args...) do { } while (0)


/*********************************************************************
 * Video4Linux file callbacks
 *********************************************************************/
static struct mgatv *MinorMap[256];

// Handle a user call to open() on a V4l2 device.
static int
mga_open(struct inode *inode, struct file *file)
{
	int minor = iminor(inode);
	struct mgatv_user *user;

	if (minor >= ARRAY_SIZE(MinorMap) || !MinorMap[minor])
		return -ENODEV;

	user = kmalloc(sizeof(*user), GFP_KERNEL);
	if (!user)
		return -ENOMEM;
	memset(user, 0, sizeof(*user));
	user->card = mgatv_get(MinorMap[minor]);

	file->private_data = user;

	return 0;
}

static int
mga_close(struct inode *inode, struct file *file)
{
	struct mgatv_user *user = file->private_data;
	struct mgatv *card = user->card;
	down(&card->lock);
	mgatv_overlay_close(user);
	mgatv_video_close(user);
	up(&card->lock);
	mgatv_put(card);
	kfree(user);
	return 0;
}

static int
mga_do_ioctl(struct inode *inode, struct file *file
	     , unsigned int cmd, void *arg)
{
	struct mgatv_user *user = file->private_data;
	struct mgatv *card = user->card;
	int ret;

	down(&card->lock);

	switch (cmd) {
	case VIDIOC_QUERYCAP: {
		struct v4l2_capability *cap = arg;
		dprintk("vidioc_querycap");
		memset(cap, 0, sizeof(*cap));
		strcpy(cap->driver, "mgatv");
                strlcpy(cap->card, card->v4l.name, sizeof(cap->card));
                sprintf(cap->bus_info, "PCI:%s", pci_name(card->pci));
		cap->version = MGATV_VERSION;
		cap->capabilities = (V4L2_CAP_VIDEO_CAPTURE
				     | V4L2_CAP_STREAMING
				     | V4L2_CAP_TUNER
				     | V4L2_CAP_VIDEO_OVERLAY);
		ret = 0;
		goto done;
	}
	}

	ret = mgatv_controls_ioctl(card, cmd, arg);
	if (ret != -ENOIOCTLCMD)
		goto done;

	ret = mgatv_videoin_ioctl(user, cmd, arg, file->f_flags & O_NONBLOCK);
	if (ret != -ENOIOCTLCMD)
		goto done;

	ret = mgatv_overlay_ioctl(user, cmd, arg);
	if (ret != -ENOIOCTLCMD)
		goto done;

	ret = -EINVAL;

 done:
	up(&card->lock);
	return ret;
}

#define MGATV_FORCECHROMA _IOR('v' , BASE_VIDIOCPRIVATE+8, int)

static int
mga_ioctl(struct inode *inode, struct file *file
	  , unsigned int cmd, unsigned long arg)
{
	if (cmd == MGATV_FORCECHROMA) {
		struct mgatv_user *user = file->private_data;
		struct mgatv *card = user->card;
		int ret;
		down(&card->lock);
		ret = mgatv_overlay_forcechroma(card, (void *)arg);
		up(&card->lock);
		return ret;
	}
	return video_usercopy(inode, file, cmd, arg, mga_do_ioctl);
}

static int
mga_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct mgatv_user *user = file->private_data;
	struct mgatv *card = user->card;
	int ret;
	down(&card->lock);
	ret = mgatv_mmap(card, vma);
	up(&card->lock);
	return ret;
}

static unsigned int
mga_poll(struct file *file, poll_table *wait)
{
	struct mgatv_user *user = file->private_data;
	struct mgatv *card = user->card;
	int ret;
	down(&card->lock);
	ret = mgatv_poll(card, file, wait);
	up(&card->lock);
	return ret;
}

static void
v4l_release_callback(struct video_device *vfd)
{
	struct mgatv *card = container_of(vfd, struct mgatv, v4l);
	mgatv_put(card);
}

// V4l File operations
static struct file_operations file_ops = {
	.owner    = THIS_MODULE,
	.open     = mga_open,
	.release  = mga_close,
	.ioctl    = mga_ioctl,
	.llseek   = no_llseek,
	.mmap     = mga_mmap,
	.poll     = mga_poll,
};


/*********************************************************************
 * Card setup and shutdown
 *********************************************************************/

// Setup the available part of the framebuffer.
static int
setup_fb_size(struct mgatv *card, struct matrox_fb_info *minfo)
{
	void *fbptr = vaddr_va(minfo->video.vbase);
	void *winptr = fbptr + minfo->video.len - fbmemuse;
	int adjsize;

	winptr = (void *) PAGE_ALIGN((unsigned long)winptr);
	adjsize = fbptr + minfo->video.len - winptr;
	if (winptr < fbptr || adjsize <= 0) {
		printk(KERN_WARNING "mgatv: fbmemsize parameter error.\n");
		return -EINVAL;
	}

	card->fbphysaddr = minfo->video.base;
	card->fbmem = winptr;
	card->fbmemsize = adjsize;
	card->fbwinoffset = winptr - fbptr;
	card->regs = vaddr_va(minfo->mmio.vbase);

	dprintk("setupfb - fbptr=0x%p fbsize=%d winptr=0x%p winsize=%d"
		" offset=%d"
		, fbptr, minfo->video.len, card->fbmem, card->fbmemsize
		, card->fbwinoffset);

	return 0;
}

// Return the i2c maven adapter of a matrox card.
static struct i2c_adapter *
get_maven(struct matrox_fb_info *minfo)
{
	// HACK - this info isn't exported, but is needed..
	struct matroxfb_dh_maven_info {
		struct i2c_bit_adapter  maven;
		struct i2c_bit_adapter  ddc1;
		struct i2c_bit_adapter  ddc2;
	};
	int i;

	for (i=0; i<minfo->drivers_count; i++)
		if (minfo->drivers[i]
		    && strcmp(minfo->drivers[i]->name, "i2c-matroxfb") == 0) {
			struct matroxfb_dh_maven_info *p;
			p = minfo->drivers_data[i];
			return &p->maven.adapter;
		}
	return NULL;
}

// Find and reserve the i2c components on an mga card.
static int
find_i2c_clients(struct mgatv *card, struct matrox_fb_info *minfo)
{
	struct i2c_adapter *adap;
	struct i2c_client *client;

	adap = get_maven(minfo);
	if (!adap) {
		printk(KERN_WARNING "mgatv: Unable to find maven i2c bus\n");
		return -ENODEV;
	}

	// Iterate through all the clients on the maven adapter.
	down(&adap->clist_lock);
	list_for_each_entry(client, &adap->clients, list) {
		switch (client->driver->id) {
		case I2C_DRIVERID_MSP3400:
			if (! i2c_use_client(client))
				card->msp3400 = client;
			break;
		case I2C_DRIVERID_TUNER:
			if (! i2c_use_client(client))
				card->tuner = client;
			break;
		case I2C_DRIVERID_KS0127:
			if (! i2c_use_client(client))
				card->ks0127 = client;
			break;
		}
	}
	up(&adap->clist_lock);

	// Verify all the necessary components were found.
	if (! card->msp3400) {
		printk(KERN_WARNING "mgatv: Unable to locate audio controller\n");
		goto out;
	}
	if (! card->tuner) {
		printk(KERN_WARNING "mgatv: Unable to locate video tuner\n");
		goto out;
	}
	if (! card->ks0127) {
		printk(KERN_WARNING "mgatv: Unable to locate video grabber\n");
		goto out;
	}

	return 0;
 out:
	if (card->msp3400)
		i2c_release_client(card->msp3400);
	if (card->tuner)
		i2c_release_client(card->tuner);
	if (card->ks0127)
		i2c_release_client(card->ks0127);
	return -ENODEV;
}

// Card setup.
static void *
card_probe(struct matrox_fb_info *minfo)
{
	static int card_nr;
	struct mgatv *card;
	int ret;

	// Allocate and initialize mgatv card structure.
	card = kmalloc(sizeof(*card), GFP_KERNEL);
	if (!card)
		return NULL;
	memset(card, 0, sizeof(*card));
	init_MUTEX(&card->lock);
	spin_lock_init(&card->irqlock);
	atomic_set(&card->refcount, 1);
	card->pci = pci_dev_get(minfo->pcidev);

	// Determine the part of the framebuffer that is available
	ret = setup_fb_size(card, minfo);
	if (ret)
		goto out1;

	// Locate and reserve i2c video capture control devices.
	ret = find_i2c_clients(card, minfo);
	if (ret)
		goto out1;

	// Initialize the i2c components
	ret = mgatv_controls_attach(card);
	if (ret)
		goto out2;

	// Initialize the video input grabbing code.
	ret = mgatv_videoin_attach(card);
	if (ret)
		goto out3;

	// Init overlay.
	mgatv_overlay_attach(card);

	// Register the v4l video_device structure.
	sprintf(card->v4l.name, "mgatv-%u", ++card_nr);
	card->v4l.fops = &file_ops;
	card->v4l.release = v4l_release_callback;
	ret = video_register_device(&card->v4l, VFL_TYPE_GRABBER, video_nr);
	if (ret) {
		printk(KERN_WARNING "mgatv: Unable to register v4l device.\n");
		goto out5;
	}
	if (card->v4l.minor >= ARRAY_SIZE(MinorMap)) {
		printk(KERN_WARNING "mgatv: v4l device minor out of range.\n");
		goto out6;
	}
	MinorMap[card->v4l.minor] = card;

	dprintk("mgatv card %p registration complete", card);

	return card;
 out6:
	video_unregister_device(&card->v4l);
 out5:
	mgatv_videoin_detach(card);
 out3:
	mgatv_controls_detach(card);
 out2:
	i2c_release_client(card->msp3400);
	i2c_release_client(card->tuner);
	i2c_release_client(card->ks0127);
 out1:
	pci_dev_put(card->pci);
	kfree(card);
	return NULL;
}

// Callback for when last reference to an mgatv struct is dropped.
void
mgatv_release(struct mgatv *card)
{
	dprintk("mgatv card %p release call", card);

	i2c_release_client(card->msp3400);
	i2c_release_client(card->tuner);
	i2c_release_client(card->ks0127);
	pci_dev_put(card->pci);
	if (card->bufs)
		kfree(card->bufs);
	kfree(card);
}

// MatroxFB callback - called when module is removed.
static void
card_remove(struct matrox_fb_info *minfo, void *data)
{
	struct mgatv *card = data;
	int minor = card->v4l.minor;

	dprintk("mgatv card %p remove call", card);

	video_unregister_device(&card->v4l);
	MinorMap[minor] = NULL;
	mgatv_videoin_detach(card);
	mgatv_controls_detach(card);
	mgatv_put(card);
}

static struct matroxfb_driver Mfb = {
	.name   = "Matrox TV Capture",
	.probe  = card_probe,
	.remove = card_remove
};


/*********************************************************************
 * Module startup/shutdown
 *********************************************************************/

// Module initialization.
static int
mgatv_init(void)
{
	int ret;
	ret = matroxfb_register_driver(&Mfb);
	return ret;
}

static void
mgatv_exit(void)
{
	matroxfb_unregister_driver(&Mfb);
}

MODULE_DESCRIPTION("Matrox TV capture driver");
MODULE_LICENSE("GPL");

module_init(mgatv_init);
module_exit(mgatv_exit);
