#include <linux/i2c.h>
#include <linux/videodev.h>
#include <asm/io.h>

// Maximum capture width/height and the capture Bytes Per Pixel.
enum {
	CAPWIDTH  = 720,
	CAPHEIGHT = 626,
	CAPBPP    = 2,
};

struct bufqueue;

// Information for a Matrox capture card:
struct mgatv {
	struct video_device v4l;
	atomic_t refcount;

	// Mutex serializing between kernel and irq handlers.
	spinlock_t irqlock;

	// Pointer to the memory mapped registers of the card.
	void *regs;

	// Area of framebuffer the driver is allowed to access.
	unsigned long fbphysaddr;
	char *fbmem;
	int fbmemsize, fbwinoffset;

	// The PCI device of the card.
	struct pci_dev *pci;

	// Mutex controlling access to this struct.
	struct semaphore lock;

	// The capture components connected to the i2c maven bus.
	struct i2c_client *msp3400;
	struct i2c_client *ks0127;
	struct i2c_client *tuner;

	// Current active input (0=composite, 1=svideo, 2=tuner)
	int input;
	// Picture format (0=ntsc, 1=pal)
	int standard;
	// Decoder color settings.
	int brightness, hue, contrast, saturation;
	// Tuner settings.
	int freq;
	// Requested capture width/height;
	struct v4l2_rect crop;

	// Videoin variables.
	int videoin_enabled;
	unsigned int capture_sequence;
	int isG400;
	struct bufqueue *bufs;
	int numbufs, bytesperline;
	struct list_head queued, filled;
	wait_queue_head_t readers;

	// Overlay variables.
	int overlay_enabled;
	int win_x, win_y, win_width, win_height;
	int bufskipbytes, start_polarity;
	u32 chromakey;
};

// Information for an open v4l device.
struct mgatv_user {
	struct mgatv *card;
	int user_enabled_video;
	int user_enabled_overlay;
};

// Helper functions for accessing mga registers.
static inline u32
mga_read(struct mgatv *card, int reg)
{
        return readl(card->regs + reg);
}

static inline void
mga_write(struct mgatv *card, int reg, u32 data)
{
        writel(data, card->regs + reg);
}

// Reference counting helper functions.
void mgatv_release(struct mgatv *card);

static inline struct mgatv *
mgatv_get(struct mgatv *card)
{
	atomic_inc(&card->refcount);
	return card;
}

static inline void
mgatv_put(struct mgatv *card)
{
	if (atomic_dec_and_test(&card->refcount))
		mgatv_release(card);
}

// I2C control helper functions.
static inline int
audio_cmd(struct mgatv *card, unsigned int cmd, void *arg)
{
	return card->msp3400->driver->command(card->msp3400, cmd, arg);
}

static inline int
encoder_cmd(struct mgatv *card, unsigned int cmd, void *arg)
{
	return card->ks0127->driver->command(card->ks0127, cmd, arg);
}

static inline int
tuner_cmd(struct mgatv *card, unsigned int cmd, void *arg)
{
	return card->tuner->driver->command(card->tuner, cmd, arg);
}

// controls.c
int mgatv_controls_ioctl(struct mgatv *card, unsigned int cmd, void *arg);
int mgatv_controls_attach(struct mgatv *card);
void mgatv_controls_detach(struct mgatv *card);

// videoin.c
int mgatv_videoin_ioctl(struct mgatv_user *user, unsigned int cmd, void *arg
			, int nonblocking);
int mgatv_mmap(struct mgatv *card, struct vm_area_struct *vma);
unsigned int mgatv_poll(struct mgatv *card, struct file *file
			, poll_table *wait);
void mgatv_video_close(struct mgatv_user *user);
int mgatv_grab_for_overlay(struct mgatv_user *user);
int mgatv_videoin_attach(struct mgatv *card);
void mgatv_videoin_detach(struct mgatv *card);

// overlay.c
int mgatv_overlay_ioctl(struct mgatv_user *user, unsigned int cmd, void *arg);
void mgatv_overlay_close(struct mgatv_user *user);
void mgatv_overlay_ackframe(struct mgatv *card, int isbottom, int fboffset);
int mgatv_overlay_forcechroma(struct mgatv *card, void *userptr);
void mgatv_overlay_attach(struct mgatv *card);
