/* Support for on-screen preview using MGA's Back-End-Scaler.
 *
 * (C) Copyright 2004 Kevin O'Connor <kevin@koconnor.net>
 *
 * This file may be distributed under the terms of the GNU GPL license.
 *
 * Numerous people have contributed to this work; see the README file.
 */

#include "mgatv.h"

// ubiquitous dprintk macro
#define dprintk(fmt, args...) printk(fmt "\n" , ##args)
//#define dprintk(fmt, args...) do { } while (0)

// Some mga registers
enum {
	CRTC    = 0x1fd4,
	CRTCEXT = 0x1fde,

	BESA1ORG    = 0x3d00,
	BESA2ORG    = 0x3d04,
	BESCTL      = 0x3d20,
	BESPITCH    = 0x3d24,
	BESHCOORD   = 0x3d28,
	BESVCOORD   = 0x3d2c,
	BESHISCAL   = 0x3d30,
	BESVISCAL   = 0x3d34,
	BESHSRCST   = 0x3d38,
	BESHSRCEND  = 0x3d3c,
	BESV1WGHT   = 0x3d48,
	BESV2WGHT   = 0x3d4c,
	BESHSRCLST  = 0x3d50,
	BESV1SRCLST = 0x3d54,
	BESV2SRCLST = 0x3d58,
	BESGLOBCTL  = 0x3dc0,
};

// Read a one byte register via writing to an mga control register.
static inline u8
__readb_reg(struct mgatv *card, int reg, int datareg, u8 subreg)
{
	writeb(subreg, card->regs + reg);
	return readb(card->regs + datareg);
}

static inline void
__writeb_reg(struct mgatv *card, int reg, int datareg, u8 subreg, u8 data)
{
	writeb(subreg, card->regs + reg);
	writeb(data, card->regs + datareg);
}

static inline u8
readb_crtc(struct mgatv *card, u8 subreg)
{
	return __readb_reg(card, CRTC, CRTC+1, subreg);
}

static inline u8
readb_crtcext(struct mgatv *card, u8 subreg)
{
	return __readb_reg(card, CRTCEXT, CRTCEXT+1, subreg);
}

static void
__get_fb_size(struct mgatv *card, int *width, int *height, int *interlace)
{
	int vtotal, htotal, tmp;

	// Width
	htotal  = (readb_crtc(card, 0x02) + 1) << 3; //XXX - reg 0x01?
        if (readb_crtcext(card, 0x01) & 0x02)
		htotal += 2048;

	// Height
        vtotal  = readb_crtc(card, 0x12) + 1;
        tmp  = readb_crtc(card, 0x07);
	if (tmp & 0x02)
		vtotal += 256;
	if (tmp & 0x40)
		vtotal += 512;
        if (readb_crtcext(card, 0x02) & 0x04)
		vtotal += 1024;

	*interlace = readb_crtcext(card, 0x00) & 0x80 ? 1 : 0;
	*width = htotal;
	*height = vtotal;
}

static void
get_fb_size(struct mgatv *card, int *width, int *height)
{
	int interlace, vtotal;

	spin_lock_irq(&card->irqlock);
	__get_fb_size(card, width, &vtotal, &interlace);
	spin_unlock_irq(&card->irqlock);
        if (interlace)
		vtotal *= 2;
	*height = vtotal;
}

// Indirect mga registers
enum {
	PALWTADD  = 0x3c00,
	X_DATAREG = 0x3c0a,

	XMULCTRL = 0x19,
	XKEYOPMODE    = 0x51,
	XCOLMSK0RED   = 0x52,
	XCOLMSK0GREEN = 0x53,
	XCOLMSK0BLUE  = 0x54,
	XCOLKEY0RED   = 0x55,
	XCOLKEY0GREEN = 0x56,
	XCOLKEY0BLUE  = 0x57,
};

static inline u8
readb_indirect(struct mgatv *card, u8 subreg)
{
	return __readb_reg(card, PALWTADD, X_DATAREG, subreg);
}

static inline void
writeb_indirect(struct mgatv *card, u8 subreg, u8 data)
{
	__writeb_reg(card, PALWTADD, X_DATAREG, subreg, data);
}

// Return the display's bytes-per-pixel.  Only needed to fool xawtv.
static int
get_fb_depth(struct mgatv *card)
{
	int ret = 4;
	spin_lock_irq(&card->irqlock);
	switch (readb_indirect(card, XMULCTRL) & 0x07) {
	case 0x00:
		ret = 1;
		break;
	case 0x01: case 0x02:
		ret = 2;
		break;
	case 0x03:
		ret = 3;
		break;
	}
	spin_unlock_irq(&card->irqlock);
	return ret;
}

static void
program_chromakey(struct mgatv *card)
{
	u32 chroma = card->chromakey;

	// Enable chroma key
	writeb_indirect(card, XKEYOPMODE, 1);

	// Turn off alpha blending
	writeb_indirect(card, XCOLMSK0RED, 0xff);
	writeb_indirect(card, XCOLMSK0GREEN, 0xff);
	writeb_indirect(card, XCOLMSK0BLUE, 0xff);

	// Set color values.
	writeb_indirect(card, XCOLKEY0RED, chroma >> 24);
	writeb_indirect(card, XCOLKEY0GREEN, chroma >> 16);
	writeb_indirect(card, XCOLKEY0BLUE, chroma >> 8);
}

static void
program_bes(struct mgatv *card)
{
	int destwidth = card->win_width, destheight = card->win_height;
	int besleft = card->win_x, besright = besleft + destwidth;
	int bestop = card->win_y, besbot = bestop + destheight;
	int offsetleft = 0, offsetright = 0, offsettop = 0, offsetbot = 0;
	int srcwidth = card->crop.width, srcheight = card->crop.height / 2;
	int srcpitch = card->bytesperline * 2;
	int cropwidth = srcwidth, cropheight = srcheight;
	int cropleft = 0, croptop = 0;
	int beshiscal, besviscal, beshsrcst, beshsrcend;
	int intrep, ifactor, Sc, vsrclst, val;
	int weight, upfield_weight;
	int displaywidth, displayheight;
	int interlace;

	// Make sure BES is disabled while programming it.
	mga_write(card, BESCTL, 0);
	// Reset globctl register to default.
	mga_write(card, BESGLOBCTL, 0);

	// Get screen dimensions.
	__get_fb_size(card, &displaywidth, &displayheight, &interlace);
	if (interlace) {
		// XXX - this is ugly.
		destheight /= 2;
		bestop /= 2;
	}

	dprintk("program buf=%dx%d@%d win=%dx%d+%d+%d disp=%dx%d@%d"
		, srcwidth, srcheight, srcpitch
		, destwidth, destheight, besleft, bestop
		, displaywidth, displayheight, interlace);

	// Check if part of the overlay window is offscreen.
	if (besleft < 0) {
		offsetleft = -besleft;
		besleft = 0;
	}
	if (besright > displaywidth) {
		offsetright = besright - displaywidth;
		besright = displaywidth;
	}
	if (bestop < 0) {
		offsettop = -bestop;
		bestop = 0;
	}
	if (besbot > displayheight) {
		offsetbot = besbot - displayheight;
		besbot = displayheight;
	}

	// Program BES window location.
	mga_write(card, BESHCOORD, besleft << 16 | (besright-1));
	mga_write(card, BESVCOORD, bestop << 16 | (besbot-1));

	mga_write(card, BESPITCH, srcpitch / CAPBPP);

	///////////////////////////// Horizontal calculations
	mga_write(card, BESHSRCLST, (srcwidth - 1) << 16);

	intrep = (destwidth == cropwidth) ? 0 : 1;
	ifactor = (((cropwidth - intrep) << 14) / (destwidth - intrep)
		   & 0x0007ffff);
	beshiscal = max(min(ifactor, (32 << 14) - 1), (srcwidth<<14) / 16384);
	mga_write(card, BESHISCAL, beshiscal << 2);

	beshsrcst = (cropleft << 14) + offsetleft * ifactor;
	mga_write(card, BESHSRCST, beshsrcst << 2);
	beshsrcend = beshsrcst + (destwidth - offsetleft
				  - offsetright - 1) * ifactor;
	mga_write(card, BESHSRCEND, beshsrcend << 2);

	///////////////////////////// Vertical calculations
	intrep = destheight == cropheight ? 0 : 1;
	ifactor = ((((cropheight - intrep) << 14) / (destheight - intrep))
		   & 0x0007ffff);
	besviscal = max(min(ifactor, (32 << 14) - 1), (srcheight<<14) / 16384);
	mga_write(card, BESVISCAL, besviscal << 2);

	if (besviscal >= (1 << 14)) {
		// Downscaling
		Sc = besviscal - (1 << 13);
		val = croptop + ((offsettop * besviscal + Sc) >> 14);
		if (offsettop == 0)
			weight = -((1<<13) + (besviscal & 0x3fff));
		else
			weight = -(1<<13) + offsettop * besviscal;
	} else {
		// Upscaling
		Sc = (((destheight - intrep) << 14)
		      / (2 * (cropheight - intrep))) & 0x7ffff;
		val = croptop + ((abs(offsettop - (Sc>>14)) * besviscal) >> 14);
		if (offsettop * besviscal >= 1<<13)
			weight = -(1<<13) + offsettop;
		else
			weight = -((1<<13) + offsettop * besviscal);
	}

	card->bufskipbytes = val * srcpitch;

	upfield_weight = offsettop * besviscal;
	mga_write(card, BESV1WGHT,
		  (((upfield_weight & 0x3FFF) << 2)
		   | (upfield_weight<0 ? 0x10000 : 0)));
	mga_write(card, BESV2WGHT,
		  ((weight & 0x3FFF) << 2) | (weight<0 ? 0x10000 : 0));

	// XXX - old driver has val only in downscaling??
	vsrclst = srcheight - 1 - (val&0x03ff);
	mga_write(card, BESV1SRCLST, vsrclst);
	mga_write(card, BESV2SRCLST, vsrclst);

	card->start_polarity = val&1;

	program_chromakey(card);
}

static int
stop_overlay(struct mgatv_user *user)
{
	struct mgatv *card = user->card;

	if (!user->user_enabled_overlay || !card->overlay_enabled)
		return -EINVAL;

	spin_lock_irq(&card->irqlock);
	mga_write(card, BESCTL, 0);
	card->overlay_enabled = 0;
	spin_unlock_irq(&card->irqlock);

	user->user_enabled_overlay = 0;
	mgatv_video_close(user);

	return 0;
}

static int
start_overlay(struct mgatv_user *user)
{
	struct mgatv *card = user->card;
	int ret;

	if (user->user_enabled_overlay) {
		// If overlay is already enabled just reprogram the bes.
		spin_lock_irq(&card->irqlock);
		program_bes(card);
		spin_unlock_irq(&card->irqlock);
		return 0;
	}

	if (card->overlay_enabled)
		return -EBUSY;

	ret = mgatv_grab_for_overlay(user);
	if (ret)
		return ret;

	spin_lock_irq(&card->irqlock);
	program_bes(card);
	card->overlay_enabled = 1;
	spin_unlock_irq(&card->irqlock);

	user->user_enabled_overlay = 1;

	return 0;
}

int
mgatv_overlay_ioctl(struct mgatv_user *user, unsigned int cmd, void *arg)
{
	struct mgatv *card = user->card;

	switch (cmd) {
	case VIDIOC_S_FBUF:
		// XXX - just ignore it.
		dprintk("vidioc_s_fbuf");
		return 0;
	case VIDIOC_G_FBUF: {
		struct v4l2_framebuffer *fb = arg;
		int width, height;
		dprintk("vidioc_g_fbuf");
		get_fb_size(card, &width, &height);
		memset(fb, 0, sizeof(*fb));
		fb->capability = (V4L2_FBUF_CAP_EXTERNOVERLAY
				  | V4L2_FBUF_CAP_CHROMAKEY);
		fb->flags = V4L2_FBUF_FLAG_CHROMAKEY;
		fb->fmt.width = width;
		fb->fmt.height = height;
		// xawtv bug - setup base address to work around it.
		fb->base = (void *)card->fbphysaddr;
		fb->fmt.bytesperline = width * get_fb_depth(card);
		return 0;
	}
	case VIDIOC_TRY_FMT:
	case VIDIOC_S_FMT: {
		struct v4l2_format *fmt = arg;
		int width, height;
		dprintk("vidioc_%s_fmt overlay"
			, cmd==VIDIOC_S_FMT ? "S" : "TRY");
		if (fmt->type != V4L2_BUF_TYPE_VIDEO_OVERLAY)
			return -ENOIOCTLCMD;
		get_fb_size(card, &width, &height);
		if (fmt->fmt.win.w.left > width)
			fmt->fmt.win.w.left = width;
		if (fmt->fmt.win.w.top > height)
			fmt->fmt.win.w.top = height;
		if (fmt->fmt.win.w.width < 0)
			fmt->fmt.win.w.width = 0;
		if (fmt->fmt.win.w.width > width)
			fmt->fmt.win.w.width = width;
		if (fmt->fmt.win.w.height < 0)
			fmt->fmt.win.w.height = 0;
		if (fmt->fmt.win.w.height > height)
			fmt->fmt.win.w.height = height;
		fmt->fmt.win.field = V4L2_FIELD_INTERLACED;
		if (cmd == VIDIOC_TRY_FMT)
			return 0;
		card->win_x = fmt->fmt.win.w.left;
		card->win_y = fmt->fmt.win.w.top;
		card->win_width = fmt->fmt.win.w.width;
		card->win_height = fmt->fmt.win.w.height;
		card->chromakey = fmt->fmt.win.chromakey;
		return 0;
	}
	case VIDIOC_G_FMT: {
		struct v4l2_format *fmt = arg;
		dprintk("vidioc_g_fmt overlay");
		if (fmt->type != V4L2_BUF_TYPE_VIDEO_OVERLAY)
			return -ENOIOCTLCMD;
		memset(fmt, 0, sizeof(*fmt));
		fmt->type = V4L2_BUF_TYPE_VIDEO_OVERLAY;
		fmt->fmt.win.w.left = card->win_x;
		fmt->fmt.win.w.top = card->win_y;
		fmt->fmt.win.w.width = card->win_width;
		fmt->fmt.win.w.height = card->win_height;
		fmt->fmt.win.field = V4L2_FIELD_INTERLACED;
		fmt->fmt.win.chromakey = card->chromakey;
		// XXX - fmt.win.colorspace?
		return 0;
	}
	case VIDIOC_OVERLAY: {
		int *flag = arg;
		dprintk("overlay=%d", *flag);
		if (*flag)
			return start_overlay(user);
		return stop_overlay(user);
	}
	}
	return -ENOIOCTLCMD;
}

// Called when a v4l file is closed.
void
mgatv_overlay_close(struct mgatv_user *user)
{
	if (user->user_enabled_overlay)
		stop_overlay(user);
}

// Called from irq handler with card's spinlock held.
void
mgatv_overlay_ackframe(struct mgatv *card, int isbottom, int fboffset)
{
	int frame, reg;
	int polarity = card->start_polarity;

	if (isbottom) {
		frame = 1;
		reg = BESA2ORG;
	} else {
		frame = 0;
		reg = BESA1ORG;
	}
//	dprintk("overirq bot=%d frame=%d pos=0x%x", isbottom, frame
//		, fboffset + card->bufskipbytes);
	mga_write(card, reg, fboffset + card->bufskipbytes);
	// Set the next frame, the start polarities, enable the BES, enable
	// horizontal filter, enable vertical filter, and enable dithering.
	mga_write(card, BESCTL, (
			  (frame<<25) | (polarity<<6) | ((polarity^1)<<7)
			  | (1<<0) | (1<<10) | (1<<11) | (1<<18)));
}

// Special ioctl to force a new chroma setting.
int
mgatv_overlay_forcechroma(struct mgatv *card, void *userptr)
{
	struct chromainput {
		int type;
		u32 rgb;
	};
	struct chromainput data;
	int ret;

	ret = copy_from_user(&data, userptr, sizeof(data));
	if (ret)
		return -EFAULT;
	if (data.type == 1)
		card->chromakey = data.rgb;
	// XXX - add a mode to snoop a pixel from the framebuffer.
	spin_lock_irq(&card->irqlock);
	program_chromakey(card);
	spin_unlock_irq(&card->irqlock);
	return 0;
}

void
mgatv_overlay_attach(struct mgatv *card)
{
	// Set a default chromakey.
	card->chromakey = (1 << 24) | (1 << 16) | (254);
}
