/* Interface to the mga's video-input hardware
 *
 * (C) Copyright 2004 Kevin O'Connor <kevin@koconnor.net>
 *
 * This file may be distributed under the terms of the GNU GPL license.
 *
 * Numerous people have contributed to this work; see the README file.
 */

#include <linux/interrupt.h>
#include <linux/pci.h>

#include "mgatv.h"

// ubiquitous dprintk macro
#define dprintk(fmt, args...) printk(fmt "\n" , ##args)
//#define dprintk(fmt, args...) do { } while (0)

enum {
	// Maximum number of buffers that may be allocated.
	MAXBUFS = 32,
	// Minimum number of buffers that must be available to do capture.
	MINBUFS = 2,
};

struct bufqueue {
	struct list_head list;
	int fboffset;
	int gotfirsthalf : 1;
	struct v4l2_buffer buf;
};

static int
allocbufs(struct mgatv *card, struct v4l2_requestbuffers *req)
{
	int bufsize, numbufs, i;
	int width = card->crop.width, height = card->crop.height;

	// For now, ignore the user request and allocate as many buffers as
	// possible.

	if (card->videoin_enabled)
		return -EBUSY;

	if (card->bufs) {
		kfree(card->bufs);
		card->bufs = NULL;
		card->numbufs = 0;
		card->bytesperline = 0;
		INIT_LIST_HEAD(&card->queued);
		INIT_LIST_HEAD(&card->filled);
	}

	// XXX - width and bpl stuff needs to move into _s_fmt

	// Increase width to the card's required alignment.
	if (card->isG400) {
		width = (width + 7) & ~0x07;
	} else {
		width = (width + 3) & ~0x03;
		// G200 can't limit maximum vertical lines
		height = CAPHEIGHT;
	}

	bufsize = PAGE_ALIGN(width * height * CAPBPP);
	numbufs = card->fbmemsize / bufsize;
	dprintk("Allocating %d bufs (memsize=%d, bufsize=%d)"
		, numbufs, card->fbmemsize, bufsize);
	if (numbufs > MAXBUFS)
		numbufs = MAXBUFS;
	if (numbufs < MINBUFS)
		return -EINVAL;

	card->bufs = kmalloc(numbufs * sizeof(card->bufs[0]), GFP_KERNEL);
	if (!card->bufs)
		return -ENOMEM;
	card->bytesperline = width * CAPBPP;
	card->numbufs = numbufs;
	memset(card->bufs, 0, numbufs * sizeof(card->bufs[0]));

	for (i=0; i<numbufs; i++) {
		struct bufqueue *q = &card->bufs[i];
		struct v4l2_buffer *buf = &q->buf;
		q->fboffset = card->fbwinoffset + i * bufsize;
		INIT_LIST_HEAD(&q->list);
		buf->index = i;
		buf->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf->flags = V4L2_BUF_FLAG_MAPPED;
		buf->memory = V4L2_MEMORY_MMAP;
		buf->m.offset = i * bufsize;
		buf->length = bufsize;
	}

	req->count = numbufs;
	return 0;
}

int
mgatv_mmap(struct mgatv *card, struct vm_area_struct *vma)
{
	int ret;
	unsigned long start = vma->vm_pgoff * PAGE_SIZE;
	unsigned long len = vma->vm_end - vma->vm_start;

	if (start + len > card->fbmemsize)
		return -EINVAL;

	// XXX - clean up card->fbmem use
	dprintk("mmap call - %lu at %lu - pa=%lx pa+start=%lx"
		, len, start, card->fbphysaddr + card->fbwinoffset
		, card->fbphysaddr + card->fbwinoffset + start);

	ret = io_remap_page_range(vma, vma->vm_start
				  , card->fbphysaddr + card->fbwinoffset + start
				  , len, PAGE_SHARED);
	return ret;
}

unsigned int
mgatv_poll(struct mgatv *card, struct file *file, poll_table *wait)
{
	int dataready;

	poll_wait(file, &card->readers, wait);

	if (! card->videoin_enabled)
		return -EINVAL;

	spin_lock_irq(&card->irqlock);
	dataready = ! list_empty(&card->filled);
	spin_unlock_irq(&card->irqlock);

//	dprintk("poll (dataready=%d)", dataready);

	if (dataready)
		return POLLIN|POLLRDNORM;
	return 0;
}

static void
queuebuf(struct mgatv *card, struct bufqueue *q)
{
	q->buf.flags |= V4L2_BUF_FLAG_QUEUED;
	q->gotfirsthalf = 0;
	spin_lock_irq(&card->irqlock);
	list_add(&q->list, &card->queued);
	spin_unlock_irq(&card->irqlock);
}

// Ughh.. There has to be a better way...
static inline int
__dataready(struct mgatv *card)
{
	int ret;
	spin_lock_irq(&card->irqlock);
	ret = ! list_empty(&card->filled);
	spin_unlock_irq(&card->irqlock);
	return ret;
}

static int
trydequeue(struct mgatv *card, struct v4l2_buffer *buf, int nonblocking)
{
	struct bufqueue *q;

 restart:
	if (!card->videoin_enabled)
		return -EINVAL;

	spin_lock_irq(&card->irqlock);
	if (list_empty(&card->filled)) {
		int ret;
		spin_unlock_irq(&card->irqlock);
		if (nonblocking)
			return -EAGAIN;

		// Release the semaphore before sleeping.
		up(&card->lock);
		ret = wait_event_interruptible(card->readers
					       , __dataready(card));
		down(&card->lock);

		if (ret)
			return ret;
		goto restart;
	}
	q = list_entry(card->filled.prev, struct bufqueue, list);
	list_del(&q->list);
	spin_unlock_irq(&card->irqlock);

	q->buf.flags = V4L2_BUF_FLAG_MAPPED;
	memcpy(buf, &q->buf, sizeof(*buf));

	return 0;
}

// Some mga register numbers.
enum {
	VSTATUS    = 0x3e30,
	VICLEAR    = 0x3e34,
	VINCTL     = 0x3E1C,
	VINHEIGHT  = 0x3E3C,
	VBICOUNT   = 0x3E18,
	VINNEXTWIN = 0x3E18,
	VIEN       = 0x3E38,
	VINADDR0   = 0x3E10,
	VINADDR1   = 0x3E14,
	VINCTL0    = 0x3E00,
	VINCTL1    = 0x3E04,
};

static void
setwindows(struct mgatv *card, struct list_head *list)
{
	struct bufqueue *q = list_entry(list, struct bufqueue, list);
	int bpl = card->bytesperline, pitch;
	// The G400 measures pitch in 8 pixel groups, the G200 in 4 pixel
	// groups.  The buffers are interlaced, so the pitch must account
	// for two lines of video.
	if (card->isG400)
		pitch = bpl*2/CAPBPP/8;
	else
		pitch = bpl*2/CAPBPP/4;

	mga_write(card, VINADDR0, q->fboffset);
	mga_write(card, VINCTL0, (pitch & 0x1FF) << 3 | 0x01);

	mga_write(card, VINADDR1, q->fboffset + bpl);
	mga_write(card, VINCTL1, (pitch & 0x1FF) << 3 | 0x01);
}

static void
g400_enable_video(struct mgatv *card)
{
	int vb0, vb1, vbi0, vbi1;

	// Check for Marvel/RRG via ks0127 address:
	if (card->ks0127->addr == 0xDA>>1) {
		// Marvel
		// vb0 = 15; vbi0 = 0; vb1 = 4; vbi1 = 0;
		vb0 = 13; vbi0 = 0; vb1 = 2; vbi1 = 0;
	} else {
		// RRG
		if (card->standard == 0) {
			// NTSC
			vb0 = 19; vbi0 = 11; vb1 = 18; vbi1 = 10;
		} else {
			// PAL
			vb0 = 21; vbi0 = 7; vb1 = 20; vbi1 = 6;
		}
	}

	// Toggle vinen from 0 to 1
	mga_write(card, VINCTL, 0x0);
	mga_write(card, VINCTL, 0x1);

	// Set VINCTL: RP check, yuy2, video-ints on both fields, 1xMode
	mga_write(card, VINCTL, 1<<24 | 1<<16 | 1<<5 | 1<<4 | 0x01);

	// Set VINHEIGHT:
	mga_write(card, VINHEIGHT, (card->crop.height/2) & 0x3FF);

	// Set VBICOUNT:
	mga_write(card, VBICOUNT, vbi1<<24 | vb1<<16 | vbi0<<8 | vb0);
}

// XXX - This function is entirely untested.
static void
g200_enable_video(struct mgatv *card)
{
	int vbi0, vbi1;

	if (card->standard == 0) {
		// NTSC
		vbi0 = 11; vbi1 = 10;
	} else {
		// PAL
		vbi0 = 15; vbi1 = 14;
	}

	// Set VINCTL: RP check, vbi-task
	mga_write(card, VINCTL, 1<<24 | 1<<16 | vbi0<<12 | vbi1<<8 | 0x01);
}

static int
start_videoin(struct mgatv_user *user)
{
	struct mgatv *card = user->card;

	if (card->videoin_enabled)
		return -EBUSY;

	spin_lock_irq(&card->irqlock);

	if (list_empty(&card->queued))
		goto out;

	if (card->isG400)
		g400_enable_video(card);
	else
		g200_enable_video(card);

	// Enable Video-Input-VSync interrupt:
	mga_write(card, VIEN, 0x01);

	// Setup the grab windows:
	setwindows(card, card->queued.prev);

	card->videoin_enabled = 1;
	card->capture_sequence = 0;
	spin_unlock_irq(&card->irqlock);

	user->user_enabled_video = 1;

	return 0;
 out:
	spin_unlock_irq(&card->irqlock);
	return -EINVAL;
}

static int
stop_videoin(struct mgatv_user *user)
{
	struct mgatv *card = user->card;

	if (!user->user_enabled_video || !card->videoin_enabled)
		return -EINVAL;

	spin_lock_irq(&card->irqlock);

	// Disable video input engine
	mga_write(card, VINCTL, 0);

	// Clear Video-Input-VSync interrupt
	mga_write(card, VIEN, 0);

	// Empty both incoming and outgoing queues.
	INIT_LIST_HEAD(&card->queued);
	INIT_LIST_HEAD(&card->filled);

	card->videoin_enabled = 0;

	spin_unlock_irq(&card->irqlock);

	user->user_enabled_video = 0;

	// Wake any sleepers.
	wake_up_interruptible_all(&card->readers);

	return 0;
}

int
mgatv_videoin_ioctl(struct mgatv_user *user, unsigned int cmd, void *arg
		    , int nonblocking)
{
	struct mgatv *card = user->card;
	int ret;

	switch (cmd) {
	case VIDIOC_TRY_FMT:
	case VIDIOC_S_FMT:
		// XXX - for now just ignore user requests and hard code a
		// single catpure setup.
	case VIDIOC_G_FMT: {
		struct v4l2_format *fmt = arg;
		dprintk("vidioc_g_fmt");
		if (fmt->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
			return -ENOIOCTLCMD;
		memset(fmt, 0, sizeof(*fmt));
		fmt->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		//fmt->fmt.pix.width = card->crop.width;
		//Increase width so there is no end-of-line padding -
		//otherwise mplayer crashes..
		fmt->fmt.pix.width = card->bytesperline / CAPBPP;
		fmt->fmt.pix.height = card->crop.height;
		fmt->fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
		fmt->fmt.pix.field = V4L2_FIELD_INTERLACED;
		fmt->fmt.pix.bytesperline = card->bytesperline;
		fmt->fmt.pix.sizeimage = card->bytesperline * card->crop.height;
		// XXX - fmt.pix.colorspace?
		return 0;
	}
	case VIDIOC_REQBUFS: {
		struct v4l2_requestbuffers *req = arg;
		dprintk("vidioc_reqbufs");
		if (req->type != V4L2_BUF_TYPE_VIDEO_CAPTURE
		    || req->memory != V4L2_MEMORY_MMAP)
			return -EINVAL;
		ret = allocbufs(card, arg);
		return ret;
	}
	case VIDIOC_QUERYBUF: {
		struct v4l2_buffer *buf = arg;
		dprintk("vidioc_querybuf");
		if (buf->index >= card->numbufs)
			return -EINVAL;
		memcpy(buf, &card->bufs[buf->index].buf, sizeof(*buf));
		return 0;
	}
	case VIDIOC_QBUF: {
		struct v4l2_buffer *buf = arg;
		struct bufqueue *q;
//		dprintk("vidioc_qbuf %d", buf->index);
		if (buf->index >= card->numbufs)
			return -EINVAL;
		q = &card->bufs[buf->index];
		if (q->buf.flags != V4L2_BUF_FLAG_MAPPED)
			return -EINVAL;
		buf->flags = q->buf.flags;
		queuebuf(card, q);
		return 0;
	}
	case VIDIOC_DQBUF: {
		struct v4l2_buffer *buf = arg;
//		dprintk("vidioc_dqbuf");
		ret = trydequeue(card, buf, nonblocking);
		return ret;
	}
	case VIDIOC_STREAMON: {
		int *type = arg;
		dprintk("vidioc_streamon");
		if (*type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
			return -EINVAL;
		ret = start_videoin(user);
		return ret;
	}
	case VIDIOC_STREAMOFF: {
		int *type = arg;
		dprintk("vidioc_streamoff");
		if (*type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
			return -EINVAL;
		ret = stop_videoin(user);
		return ret;
	}
	}
	return -ENOIOCTLCMD;
}

// Called when a v4l file is closed.
void
mgatv_video_close(struct mgatv_user *user)
{
	if (user->user_enabled_video)
		stop_videoin(user);
}

int
mgatv_grab_for_overlay(struct mgatv_user *user)
{
	struct mgatv *card = user->card;
	struct v4l2_requestbuffers req;
	int i, ret;

	memset(&req, 0, sizeof(req));
	ret = allocbufs(card, &req);
	if (ret)
		return ret;

	for (i=0; i<card->numbufs; i++)
		queuebuf(card, &card->bufs[i]);

	ret = start_videoin(user);
	if (ret)
		return ret;
	// XXX - user can now stop videoin via ioctl!?

	return 0;
}


/*********************************************************************
 * Interrupts
 *********************************************************************/

static irqreturn_t
irq_handler(int irq, void *data, struct pt_regs *regs)
{
	struct bufqueue *q;
	struct mgatv *card = data;
	u32 vstatus;
	int isbottom;

	spin_lock(&card->irqlock);

	vstatus = mga_read(card, VSTATUS);

	// Check for VINVSYNC interrupt.
	if (! (vstatus & 0x01))
		goto nowork;

	// Acknowledge the VINVSYNC interrupt.
	mga_write(card, VICLEAR, 0x01);

//	dprintk("irq vstatus=%x capseq=%d viden=%d"
//		, vstatus, card->capture_sequence, card->videoin_enabled);

	if (! card->videoin_enabled)
		// Hrrm..  Interrupt disable race?
		goto nowork;

	isbottom = vstatus & (1<<8);

	// If the capture is just starting don't change windows.
	if (card->capture_sequence == 0) {
		card->capture_sequence = 1;
		goto done;
	}

	q = list_entry(card->queued.prev, struct bufqueue, list);

	// Program the BES if enabled.
	if (card->overlay_enabled) {
		int offset = q->fboffset;
		if (isbottom)
			offset += card->bytesperline;
		mgatv_overlay_ackframe(card, isbottom, offset);
	}

	// Handle frame.
	if (isbottom) {
		// Bottom field
		if (card->queued.prev == card->queued.next) {
			// Only one item in queue; can't dequeue it, so
			// leave things as they are.
//			printk("Queue empty - dropping frame %d\n"
//			       , card->capture_sequence);
			goto done;
		}
		// Record that the first part of an interlaced image was
		// captured.
		q->gotfirsthalf = 1;
		// Program next window.
		setwindows(card, q->list.prev);
	} else {
		// Top field.
		card->capture_sequence++;

		if (! q->gotfirsthalf) {
			// Didn't get first half of this image, so leave
			// things as they are.
//			printk("No firsthalf %d\n"
//			       , card->capture_sequence - 1);
			goto done;
		}
		// Full frame captured, move it to filled list.
		list_move(&q->list, &card->filled);
		q->buf.sequence = card->capture_sequence - 1;
		do_gettimeofday(&q->buf.timestamp);

		// Notify any waiters.
		wake_up_interruptible(&card->readers);

		// XXX - Ughh.
		if (card->overlay_enabled) {
			q->buf.flags |= V4L2_BUF_FLAG_QUEUED;
			q->gotfirsthalf = 0;
			list_add(&q->list, &card->queued);
		}
	}
 done:
	// Program the G200 next window register.
	if (!card->isG400)
		mga_write(card, VINNEXTWIN, isbottom ? 1 : 0);

	spin_unlock(&card->irqlock);
	return IRQ_HANDLED;
 nowork:
	spin_unlock(&card->irqlock);
	return IRQ_NONE;
}


/*********************************************************************
 * Card startup
 *********************************************************************/

int
mgatv_videoin_attach(struct mgatv *card)
{
	int ret;

	init_waitqueue_head(&card->readers);

	INIT_LIST_HEAD(&card->queued);
	INIT_LIST_HEAD(&card->filled);

	// Register the interrupt handler.
	ret = request_irq(card->pci->irq, irq_handler, SA_INTERRUPT | SA_SHIRQ
			  , "mgatv", card);
	if (ret) {
		printk(KERN_WARNING "mgatv: Unable to register interrupt\n");
		return -ENOSYS;
	}

	if (card->pci->device == PCI_DEVICE_ID_MATROX_G400)
		card->isG400 = 1;

	return 0;
}

void
mgatv_videoin_detach(struct mgatv *card)
{
	free_irq(card->pci->irq, card);
}
