/*
 * Prism54 USB driver
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/smp_lock.h>
#include <linux/completion.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/usb.h>
#include <linux/pci.h>
#include <linux/firmware.h>
#include <asm/uaccess.h>
#include "prism54_usb.h"
#include "isl_38xx.h"

void p54u_mdelay(int ms)
{
	int t = (ms * HZ + 500) / 1000;
	
	if(!t && ms)
		t = 1;
	
	set_current_state(TASK_UNINTERRUPTIBLE);
	while(t) {
		//printk("t: %i\n", t);
		t = schedule_timeout(t);
	}
	
	return;
}


void p54u_data_debug(struct net_device *netdev, unsigned int ep, void *_data, int len)
{
	struct p54u *p54u = netdev_priv(netdev);
	struct usb_device *usbdev = p54u->usbdev;
	unsigned char *data = _data;
	int k, l, in;
	
	in = ep & USB_DIR_IN;
	if(debug > 1) {
		for(k = 0; k < len; k+= 16) {
			if(k == 0) {
				printk(KERN_CRIT "[%s >%-6lld<%-6lld+%-6lld]     %02x %s 00000000:", in?"IN ":"OUT", 0LL, 0LL, 0LL, ep, in?"<-":"->");
			} else {
				printk(KERN_CRIT "                                   -> %08x:", k);
			}
			for (l = 0; (l < 16) && (k + l < len); ++l) {
				printk(" %02x", ((unsigned char *)data)[k + l]);
			}
			printk("\n");
		}
	}
}

int p54u_bulk_msg(struct net_device *netdev, unsigned int ep, void *_data, int len)
{
	struct p54u *p54u = netdev_priv(netdev);
	struct usb_device *usbdev = p54u->usbdev;
	unsigned char *data = _data;
	size_t alen;
	int pipe, err, k, l;

	p54u_data_debug(netdev, ep, data, len);
	
	pipe = usb_sndbulkpipe(usbdev, ep);
	err = usb_bulk_msg(usbdev, pipe, data, len, &alen, 2 * HZ);
	if(err)
		p54u_info("bulk submit failed: %i\n", err);
	
	return err;
}


u32 p54u_reg_rw(struct net_device *netdev, int write, int ep, int port, u32 addr, u32 val)
{
	struct p54u *p54u = netdev_priv(netdev);
	struct usb_device *usbdev = p54u->usbdev;
	struct p54u_reg reg;
	unsigned int pipe;
	size_t len;
	int err, i, epd;
	u64 t, t1, t2;
	
	reg.port = cpu_to_le16(port);
	reg.addr = cpu_to_le32(addr);
	reg.val = ((port & P54U_PORT_U32) == P54U_PORT_U32) ? cpu_to_le32(val) : cpu_to_le16(val);
	len = write ? P54U_REG_WSIZE : P54U_REG_RSIZE;
	pipe = usb_sndbulkpipe(usbdev, ep & USB_ENDPOINT_NUMBER_MASK);
	
	t1 = p54u_time();
	err = usb_bulk_msg(usbdev, pipe, &reg, len, &len, HZ);
	t2 = p54u_time();
	t = t2 - t1;
	
	p54u_data_debug(netdev, ep, &reg, len);

	if(err) {
		p54u_err("%s %02x %04x %08x %08x: failed: %i", write ? "Write": "Read", ep, port, addr, val, err);
		p54u->err = err;
		return 0;
	}
	
	if(write) {
//		udelay(ISL38XX_WRITEIO_DELAY);
		return 0;
	}
	
	pipe = usb_rcvbulkpipe(usbdev, ep & USB_ENDPOINT_NUMBER_MASK);
	len = sizeof(reg.val);
	
	t1 = p54u_time();
	err = usb_bulk_msg(usbdev, pipe, &reg.val, len, &len, HZ);
	t2 = p54u_time();
	t = t2 - t1;
	
	if(err) {
		p54u_err("Register read %02x %04x %08x %08x: failed: %i", ep, port, addr, val, err);
		p54u->err = err;
		return 0;
	}
	
	p54u_data_debug(netdev, ep | USB_DIR_IN, &reg.val, len);
	
	return ((port & P54U_PORT_U32) == P54U_PORT_U32) ? le32_to_cpu(reg.val) : le16_to_cpu(reg.val);
}


int p54u_wait_int(struct net_device *netdev)
{
	struct p54u *p54u = netdev_priv(netdev);
	struct usb_device *usbdev = p54u->usbdev;
	int err, pending;
	
	wait_for_completion(&p54u->int_rx.comp);
	printk("readi: %08x\n", p54u->pending);
	pending = p54u->pending;

	return pending;
}


void p54u_int_rx_cb(struct urb *urb, struct pt_regs *p)
{
	struct net_device *netdev = urb->context;
	struct p54u *p54u = netdev_priv(netdev);
	struct usb_device *usbdev = p54u->usbdev;
	int err, pending;
	u32 reg;
	u64 t;
	
	t = p54u_time();
	
	pending = *(int *)urb->transfer_buffer;
	p54u_data_debug(netdev, 0x8f, urb->transfer_buffer, urb->actual_length);
	p54u->pending = pending;
	complete(&p54u->int_rx.comp);
	
	return;
}

void p54u_data_rx_cb(struct urb *urb, struct pt_regs *p)
{
	struct net_device *netdev = urb->context;
	struct p54u *p54u = netdev_priv(netdev);
	struct usb_device *usbdev = p54u->usbdev;
	int err;
	
	p54u_data_debug(netdev, 0x81, urb->transfer_buffer, urb->actual_length);
	
	return;
}

void p54u_mgmt_rx_cb(struct urb *urb, struct pt_regs *p)
{
	struct net_device *netdev = urb->context;
	struct p54u *p54u = netdev_priv(netdev);
	struct usb_device *usbdev = p54u->usbdev;
	unsigned char *data;
	size_t len;
	int err, k, l;
	u64 t;
	
	p54u_data_debug(netdev, 0x82, urb->transfer_buffer, urb->actual_length);
	
	return;
}

