/*
 * voodoo.c - Part of v4l voodoo3 drivers
 *
 * $Id: voodoo.c,v 1.1.1.1 2004/05/23 04:46:51 scottsman Exp $
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <asm/io.h>
#include <linux/init.h>
#include <linux/videodev.h>
#include "voodoo.h"

#define V3_VERSION "0.1-Multistandard for V4L2"

/* Global data struct.
 * We can have only a single voodoo3 3500 TV card anyway,
 * because there aren't any dual AGP slot motherboards, right?
 */
struct voodoo_data *voodoo = NULL;

/* Module parameters */
#ifdef INCLUDE_I2C
int ddc = 1;
#endif
int debug = 0;
int model = 0;
int video_nr = -1;
int vbi_nr = -1;
int radio_nr = -1;

static struct conf_struct voodoo3_init_reg[] = {
	{V3REG_vidInFormat, -1, 0, 0},	/* Reserved - leave defaults                 */
	{V3REG_vidInFormat, 6, 3, 1},	/* 110 = 8bit YCbCr 4:2:2 UYVY               */
	{V3REG_vidInFormat, 0, 4, 4},	/* No De-interlacing                         */
	{V3REG_vidInFormat, -1, 7, 5},	/* Polarities - leave defaults               */
	{V3REG_vidInFormat, 1, 8, 8},	/* Brooktree tv-out                          */
	{V3REG_vidInFormat, 2, 10, 9},	/* VideoIn triple buffering (FIXME:BUGGY???) */
	{V3REG_vidInFormat, 0, 11, 11},	/* VideoInBuffer linear space                */
	{V3REG_vidInFormat, -1, 13, 12},	/* polarities - leave defaults               */
	{V3REG_vidInFormat, -1, 14, 14},	/* VMI-interface enable (not yet)            */
	{V3REG_vidInFormat, 1, 15, 15},	/* TV-Out-interface enable                   */
	{V3REG_vidInFormat, 0, 16, 16},	/* Genlock disable(VMI-slave)                */
	{V3REG_vidInFormat, 0, 17, 17},	/* Use vga timing                            */
	{V3REG_vidInFormat, -1, 18, 18},	/* Genlock source - with bit16=0 no effect   */
	{V3REG_vidInFormat, -1, 19, 19},	/* Use display_ena for driving tv-out        */
	{V3REG_vidInFormat, 0, 20, 20},	/* VideoIn Horizontal Decimation disable     */
	{V3REG_vidInFormat, 0, 21, 21},	/* VideoIn Vertical Decimation disable       */
	{V3REG_vidInFormat, -1, 31, 22},	/* Reserved - leave defaults                 */
	{V3REG_vidSerialParallelPort, 0, 0, 0},	/* VMI parallel port disable                 */
	{V3REG_vidSerialParallelPort, -1, 1, 1},	/* VMI Chip Select (default)                 */
	{V3REG_vidSerialParallelPort, -1, 4, 2},	/* VMI send/rcv data ctrl - leave defaults   */
	{V3REG_vidSerialParallelPort, 0, 5, 5},	/* VMI data output disable                   */
	{V3REG_vidSerialParallelPort, -1, 13, 6},	/* VMI send/rcv data - leave defaults        */
	{V3REG_vidSerialParallelPort, -1, 17, 14},	/* VMI send/rcv addrs - leave defaults       */
#ifdef INCLUDE_I2C
	{V3REG_vidSerialParallelPort, 1, 22, 18},	/* Activate ddc-interface(i2c for monitor)   */
	{V3REG_vidSerialParallelPort, 1, 27, 23},	/* Activate i2c-interface(i2c on card)       */
#else
	{V3REG_vidSerialParallelPort, -1, 22, 18},	/* Activate ddc-interface(i2c for monitor)   */
	{V3REG_vidSerialParallelPort, -1, 27, 23},	/* Activate i2c-interface(i2c on card)       */
#endif
	{V3REG_vidSerialParallelPort, 1, 28, 28},	/* VMI normal operation                      */
	{V3REG_vidSerialParallelPort, 1, 29, 29},	/* GPIO[1] - VMI sync_oe                     */
	{V3REG_vidSerialParallelPort, -1, 30, 30},	/* GPIO[2] - Input only                      */
	{V3REG_vidSerialParallelPort, 1, 31, 31},	/* TV out normal operation                   */
};

/* Detect whether a Voodoo3 can be found, and map its memory. */
static int v3tv_probe(void)
{
	struct pci_dev *dev = NULL;
	u16 subsys_id;

	while ((dev = pci_find_device(PCI_VENDOR_ID_3DFX,
				      PCI_DEVICE_ID_3DFX_VOODOO3, dev))) {
		subsys_id = dev->subsystem_device;
		if ((subsys_id == VOODOO3_MODEL_NTSC) ||
		    (subsys_id == VOODOO3_MODEL_PAL) ||
		    (subsys_id == VOODOO3_MODEL_SECAM))
			break;
	}

	if (!dev) {
		printk(KERN_INFO "v3tv: No Voodoo3 3500 TV found.\n");
		return -ENODEV;
	}

	printk(KERN_INFO "v3tv: Voodoo3 found.\n");
	if (!model)
		if (debug)
			printk(KERN_INFO
			       "v3tv: Autodetecting Voodoo3 model.\n");
	voodoo->model = (model) ? model : dev->subsystem_device;

	/* map Voodoo3 memory. */
	voodoo->io_base = ioremap_nocache(dev->resource[0].start &
					  PCI_BASE_ADDRESS_MEM_MASK, 0x200400);
	if (!voodoo->io_base)
		return (-ENOMEM);

	voodoo->fb_base = ioremap(dev->resource[1].start &
				  PCI_BASE_ADDRESS_MEM_MASK, 1024 * 1024 * 16);
	if (!voodoo->fb_base) {
		iounmap(voodoo->io_base);
		return -ENOMEM;
	}

	if (debug)
		printk
		    ("v3tv: Registers mapped @ 0x%lx, Frame buffer mapped @ 0x%lx\n",
		     (unsigned long)voodoo->io_base,
		     (unsigned long)voodoo->fb_base);

	voodoo->dev = dev;

	return 0;
}

static void v3tv_cleanup_module(void)
{
	if (voodoo->initialized & I_v4l2)
		v3tv_video_cleanup(voodoo);

	v3tv_i2c_setup(0);

	if (voodoo->io_base)
		iounmap(voodoo->io_base);

	if (voodoo->fb_base)
		iounmap(voodoo->fb_base);

	kfree(voodoo);
}

static int v3tv_init_module(void)
{
	int res;

	printk("v3tv: version %s (%s)\n", V3_VERSION, __DATE__);

	voodoo = kmalloc(sizeof(struct voodoo_data), GFP_KERNEL);
	if (!voodoo)
		return -ENOMEM;

	memset(voodoo, 0, sizeof(struct voodoo_data));	/* Paranoia: Clear data */

	if ((res = v3tv_probe())) {
		printk("v3tv: Voodoo3 not detected, module not inserted.\n");
		v3tv_cleanup_module();
		return res;
	}

//	init_MUTEX(&voodoo->lock);
	/* Configure the chip to initial state */
	v3tv_conf(voodoo3_init_reg, CONF_SIZE(voodoo3_init_reg));

	/* activate i2c */
	if (v3tv_i2c_setup(1))
		return -EINVAL;

	if (v3tv_video_init(voodoo))
		return -EINVAL;

	voodoo->initialized |= I_v4l2;

	/* set up overlay and videobuffers */
	v3tv_overlay_setup(voodoo);

	return 0;
}

#ifdef MODULE
MODULE_AUTHOR("Juha Valkama <juhisv@surfree.com>");
MODULE_DESCRIPTION("Voodoo3 3500 TV Video4Linux driver");
MODULE_LICENSE("GPL");

MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debug flag (0, 1 or 2)");
MODULE_PARM(model, "i");
MODULE_PARM_DESC(model, "Voodoo3 submodel ID (0x60, 0x61, or 0x62");
#ifdef INCLUDE_I2C
MODULE_PARM(ddc, "i");
MODULE_PARM_DESC(ddc,
		 "Set to 0 to disable DDC init (if you don't have a DDC monitor)");
#endif

MODULE_PARM(video_nr,"i");
MODULE_PARM(vbi_nr,"i");
MODULE_PARM(radio_nr,"i");

module_init(v3tv_init_module);
module_exit(v3tv_cleanup_module);
#endif				/* MODULE */

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
