/*
 * USB Orinoco driver
 *
 * Copyright (c) 2003 Manuel Estrada Sainz <ranty@debian.org>
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License
 * at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and
 * limitations under the License.
 *
 * Alternatively, the contents of this file may be used under the
 * terms of the GNU General Public License version 2 (the "GPL"), in
 * which case the provisions of the GPL are applicable instead of the
 * above.  If you wish to allow the use of your version of this file
 * only under the terms of the GPL and not to allow others to use your
 * version of this file under the MPL, indicate your decision by
 * deleting the provisions above and replace them with the notice and
 * other provisions required by the GPL.  If you do not delete the
 * provisions above, a recipient may use your version of this file
 * under either the MPL or the GPL.  
 *
 * Queueing code based on linux-wlan-ng 0.2.1-pre5
 *
 * Copyright (C) 1999 AbsoluteValue Systems, Inc.  All Rights Reserved.
 *
 * 	The license is the same as above.
 *
 * Initialy based on USB Skeleton driver - 0.7
 *
 * Copyright (c) 2001 Greg Kroah-Hartman (greg@kroah.com)
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation; either version 2 of
 *	the License, or (at your option) any later version.
 *
 * NOTE: The original USB Skeleton driver is GPL, but all that code is
 * gone so MPL/GPL applies.
 */

#define DRIVER_NAME "orinoco_usb"
#define PFX DRIVER_NAME ": "

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/fcntl.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/smp_lock.h>
#include <linux/usb.h>
#include <linux/timer.h>
#include <asm/io.h>

#include <linux/netdevice.h>
#include <linux/if_arp.h>
#include <linux/etherdevice.h>
#include <linux/wireless.h>

#include "firmware.h"
#include "orinoco.h"

/* Forward compatibility macros */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
# define return_probe(priv_data, retval) return retval
# define usb_get_desc(iface) (iface.desc)
# define fw_dev_param (&interface->dev)
#endif

struct ez_usb_fw {
	u16 size;
	u16 card_variant_offset;
	u8 *code;
};

/*
 * If EZUSB_FW_INCLUDED is defined, the firmware is hardcoded into
 * the driver.  Use get_ezusb_fw script to generate orinoco_usb_fw.h and
 * copy it to the same directory as orinoco_usb.c.
 *
 * If EZUSB_FW_INCLUDED is not defined, the firmware is loaded at the
 * runtime using hotplug.  Use the same get_ezusb_fw script to generate
 * file orinoco_ezusb_fw, copy it to the hotplug firmware directory
 * (typically /usr/lib/hotplug/firmware) and make sure that you have
 * hotplug installed and enabled in the kernel.
 */
/* #define EZUSB_FW_INCLUDED 1 */

#ifdef EZUSB_FW_INCLUDED
/* Header with the firmware */
#include "orinoco_usb_fw.h"
static struct ez_usb_fw firmware = {
	.size = sizeof(ezusb_fw),
	.code = ezusb_fw,
};
#else				/* !EZUSB_FW_INCLUDED */
static struct ez_usb_fw firmware = {
	.size = 0,
	.code = NULL,
};
#endif				/* !EZUSB_FW_INCLUDED */

#ifdef CONFIG_USB_DEBUG
static int debug = 1;
#else
static int debug;
#endif

/* Debugging macros */
#undef dbg
#define dbg(format, arg...) \
	do { if (debug) printk(KERN_DEBUG PFX "%s: " format "\n", \
			       __FUNCTION__ , ## arg); } while (0)
#undef err
#define err(format, arg...) \
	do { printk(KERN_ERR PFX format "\n", ## arg); } while (0)

/* Module paramaters */
module_param(debug, int, 0644);
MODULE_PARM_DESC(debug, "Debug enabled or not");

/*
 * Under some conditions, the card gets stuck and stops paying attention
 * to the world (i.e. data communication stalls) until we do something to
 * it.  Sending an INQ_TALLIES command seems to be enough and should be
 * harmless otherwise.  This behaviour has been observed when using the
 * driver on a systemimager client during installation.  In the past a
 * timer was used to send INQ_TALLIES commands when there was no other
 * activity, but it was troublesome and was removed.
 */

static int force_unsupported = 0;
module_param(force_unsupported, int, 0644);
MODULE_PARM_DESC(force_unsupported, "Try to handle untested hardware.");

static int default_card_variant = 0;
module_param(default_card_variant, int, 0644);
MODULE_PARM_DESC(default_card_variant,
		 "Card variant to use as a first guess");
/* Most of this Device/Vendor pairs are untested */

#define USB_COMPAQ_VENDOR_ID		0x049f	/* Compaq Computer Corp. */
#define USB_COMPAQ_WL215_ID		0x001f	/* Compaq WL215 USB Adapter */
#define USB_HP_WL215_ID			0x0082	/* Compaq WL215 USB Adapter */
#define USB_COMPAQ_W200_ID		0x0076	/* Compaq W200 USB Adapter */

#define USB_MELCO_VENDOR_ID		0x0411
#define USB_BUFFALO_AIRSTA_ID		0x000B	/* Buffalo airstation usb */

#define USB_LUCENT_VENDOR_ID		0x047E	/* Lucent Technologies */
#define USB_LUCENT_ORINOCO_ID		0x0300	/* Lucent/Agere Orinoco USB
						   Client */
#define USB_AVAYA8_VENDOR_ID		0x0D98
#define USB_AVAYAE_VENDOR_ID		0x0D9E
#define USB_AVAYA_WIRELESS_ID		0x0300	/* Avaya Wireless USB Card */

#define USB_AGERE_VENDOR_ID		0x0D4E	/* Agere Systems */
#define USB_AGERE_ORINOCO_ID		0x1000	/*Wireless USB Card Model 0801 */
#define USB_AGERE_REBRANDED_ID		0x047A	/* WLAN USB Card */

#define USB_ELSA_VERNDOR_ID		0x05CC
#define USB_ELSA_AIRLANCER_ID		0x3100	/* ELSA AirLancer USB-11 */

#define USB_LEGEND_VENDOR_ID		0x0E7C
#define USB_LEGEND_JOYNET_ID		0x0300	/* LEGEND Joynet WLAN USB Card */

#define USB_SAMSUNG_VENDOR_ID		0x04E8
#define USB_SAMSUNG_SEW2001U_ID		0x5002	/* Samsung SEW-2001u Card */
#define USB_SANSUNG_SEW2001U2_ID	0x5B11	/* Samsung SEW-2001u Card */

#define USB_IGATE_VENDOR_ID		0x0681
#define USB_IGATE_IGATE_11M_ID		0x0012	/* I-GATE 11M USB Card */

#define USB_FUJITSU_VENDOR_ID		0x0BF
#define USB_FUJITSU_E1100_ID		0x1002	/* Fujitsu Siemens Computers
						   connect2AIR WLAN E-1100 USB */
#define USB_2WIRE_VENDOR_ID		0x1630
#define USB_2WIRE_WIRELESS_ID		0xff81	/* 2Wire Wireless USB adapter */


#define EZUSB_REQUEST_FW_TRANS		0xA0
#define EZUSB_REQUEST_TRIGER		0xAA
#define EZUSB_REQUEST_TRIG_AC		0xAC
#define EZUSB_CPUCS_REG			0x7F92

#define EZUSB_RID_TX			0x0700
#define EZUSB_RID_RX			0x0701
#define EZUSB_RID_INIT1			0x0702
#define EZUSB_RID_ACK			0x0710
#define EZUSB_RID_DOCMD			0x0860

/* Recognize info frames */
#define EZUSB_IS_INFO(id)		((id >= 0xF000) && (id <= 0xF2FF))

#define EZUSB_MAGIC			0x0210

#define EZUSB_FRAME_DATA		1
#define EZUSB_FRAME_CONTROL		2

#define DEF_TIMEOUT			(3*HZ)

#define BULK_BUF_SIZE			2048

struct ezusb_packet {
	u16 magic;		/* 0x0210 */
	u8 req_reply_count;
	u8 ans_reply_count;
	u16 frame_type;		/* 0x01 for data frames, 0x02 otherwise */
	u16 size;		/* transport size */
	u16 crc;		/* CRC up to here */
	u16 hermes_len;
	u16 hermes_rid;
	u8 data[0];
} __attribute__ ((packed));

/* table of devices that may work with this driver */
/* We will refuse to touch any untested device unless force_unsupported is set
 */
static struct usb_device_id ezusb_table[] = {
	{USB_DEVICE(USB_COMPAQ_VENDOR_ID, USB_COMPAQ_WL215_ID)},
	{USB_DEVICE(USB_COMPAQ_VENDOR_ID, USB_HP_WL215_ID)},
	{USB_DEVICE(USB_COMPAQ_VENDOR_ID, USB_COMPAQ_W200_ID)},
	{USB_DEVICE(USB_MELCO_VENDOR_ID, USB_BUFFALO_AIRSTA_ID)},
	{USB_DEVICE(USB_LUCENT_VENDOR_ID, USB_LUCENT_ORINOCO_ID)},
	{USB_DEVICE(USB_AVAYA8_VENDOR_ID, USB_AVAYA_WIRELESS_ID)},
	{USB_DEVICE(USB_AVAYAE_VENDOR_ID, USB_AVAYA_WIRELESS_ID)},
	{USB_DEVICE(USB_AGERE_VENDOR_ID, USB_AGERE_ORINOCO_ID)},
	{USB_DEVICE(USB_ELSA_VERNDOR_ID, USB_ELSA_AIRLANCER_ID)},
	{USB_DEVICE(USB_LEGEND_VENDOR_ID, USB_LEGEND_JOYNET_ID)},
	{USB_DEVICE_VER(USB_SAMSUNG_VENDOR_ID, USB_SAMSUNG_SEW2001U_ID,
			0, 0)},
	{USB_DEVICE(USB_SAMSUNG_VENDOR_ID, USB_SANSUNG_SEW2001U2_ID)},
	{USB_DEVICE(USB_IGATE_VENDOR_ID, USB_IGATE_IGATE_11M_ID)},
	{USB_DEVICE(USB_FUJITSU_VENDOR_ID, USB_FUJITSU_E1100_ID)},
	{USB_DEVICE(USB_2WIRE_VENDOR_ID, USB_2WIRE_WIRELESS_ID)},
	{USB_DEVICE(USB_AGERE_VENDOR_ID, USB_AGERE_REBRANDED_ID)},
	{}			/* Terminating entry */
};

MODULE_DEVICE_TABLE(usb, ezusb_table);

/* Devices should be added to 'ezusb_table' and one of
 * ezusb_varian1_table or ezusb_varian2_table. Devices on the
 * 'ezusb_table' but not on any of 'ezusb_varian1_table' or
 * 'ezusb_varian2_table' will require force_unsupported=1.  */

/* Variant 1 has a PCMCIA card inside */
static struct usb_device_id ezusb_varian1_table[] = {
	{USB_DEVICE(USB_COMPAQ_VENDOR_ID, USB_COMPAQ_WL215_ID)},
	{USB_DEVICE(USB_COMPAQ_VENDOR_ID, USB_HP_WL215_ID)},
	{USB_DEVICE(USB_COMPAQ_VENDOR_ID, USB_COMPAQ_W200_ID)},
	{USB_DEVICE(USB_LUCENT_VENDOR_ID, USB_LUCENT_ORINOCO_ID)},
	{USB_DEVICE(USB_AVAYA8_VENDOR_ID, USB_AVAYA_WIRELESS_ID)},
	{USB_DEVICE(USB_MELCO_VENDOR_ID, USB_BUFFALO_AIRSTA_ID)},
	{USB_DEVICE(USB_AGERE_VENDOR_ID, USB_AGERE_ORINOCO_ID)},
	{USB_DEVICE(USB_AGERE_VENDOR_ID, USB_AGERE_REBRANDED_ID)},
	{}
};

/* Variant 2 is all build in a single board */
static struct usb_device_id ezusb_varian2_table[] = {
	{}
};

/* Structure to hold all of our device specific stuff */
struct ezusb_priv {
	struct usb_device *udev;
	struct net_device *net_dev;
	struct semaphore sem;
	spinlock_t req_lock;
	struct list_head req_pending;
	struct list_head req_active;
	spinlock_t reply_count_lock;
	u16 hermes_reg_fake[0x40];
	u8 *bap_buf;
	struct urb *read_urb;
	int read_pipe;
	int write_pipe;
	u8 reply_count;
	u8 card_variant;
};

typedef enum {
	EZUSB_CTX_START,
	EZUSB_CTX_QUEUED,
	EZUSB_CTX_REQ_SUBMITTED,
	EZUSB_CTX_REQ_COMPLETE,
	EZUSB_CTX_RESP_RECEIVED,
	EZUSB_CTX_REQ_TIMEOUT,
	EZUSB_CTX_REQ_FAILED,
	EZUSB_CTX_RESP_TIMEOUT,
	EZUSB_CTX_REQSUBMIT_FAIL,
	EZUSB_CTX_COMPLETE,
} ezusb_state_t;

struct request_context {
	struct list_head list;
	atomic_t refcount;
	struct completion done;	/* Signals that CTLX is dead */
	int killed;
	struct urb *outurb;	/* OUT for req pkt */
	struct ezusb_priv *dev;
	struct ezusb_packet *buf;
	int buf_length;
	struct timer_list timer;	/* Timeout handling */
	ezusb_state_t state;	/* Current state */
	/* the RID that we will wait for */
	u16 out_rid;
	u16 in_rid;
};


/* Forward declarations */
static void ezusb_ctx_complete(struct request_context *ctx);
static void ezusb_req_queue_run(struct ezusb_priv *dev);
static void ezusb_bulk_in_callback(struct urb *urb, struct pt_regs *);


/* Debug functions */
static inline void hex_dump(const void *data_p, int size)
{
	int i;
	const u8 *data = data_p;

	printk(KERN_DEBUG "Dump at %p, length: %d\n", data, size);

	for (i = 0; i < size; i++) {
		if ((i & 0xf) == 0)
			printk(KERN_DEBUG "0x%04X: ", i);
		if ((i & 3) == 0)
			printk(" ");
		printk("%02X ", data[i]);
		if ((i & 0xf) == 0xf)
			printk("\n");
	}
	if ((i & 0xf) != 0)
		printk("\n");
}


static inline u8 ezusb_reply_inc(u8 count)
{
	if (count < 0x7F)
		return count + 1;
	else
		return 1;
}

static void ezusb_request_context_put(struct request_context *ctx)
{
	if (!atomic_dec_and_test(&ctx->refcount))
		return;

	WARN_ON(!ctx->done.done);
	BUG_ON(ctx->outurb->status == -EINPROGRESS);
	BUG_ON(timer_pending(&ctx->timer));
	usb_free_urb(ctx->outurb);
	kfree(ctx->buf);
	kfree(ctx);
}

static inline void ezusb_mod_timer(struct ezusb_priv *dev,
				   struct timer_list *timer,
				   unsigned long expire)
{
	if (!dev->udev)
		return;
	mod_timer(timer, expire);
}

static void ezusb_request_timerfn(u_long _ctx)
{
	struct request_context *ctx = (void *) _ctx;

	ctx->outurb->transfer_flags |= URB_ASYNC_UNLINK;
	if (usb_unlink_urb(ctx->outurb) == -EINPROGRESS) {
		ctx->state = EZUSB_CTX_REQ_TIMEOUT;
	} else {
		ctx->state = EZUSB_CTX_RESP_TIMEOUT;
		dbg("couldn't unlink");
		atomic_inc(&ctx->refcount);
		ctx->killed = 1;
		ezusb_ctx_complete(ctx);
		ezusb_request_context_put(ctx);
	}
};

struct request_context *ezusb_alloc_ctx(struct ezusb_priv *dev,
					u16 out_rid, u16 in_rid)
{
	struct request_context *ctx;

	ctx = kmalloc(sizeof(*ctx), GFP_ATOMIC);
	if (!ctx) {
		return NULL;
	}
	memset(ctx, 0, sizeof(*ctx));

	ctx->buf = kmalloc(BULK_BUF_SIZE, GFP_ATOMIC);
	if (!ctx->buf) {
		kfree(ctx);
		return NULL;
	}
	ctx->outurb = usb_alloc_urb(0, GFP_ATOMIC);
	if (!ctx->outurb) {
		kfree(ctx->buf);
		kfree(ctx);
		return NULL;
	}

	ctx->dev = dev;
	ctx->state = EZUSB_CTX_START;
	ctx->out_rid = out_rid;
	ctx->in_rid = in_rid;

	atomic_set(&ctx->refcount, 1);
	init_completion(&ctx->done);

	init_timer(&ctx->timer);
	ctx->timer.function = ezusb_request_timerfn;
	ctx->timer.data = (u_long) ctx;
	return ctx;
}


/* Hopefully the real complete_all will soon be exported, in the mean
 * while this should work. */
static inline void ezusb_complete_all(struct completion *comp)
{
	complete(comp);
	complete(comp);
	complete(comp);
	complete(comp);
}

static void ezusb_ctx_complete(struct request_context *ctx)
{
	struct ezusb_priv *dev = ctx->dev;
	unsigned long flags;

	spin_lock_irqsave(&dev->req_lock, flags);

	list_del_init(&ctx->list);
	if (dev->udev) {
		spin_unlock_irqrestore(&dev->req_lock, flags);
		ezusb_req_queue_run(dev);
		spin_lock_irqsave(&dev->req_lock, flags);
	}

	switch (ctx->state) {
	case EZUSB_CTX_COMPLETE:
	case EZUSB_CTX_REQSUBMIT_FAIL:
	case EZUSB_CTX_REQ_FAILED:
	case EZUSB_CTX_REQ_TIMEOUT:
	case EZUSB_CTX_RESP_TIMEOUT:
		spin_unlock_irqrestore(&dev->req_lock, flags);

		if ((ctx->out_rid == EZUSB_RID_TX) && dev->net_dev) {
			struct net_device *net_dev = dev->net_dev;
			struct orinoco_private *priv = netdev_priv(net_dev);
			struct net_device_stats *stats = &priv->stats;

			if (ctx->state != EZUSB_CTX_COMPLETE)
				stats->tx_errors++;
			else
				stats->tx_packets++;

			netif_wake_queue(net_dev);
		}
		ezusb_complete_all(&ctx->done);
		ezusb_request_context_put(ctx);
		break;

	default:
		spin_unlock_irqrestore(&dev->req_lock, flags);
		if (!dev->udev) {
			/* This is normal, as all request contexts get flushed
			 * when the device is disconnected */
			err("Called, CTLX not terminating, but device gone");
			ezusb_complete_all(&ctx->done);
			ezusb_request_context_put(ctx);
			break;
		}

		err("Called, CTLX not in terminating state.");
		/* Things are really bad if this happens. Just leak
		 * the CTLX because it may still be linked to the 
		 * queue or the OUT urb may still be active.
		 * Just leaking at least prevents an Oops or Panic.
		 */
		break;
	}
}

/**
 * ezusb_req_queue_run:
 * Description:
 *	Note: Only one active CTX at any one time, because there's no
 *	other (reliable) way to match the response URB to the correct
 *	CTX.
 **/
static void ezusb_req_queue_run(struct ezusb_priv *dev)
{
	unsigned long flags;
	struct request_context *ctx;
	int result;

	spin_lock_irqsave(&dev->req_lock, flags);

	if (!list_empty(&dev->req_active))
		goto unlock;

	if (list_empty(&dev->req_pending))
		goto unlock;

	ctx =
	    list_entry(dev->req_pending.next, struct request_context,
		       list);

	if (!ctx->dev->udev)
		goto unlock;

	/* We need to split this off to avoid a race condition */
	list_move_tail(&ctx->list, &dev->req_active);

	if (ctx->state == EZUSB_CTX_QUEUED) {
		atomic_inc(&ctx->refcount);
		result = usb_submit_urb(ctx->outurb, GFP_ATOMIC);
		if (result) {
			ctx->state = EZUSB_CTX_REQSUBMIT_FAIL;

			spin_unlock_irqrestore(&dev->req_lock, flags);

			err("Fatal, failed to submit command urb."
			    " error=%d\n", result);

			ezusb_ctx_complete(ctx);
			ezusb_request_context_put(ctx);
			goto done;
		}

		ctx->state = EZUSB_CTX_REQ_SUBMITTED;
		ezusb_mod_timer(ctx->dev, &ctx->timer,
				jiffies + DEF_TIMEOUT);
	}

      unlock:
	spin_unlock_irqrestore(&dev->req_lock, flags);

      done:
	return;
}

static void ezusb_req_enqueue_run(struct ezusb_priv *dev,
				  struct request_context *ctx)
{
	unsigned long flags;

	spin_lock_irqsave(&dev->req_lock, flags);

	if (!ctx->dev->udev) {
		spin_unlock_irqrestore(&dev->req_lock, flags);
		goto done;
	}
	atomic_inc(&ctx->refcount);
	list_add_tail(&ctx->list, &dev->req_pending);
	spin_unlock_irqrestore(&dev->req_lock, flags);

	ctx->state = EZUSB_CTX_QUEUED;
	ezusb_req_queue_run(dev);

      done:
	return;
}

static void ezusb_request_out_callback(struct urb *urb,
				       struct pt_regs *pt_regs)
{
	unsigned long flags;
	ezusb_state_t state;
	struct request_context *ctx = urb->context;
	struct ezusb_priv *dev = ctx->dev;

	spin_lock_irqsave(&dev->req_lock, flags);

	del_timer(&ctx->timer);

	if (ctx->killed) {
		spin_unlock_irqrestore(&dev->req_lock, flags);
		warn("interrupt called with dead ctx");
		goto out;
	}

	state = ctx->state;

	if (urb->status == 0) {
		switch (state) {
		case EZUSB_CTX_REQ_SUBMITTED:
			if (ctx->in_rid) {
				ctx->state = EZUSB_CTX_REQ_COMPLETE;
				/* reply URB still pending */
				ezusb_mod_timer(dev, &ctx->timer,
						jiffies + DEF_TIMEOUT);
				spin_unlock_irqrestore(&dev->req_lock,
						       flags);
				break;
			}
			/* fall through */
		case EZUSB_CTX_RESP_RECEIVED:
			/* IN already received before this OUT-ACK */
			ctx->state = EZUSB_CTX_COMPLETE;
			spin_unlock_irqrestore(&dev->req_lock, flags);
			ezusb_ctx_complete(ctx);
			break;

		default:
			spin_unlock_irqrestore(&dev->req_lock, flags);
			err("Unexpected state(0x%x, %d) in OUT URB",
			    state, urb->status);
			break;
		}
	} else {
		/* If someone cancels the OUT URB then its status
		 * should be either -ECONNRESET or -ENOENT.
		 */
		switch (state) {
		case EZUSB_CTX_REQ_SUBMITTED:
		case EZUSB_CTX_RESP_RECEIVED:
			ctx->state = EZUSB_CTX_REQ_FAILED;
			/* fall through */

		case EZUSB_CTX_REQ_FAILED:
		case EZUSB_CTX_REQ_TIMEOUT:
			spin_unlock_irqrestore(&dev->req_lock, flags);

			ezusb_ctx_complete(ctx);
			break;

		default:
			spin_unlock_irqrestore(&dev->req_lock, flags);

			err("Unexpected state(0x%x, %d) in OUT URB",
			    state, urb->status);
			break;
		}
	}
      out:
	ezusb_request_context_put(ctx);
}

static void ezusb_request_in_callback(struct ezusb_priv *dev,
				      struct urb *urb)
{
	struct ezusb_packet *ans = urb->transfer_buffer;
	struct request_context *ctx = NULL;
	ezusb_state_t state;
	unsigned long flags;

	/* Find the CTLX on the active queue that requested this URB */
	spin_lock_irqsave(&dev->req_lock, flags);
	if (dev->udev) {
		struct list_head *item;

		list_for_each(item, &dev->req_active) {
			struct request_context *c;
			int reply_count;

			c = list_entry(item, struct request_context, list);
			reply_count =
			    ezusb_reply_inc(c->buf->req_reply_count);
			if ((ans->ans_reply_count == reply_count)
			    && (le16_to_cpu(ans->hermes_rid) == c->in_rid)) {
				ctx = c;
				break;
			}
			dbg("Skipped (0x%x/0x%x) (%d/%d)",
			    le16_to_cpu(ans->hermes_rid),
			    c->in_rid, ans->ans_reply_count, reply_count);
		}
	}

	if (ctx == NULL) {
		spin_unlock_irqrestore(&dev->req_lock, flags);
		err("%s: got unexpected RID: 0x%04X", __FUNCTION__,
		    le16_to_cpu(ans->hermes_rid));
		ezusb_req_queue_run(dev);
		return;
	}

	/* The data we want is in the in buffer, exchange */
	urb->transfer_buffer = ctx->buf;
	ctx->buf = (void *) ans;
	ctx->buf_length = urb->actual_length;

	state = ctx->state;
	switch (state) {
	case EZUSB_CTX_REQ_SUBMITTED:
		/* We have received our response URB before
		 * our request has been acknowledged. Do NOT
		 * destroy our CTX yet, because our OUT URB
		 * is still alive ...
		 */
		ctx->state = EZUSB_CTX_RESP_RECEIVED;
		spin_unlock_irqrestore(&dev->req_lock, flags);

		/* Let the machine continue running. */
		break;

	case EZUSB_CTX_REQ_COMPLETE:
		/* This is the usual path: our request
		 * has already been acknowledged, and
		 * we have now received the reply.
		 */
		ctx->state = EZUSB_CTX_COMPLETE;

		/* Stop the intimer */
		del_timer(&ctx->timer);
		spin_unlock_irqrestore(&dev->req_lock, flags);

		/* Call the completion handler */
		ezusb_ctx_complete(ctx);
		break;

	default:
		spin_unlock_irqrestore(&dev->req_lock, flags);

		warn("Matched IN URB, unexpected context state(0x%x)",
		     state);
		/* Throw this CTX away and try submitting another */
		del_timer(&ctx->timer);
		ctx->outurb->transfer_flags |= URB_ASYNC_UNLINK;
		usb_unlink_urb(ctx->outurb);
		ezusb_req_queue_run(dev);
		break;
	}			/* switch */
}


static void ezusb_req_ctx_wait(struct ezusb_priv *dev,
			       struct request_context *ctx)
{
	switch (ctx->state) {
	case EZUSB_CTX_QUEUED:
	case EZUSB_CTX_REQ_SUBMITTED:
	case EZUSB_CTX_REQ_COMPLETE:
	case EZUSB_CTX_RESP_RECEIVED:
		if (in_atomic()) {
			/* If we get called from a timer, timeout timers don't
			 * get the chance to run themselfs. So we make sure
			 * that we don't sleep for ever */
			int msecs = DEF_TIMEOUT * (1000 / HZ);
			while (!ctx->done.done && msecs--)
				udelay(1000);
		} else {
			wait_event_interruptible(ctx->done.wait,
						 ctx->done.done);
		}
		break;
	default:
		/* Done or failed - nothing to wait for */
		break;
	}
}

static inline u16 build_crc(u8 *data)
{
	u16 crc = 0;
	int i;
	for (i = 0; i < 8; i++) {
		crc = (crc << 1) + data[i];
	}
	return crc;
}

static inline void ezusb_req_le_to_cpu(struct ezusb_packet *req)
{
	le16_to_cpus(&req->magic);
	le16_to_cpus(&req->frame_type);
	le16_to_cpus(&req->size);
	le16_to_cpus(&req->crc);
	/* don't translate req->hermes_len and req->hermes_rid,
	   orinoco.c expects them to be little endian */
}

/**
 * ezusb_fill_req:
 * 
 * if data == NULL and length > 0 the data is assumed to be already in
 * the target buffer and only the header is filled.
 *
 */
static int ezusb_fill_req(struct ezusb_packet *req, u16 length, u16 rid,
			  const void *data, u16 frame_type, u8 reply_count)
{
	int total_size = sizeof(*req) + length;

	BUG_ON(total_size > BULK_BUF_SIZE);

	req->magic = cpu_to_le16(EZUSB_MAGIC);
	req->req_reply_count = reply_count;
	req->ans_reply_count = 0;
	req->frame_type = cpu_to_le16(frame_type);
	req->size = cpu_to_le16(length + 4);
	req->crc = cpu_to_le16(build_crc((u8 *) req));
	req->hermes_len = cpu_to_le16(HERMES_BYTES_TO_RECLEN(length));
	req->hermes_rid = cpu_to_le16(rid);
	if (data)
		memcpy(req->data, data, length);
	return total_size;
}

static int ezusb_submit_in_urb(struct ezusb_priv *dev)
{
	int retval = 0;
	void *cur_buf = dev->read_urb->transfer_buffer;

	if (dev->read_urb->status == -EINPROGRESS) {
		dbg("urb busy, not resubmiting");
		retval = -EBUSY;
		goto exit;
	}
	usb_fill_bulk_urb(dev->read_urb, dev->udev, dev->read_pipe,
			  cur_buf, BULK_BUF_SIZE,
			  ezusb_bulk_in_callback, dev);
	dev->read_urb->transfer_flags = 0;
	retval = usb_submit_urb(dev->read_urb, GFP_ATOMIC);
	if (retval)
		err("%s submit failed %d", __FUNCTION__, retval);

      exit:
	return retval;
}

static int ezusb_remove_in_urb(struct ezusb_priv *dev)
{
	int retval = 0;

	if (dev->read_urb->status != -EINPROGRESS) {
		dbg("no urb to remove");
		retval = -ENOENT;
		goto exit;
	}
	retval = usb_unlink_urb(dev->read_urb);
      exit:
	return retval;
}

static inline int ezusb_8051_cpucs(struct ezusb_priv *dev, int reset)
{
	u8 res_val = reset;	/* avoid argument promotion */

	if (!dev->udev) {
		err("%s: !dev->udev", __FUNCTION__);
		return -EFAULT;
	}
	return usb_control_msg(dev->udev,
			       usb_sndctrlpipe(dev->udev, 0),
			       EZUSB_REQUEST_FW_TRANS,
			       USB_TYPE_VENDOR | USB_RECIP_DEVICE |
			       USB_DIR_OUT, EZUSB_CPUCS_REG, 0, &res_val,
			       sizeof(res_val), DEF_TIMEOUT);
}

static int find_fw_variant_offset(struct ez_usb_fw *fw)
{
	int offset;
	offset = *(short *) (fw->code + 1);
	offset = be16_to_cpu(offset);
	for (; offset < fw->size; offset++) {
		if (memcmp("\x02\x03\x58", &fw->code[offset], 3) == 0)
			break;
	}
	if (offset > (fw->size - 4)) {
		dbg("couldn't find card_variant_offset");
		return -1;
	}
	fw->card_variant_offset = offset + 3;
	dbg("card_variant_offset = 0x%04X", fw->card_variant_offset);
	return 0;
}

static int ezusb_firmware_download(struct ezusb_priv *dev,
				   struct ez_usb_fw *fw)
{
	int retval, addr;
	int packet_size = dev->udev->descriptor.bMaxPacketSize0;
	int variant_offset = fw->card_variant_offset % packet_size;
	int variant_base = fw->card_variant_offset - variant_offset;
	u8 *transfer_buffer = kmalloc(packet_size, GFP_KERNEL);

	if (!transfer_buffer) {
		printk(KERN_ERR PFX "kmalloc(%d) failed\n", packet_size);
		retval = -ENOMEM;
		goto fail;
	}

	retval = ezusb_8051_cpucs(dev, 1);
	if (retval < 0)
		goto fail;
	for (addr = 0; addr < fw->size; addr += packet_size) {
		/* 0x100-0x300 should be left alone, it contains card
		 * specific data, like USB enumeration information */
		if ((addr >= 0x100) && (addr < 0x300))
			continue;

		memcpy(transfer_buffer, &fw->code[addr], packet_size);
		if (addr == variant_base) {
			dbg("Patching card_variant byte 0x%02X at 0x%04X",
			    dev->card_variant, fw->card_variant_offset);
			transfer_buffer[variant_offset] =
			    dev->card_variant;
		}
		retval = usb_control_msg(dev->udev,
					 usb_sndctrlpipe(dev->udev, 0),
					 EZUSB_REQUEST_FW_TRANS,
					 USB_TYPE_VENDOR | USB_RECIP_DEVICE
					 | USB_DIR_OUT,
					 addr, 0x0,
					 transfer_buffer, packet_size,
					 DEF_TIMEOUT);

		if (retval < 0)
			goto fail;
	}
	retval = ezusb_8051_cpucs(dev, 0);
	if (retval < 0)
		goto fail;

	goto exit;
      fail:
	printk(KERN_ERR PFX "Firmware download failed, error %d\n",
	       retval);
      exit:
	if (transfer_buffer)
		kfree(transfer_buffer);
	return retval;
}

static int ezusb_access_ltv(struct ezusb_priv *dev,
			    struct request_context *ctx,
			    u16 length, const void *data, u16 frame_type,
			    void *ans_buff, int ans_size, u16 *ans_length)
{
	int req_size;
	int retval = 0;
	ezusb_state_t state;

	BUG_ON(in_irq());

	if (!dev->udev) {
		dbg("Device disconnected");
		return -ENODEV;
	}

	if (dev->read_urb->status != -EINPROGRESS) {
		err("%s: in urb not pending", __FUNCTION__);
	}

	/* protect dev->reply_count, guarantee sequential numbers */
	spin_lock_bh(&dev->reply_count_lock);
	req_size = ezusb_fill_req(ctx->buf, length, ctx->out_rid, data,
				  frame_type, dev->reply_count);
	usb_fill_bulk_urb(ctx->outurb, dev->udev, dev->write_pipe,
			  ctx->buf, req_size,
			  ezusb_request_out_callback, ctx);

	if (ctx->in_rid)
		dev->reply_count = ezusb_reply_inc(dev->reply_count);

	ezusb_req_enqueue_run(dev, ctx);

	spin_unlock_bh(&dev->reply_count_lock);

	if (ctx->in_rid)
		ezusb_req_ctx_wait(dev, ctx);

	state = ctx->state;
	switch (state) {
	case EZUSB_CTX_COMPLETE:
		retval = ctx->outurb->status;
		break;

	case EZUSB_CTX_QUEUED:
	case EZUSB_CTX_REQ_SUBMITTED:
		if (!ctx->in_rid)
			break;
	default:
		err("%s: Unexpected context state %d", __FUNCTION__,
		    state);
		/* fall though */
	case EZUSB_CTX_REQ_TIMEOUT:
	case EZUSB_CTX_REQ_FAILED:
	case EZUSB_CTX_RESP_TIMEOUT:
	case EZUSB_CTX_REQSUBMIT_FAIL:
		printk(KERN_ERR PFX "Access failed, resetting (state %d,"
		       " reply_count %d)\n", state, dev->reply_count);
		dev->reply_count = 0;
		if (state == EZUSB_CTX_REQ_TIMEOUT
		    || state == EZUSB_CTX_RESP_TIMEOUT) {
			printk(KERN_ERR PFX "ctx timed out\n");
			retval = -ETIMEDOUT;
		} else {
			printk(KERN_ERR PFX "ctx failed\n");
			retval = -EFAULT;
		}
		goto exit;
		break;
	}
	if (ctx->in_rid) {
		struct ezusb_packet *ans = ctx->buf;
		int exp_len;

		if (ans->hermes_len != 0)
			exp_len = le16_to_cpu(ans->hermes_len) * 2 + 12;
		else
			exp_len = 14;

		if (exp_len != ctx->buf_length) {
			err("%s: lenght mismatch for RID 0x%04x: "
			    "expected %d, got %d", __FUNCTION__,
			    ctx->in_rid, exp_len, ctx->buf_length);
			retval = -EIO;
			goto exit;
		}

		if (ans_buff)
			memcpy(ans_buff, ans->data,
			       min_t(int, exp_len, ans_size));
		if (ans_length)
			*ans_length = le16_to_cpu(ans->hermes_len);
	}
      exit:
	ezusb_request_context_put(ctx);
	return retval;
}

static int ezusb_write_ltv(hermes_t *hw, int bap, u16 rid,
			   u16 length, const void *data)
{
	struct ezusb_priv *dev = hw->priv;
	u16 frame_type;
	struct request_context *ctx;

	if (length == 0)
		return -EINVAL;

	length = HERMES_RECLEN_TO_BYTES(length);

	/* On memory mapped devices HERMES_RID_CNFGROUPADDRESSES can be
	 * set to be empty, but the USB bridge doesn't like it */
	if (length == 0)
		return 0;

	ctx = ezusb_alloc_ctx(dev, rid, EZUSB_RID_ACK);
	if (!ctx)
		return -ENOMEM;

	if (rid == EZUSB_RID_TX)
		frame_type = EZUSB_FRAME_DATA;
	else
		frame_type = EZUSB_FRAME_CONTROL;

	return ezusb_access_ltv(dev, ctx, length, data, frame_type,
				NULL, 0, NULL);
}

static int ezusb_read_ltv(hermes_t *hw, int bap, u16 rid,
			  unsigned bufsize, u16 *length, void *buf)
{
	struct ezusb_priv *dev = hw->priv;
	struct request_context *ctx;

	if ((bufsize < 0) || (bufsize % 2))
		return -EINVAL;

	ctx = ezusb_alloc_ctx(dev, rid, rid);
	if (!ctx)
		return -ENOMEM;

	return ezusb_access_ltv(dev, ctx, 0, NULL, EZUSB_FRAME_CONTROL,
				buf, bufsize, length);
}

static int ezusb_docmd_wait(hermes_t *hw, u16 cmd, u16 parm0,
			    struct hermes_response *resp)
{
	struct ezusb_priv *dev = hw->priv;
	struct request_context *ctx;

	u16 data[4] = {
		cpu_to_le16(cmd),
		cpu_to_le16(parm0),
		0,
		0,
	};
	dbg("0x%04X, parm0 0x%04X", cmd, parm0);
	ctx = ezusb_alloc_ctx(dev, EZUSB_RID_DOCMD, EZUSB_RID_ACK);
	if (!ctx)
		return -ENOMEM;

	return ezusb_access_ltv(dev, ctx, sizeof(data), &data,
				EZUSB_FRAME_CONTROL, NULL, 0, NULL);
}

static int ezusb_bap_pread(struct hermes *hw, int bap,
			   void *buf, unsigned len, u16 id, u16 offset)
{
	struct ezusb_priv *dev = hw->priv;
	struct ezusb_packet *ans = (void *) dev->read_urb->transfer_buffer;
	int actual_length = dev->read_urb->actual_length;

	if (id == EZUSB_RID_RX) {
		if ((sizeof(*ans) + offset + len) > actual_length) {
			printk(KERN_ERR PFX "BAP read beyond buffer end "
			       "in rx frame\n");
			return -EINVAL;
		}
		memcpy(buf, ans->data + offset, len);
		return 0;
	}

	if (EZUSB_IS_INFO(id)) {
		/* Include 4 bytes for length/type */
		if ((sizeof(*ans) + offset + len - 4) > actual_length) {
			printk(KERN_ERR PFX "BAP read beyond buffer end "
			       "in info frame\n");
			return -EFAULT;
		}
		memcpy(buf, ans->data + offset - 4, len);
	} else {
		printk(KERN_ERR PFX "Unexpected fid 0x%04x\n", id);
		return -EINVAL;
	}

	return 0;
}

static int ezusb_xmit(struct sk_buff *skb, struct net_device *net_dev)
{
	struct orinoco_private *priv =
	    (struct orinoco_private *) netdev_priv(net_dev);
	struct net_device_stats *stats = &priv->stats;
	struct ezusb_priv *dev = priv->card;
	int err = 0;
	char *p;
	struct ethhdr *eh;
	int len, data_len, data_off;
	u16 tx_control;
	unsigned long flags;
	struct request_context *ctx;
	u8 *buf;
	int tx_size;

	TRACE_ENTER(net_dev->name);

	if (!netif_running(net_dev)) {
		printk(KERN_ERR "%s: Tx on stopped device!\n",
		       net_dev->name);
		TRACE_EXIT(net_dev->name);
		return 1;
	}

	if (netif_queue_stopped(net_dev)) {
		printk(KERN_DEBUG "%s: Tx while transmitter busy!\n",
		       net_dev->name);
		TRACE_EXIT(net_dev->name);
		return 1;
	}

	if (orinoco_lock(priv, &flags) != 0) {
		printk(KERN_ERR
		       "%s: orinoco_xmit() called while hw_unavailable\n",
		       net_dev->name);
		TRACE_EXIT(net_dev->name);
		return 1;
	}

	if (!netif_carrier_ok(net_dev) ||
	    (priv->iw_mode == IW_MODE_MONITOR)) {
		/* Oops, the firmware hasn't established a connection,
		   silently drop the packet (this seems to be the
		   safest approach). */
		stats->tx_errors++;
		orinoco_unlock(priv, &flags);
		dev_kfree_skb(skb);
		TRACE_EXIT(net_dev->name);
		return 0;
	}

	ctx = ezusb_alloc_ctx(dev, EZUSB_RID_TX, 0);
	if (!ctx) {
		err = -ENOMEM;
		goto fail;
	}
	memset(ctx->buf, 0, BULK_BUF_SIZE);
	buf = ctx->buf->data;

	/* Length of the packet body */
	/* FIXME: what if the skb is smaller than this? */
	len = max_t(int, skb->len - ETH_HLEN, ETH_ZLEN - ETH_HLEN);

	eh = (struct ethhdr *) skb->data;

	tx_control = cpu_to_le16(0);
	memcpy(buf, &tx_control, sizeof(tx_control));
	buf += sizeof(tx_control);
	/* Encapsulate Ethernet-II frames */
	if (ntohs(eh->h_proto) > ETH_DATA_LEN) {	/* Ethernet-II frame */
		struct header_struct *hdr = (void *) buf;
		buf += sizeof(*hdr);
		data_len = len;
		data_off = sizeof(tx_control) + sizeof(*hdr);
		p = skb->data + ETH_HLEN;

		/* 802.3 header */
		memcpy(hdr->dest, eh->h_dest, ETH_ALEN);
		memcpy(hdr->src, eh->h_source, ETH_ALEN);
		hdr->len = htons(data_len + ENCAPS_OVERHEAD);

		/* 802.2 header */
		memcpy(&hdr->dsap, &encaps_hdr, sizeof(encaps_hdr));

		hdr->ethertype = eh->h_proto;
	} else {		/* IEEE 802.3 frame */
		data_len = len + ETH_HLEN;
		data_off = sizeof(tx_control);
		p = skb->data;
	}

	memcpy(buf, p, data_len);
	buf += data_len;

	/* Finally, we actually initiate the send */
	netif_stop_queue(net_dev);

	/* The card may behave better if we send evenly sized usb transfers */
	tx_size = ALIGN(buf - ctx->buf->data, 2);

	err = ezusb_access_ltv(dev, ctx, tx_size, NULL,
			       EZUSB_FRAME_DATA, NULL, 0, NULL);

	if (err) {
		netif_start_queue(net_dev);
		printk(KERN_ERR "%s: Error %d transmitting packet\n",
		       net_dev->name, err);
		stats->tx_errors++;
		goto fail;
	}

	net_dev->trans_start = jiffies;
	stats->tx_bytes += data_off + data_len;

	orinoco_unlock(priv, &flags);

	dev_kfree_skb(skb);

	TRACE_EXIT(net_dev->name);

	return 0;
      fail:
	TRACE_EXIT(net_dev->name);

	orinoco_unlock(priv, &flags);
	return err;
}

static int ezusb_allocate(struct hermes *hw, u16 size, u16 *fid)
{
	*fid = EZUSB_RID_TX;
	return 0;
}


static int ezusb_hard_reset(struct orinoco_private *priv)
{
	struct ezusb_priv *dev = priv->card;
	int retval = ezusb_8051_cpucs(dev, 1);

	if (retval < 0) {
		err("Failed to reset");
		return retval;
	}

	retval = ezusb_8051_cpucs(dev, 0);
	if (retval < 0) {
		err("Failed to unreset");
		return retval;
	}

	dbg("sending control message");
	retval = usb_control_msg(dev->udev,
				 usb_sndctrlpipe(dev->udev, 0),
				 EZUSB_REQUEST_TRIGER,
				 USB_TYPE_VENDOR | USB_RECIP_DEVICE |
				 USB_DIR_OUT, 0x0, 0x0, NULL, 0,
				 DEF_TIMEOUT);
	if (retval < 0) {
		err("EZUSB_REQUEST_TRIGER failed retval %d", retval);
		return retval;
	}
#if 0
	dbg("Sending EZUSB_REQUEST_TRIG_AC");
	retval = usb_control_msg(dev->udev,
				 usb_sndctrlpipe(dev->udev, 0),
				 EZUSB_REQUEST_TRIG_AC,
				 USB_TYPE_VENDOR | USB_RECIP_DEVICE |
				 USB_DIR_OUT, 0x00FA, 0x0, NULL, 0,
				 DEF_TIMEOUT);
	if (retval < 0) {
		err("EZUSB_REQUEST_TRIG_AC failed retval %d", retval);
		return retval;
	}
#endif

	return 0;
}


static int ezusb_init(hermes_t *hw)
{
	struct ezusb_priv *dev = hw->priv;
	int retval;

	BUG_ON(in_interrupt());
	BUG_ON(!dev);

	dev->reply_count = 0;
	/* Write the MAGIC number on the simulated registers to keep
	 * orinoco.c happy */
	hermes_write_regn(hw, SWSUPPORT0, HERMES_MAGIC);
	hermes_write_regn(hw, RXFID, EZUSB_RID_RX);

	ezusb_remove_in_urb(dev);
	ezusb_submit_in_urb(dev);

	retval = ezusb_write_ltv(hw, 0, EZUSB_RID_INIT1,
				 HERMES_BYTES_TO_RECLEN(2), "\x10\x00");
	if (retval < 0) {
		printk(KERN_ERR PFX "EZUSB_RID_INIT1 error %d\n", retval);
		return retval;
	}

	retval = ezusb_docmd_wait(hw, HERMES_CMD_INIT, 0, NULL);
	if (retval < 0) {
		printk(KERN_ERR PFX "HERMES_CMD_INIT error %d\n", retval);
		return retval;
	}

	return 0;
}

static void ezusb_bulk_in_callback(struct urb *urb, struct pt_regs *regs)
{
	struct ezusb_priv *dev = (struct ezusb_priv *) urb->context;
	struct ezusb_packet *ans = urb->transfer_buffer;
	u16 crc;
	u16 hermes_rid;

	if (dev->udev == NULL) {
		dbg("disconnected");
		return;
	}

	if (urb->status == -ETIMEDOUT) {
		/* When a device gets unplugged we get this every time
		 * we resubmit, flooding the logs.  Since we don't use
		 * USB timeouts, it shouldn't happen any other time*/
		warn("%s: urb timed out, not resubmiting", __FUNCTION__);
		return;
	}
	if (urb->status == -ECONNABORTED) {
		warn("%s: connection abort, resubmiting urb",
		     __FUNCTION__);
		goto resubmit;
	}
	if ((urb->status == -EILSEQ)
	    || (urb->status == -ENOENT)
	    || (urb->status == -ECONNRESET)) {
		dbg("status %d, not resubmiting", urb->status);
		return;
	}
	if (urb->status)
		dbg("status: %d length: %d",
		    urb->status, urb->actual_length);
	if (urb->actual_length < sizeof(*ans)) {
		err("%s: short read, ignoring", __FUNCTION__);
		goto resubmit;
	}
	crc = build_crc((u8 *) ans);
	ezusb_req_le_to_cpu(ans);
	if (ans->crc != crc) {
		err("CRC error, ignoring packet");
		goto resubmit;
	}

	hermes_rid = le16_to_cpu(ans->hermes_rid);
	if ((hermes_rid != EZUSB_RID_RX) && !EZUSB_IS_INFO(hermes_rid)) {
		ezusb_request_in_callback(dev, urb);
	} else if (dev->net_dev) {
		struct net_device *net_dev = dev->net_dev;
		struct orinoco_private *priv = netdev_priv(net_dev);
		hermes_t *hw = &priv->hw;

		if (hermes_rid == EZUSB_RID_RX) {
			__orinoco_ev_rx(net_dev, hw);
		} else {
			hermes_write_regn(hw, INFOFID,
					  le16_to_cpu(ans->hermes_rid));
			__orinoco_ev_info(net_dev, hw);
		}
	}

      resubmit:
	if (dev->udev)
		ezusb_submit_in_urb(dev);
}

static inline void ezusb_delete(struct ezusb_priv *dev)
{
	struct net_device *net_dev;
	struct list_head *item;
	struct list_head *tmp_item;
	unsigned long flags;
	int retval;

	BUG_ON(in_interrupt());
	BUG_ON(!dev);

	net_dev = dev->net_dev;
	down(&dev->sem);

	dev->udev = NULL;	/* No timer will be rearmed from here */

	retval = ezusb_remove_in_urb(dev);
	if (retval) {
		dbg("retval %d status %d", retval, dev->read_urb->status);
	}

      restart_list:
	spin_lock_irqsave(&dev->req_lock, flags);
	list_for_each_safe(item, tmp_item, &dev->req_active) {
		struct request_context *ctx;

		ctx = list_entry(item, struct request_context, list);
		atomic_inc(&ctx->refcount);

		ctx->outurb->transfer_flags |= URB_ASYNC_UNLINK;
		if (usb_unlink_urb(ctx->outurb) == -EINPROGRESS) {
			spin_unlock_irqrestore(&dev->req_lock, flags);
			wait_for_completion(&ctx->done);
		} else {
			spin_unlock_irqrestore(&dev->req_lock, flags);
		}
		del_timer_sync(&ctx->timer);
		/* FIXME: there is an slight chance for the irq handler to
		 * be running */
		if (!list_empty(&ctx->list))
			ezusb_ctx_complete(ctx);

		ezusb_request_context_put(ctx);
		goto restart_list;
	}
	spin_unlock_irqrestore(&dev->req_lock, flags);

	list_for_each_safe(item, tmp_item, &dev->req_pending)
	    ezusb_ctx_complete(list_entry(item,
					  struct request_context, list));

	if (dev->read_urb->status == -EINPROGRESS) {
		printk(KERN_ERR PFX "Some URB in progress\n");
	}
	up(&dev->sem);

	kfree(dev->read_urb->transfer_buffer);
	if (dev->bap_buf != NULL)
		kfree(dev->bap_buf);
	if (dev->read_urb != NULL)
		usb_free_urb(dev->read_urb);
	if (dev->net_dev) {
		unregister_netdev(dev->net_dev);
		free_orinocodev(dev->net_dev);
	}
}

static const struct hermes_ops ezusb_ops = {
	.init = ezusb_init,
	.docmd_wait = ezusb_docmd_wait,
	.allocate = ezusb_allocate,
	.read_ltv = ezusb_read_ltv,
	.write_ltv = ezusb_write_ltv,
	.bap_pread = ezusb_bap_pread
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static void *ezusb_probe(struct usb_device *udev, unsigned int ifnum,
			 const struct usb_device_id *id)
{
	struct usb_interface *interface =
	    &udev->actconfig->interface[ifnum];
#else
static int ezusb_probe(struct usb_interface *interface,
		       const struct usb_device_id *id)
{
	struct usb_device *udev = interface_to_usbdev(interface);
#endif
	struct orinoco_private *priv;
	struct net_device *net_dev;
	hermes_t *hw;
	struct ezusb_priv *dev = NULL;
	struct usb_interface_descriptor *iface_desc;
	struct usb_endpoint_descriptor *ep;
	int card_variant = -1;
	const struct firmware *fw_entry;
	int retval = 0;
	int i;

	/* See if the device offered us matches what we can accept */

	if (usb_match_id(interface, ezusb_varian1_table))
		card_variant = 0;
	if (usb_match_id(interface, ezusb_varian2_table))
		card_variant = 1;

	if (force_unsupported)
		card_variant = -1;

	if (card_variant == -1) {
		err("Device is not supported (you may want to set"
		    " force_unsupported=1)");
		if (!force_unsupported)
			return_probe(NULL, -ENODEV);
		err("Trying to handle device anyway as requested");
	}

	net_dev = alloc_orinocodev(sizeof(*dev), ezusb_hard_reset);
	if (!net_dev) {
		err("Couldn't allocate net_dev");
		goto exit;
	}

	priv = netdev_priv(net_dev);
	hw = &priv->hw;

	dev = priv->card;

	init_MUTEX(&dev->sem);
	spin_lock_init(&dev->reply_count_lock);

	spin_lock_init(&dev->req_lock);
	INIT_LIST_HEAD(&dev->req_pending);
	INIT_LIST_HEAD(&dev->req_active);

	dev->udev = udev;

	hw->iobase = (unsigned long) &dev->hermes_reg_fake;
	hw->io_space = HERMES_MEM;
	hw->reg_spacing = HERMES_16BIT_REGSPACING;
	hw->priv = dev;
	hw->ops = &ezusb_ops;
	net_dev->hard_start_xmit = ezusb_xmit;

	priv->irq_no_disable = 1;

	/* set up the endpoint information */
	/* check out the endpoints */

	iface_desc = &usb_get_desc(interface->altsetting[0]);
	for (i = 0; i < iface_desc->bNumEndpoints; ++i) {
		ep = &usb_get_desc(interface->altsetting[0].endpoint[i]);

		if (((ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
		     == USB_DIR_IN) &&
		    ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
		     == USB_ENDPOINT_XFER_BULK)) {
			/* we found a bulk in endpoint */
			if (dev->read_urb != NULL) {
				warn("Found a second bulk in ep, ignored");
				continue;
			}

			dev->read_urb = usb_alloc_urb(0, GFP_KERNEL);
			if (!dev->read_urb) {
				err("No free urbs available");
				goto error;
			}
			if (ep->wMaxPacketSize != 64)
				warn("bulk in: wMaxPacketSize!= 64");
			if (ep->bEndpointAddress != (2 | USB_DIR_IN))
				warn("bulk in: bEndpointAddress: %d",
				     ep->bEndpointAddress);
			dev->read_pipe = usb_rcvbulkpipe(udev,
							 ep->
							 bEndpointAddress);
			dev->read_urb->transfer_buffer =
			    kmalloc(BULK_BUF_SIZE, GFP_KERNEL);
			if (!dev->read_urb->transfer_buffer) {
				err("Couldn't allocate IN buffer");
				goto error;
			}
		}

		if (((ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
		     == USB_DIR_OUT) &&
		    ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
		     == USB_ENDPOINT_XFER_BULK)) {
			/* we found a bulk out endpoint */
			if (dev->bap_buf != NULL) {
				warn("Found a second bulk out ep, ignored");
				continue;
			}

			if (ep->wMaxPacketSize != 64)
				warn("bulk out: wMaxPacketSize != 64");
			if (ep->bEndpointAddress != 2)
				warn("bulk out: bEndpointAddress: %d",
				     ep->bEndpointAddress);
			dev->write_pipe = usb_sndbulkpipe(udev,
							  ep->
							  bEndpointAddress);
			dev->bap_buf = kmalloc(BULK_BUF_SIZE, GFP_KERNEL);
			if (!dev->bap_buf) {
				err("Couldn't allocate bulk_out_buffer");
				goto error;
			}
		}
	}
	if (!dev->bap_buf || !dev->read_urb) {
		err("Didn't find the required bulk endpoints");
		goto error;
	}

	if (card_variant != -1)
		dev->card_variant = card_variant;
	else
		dev->card_variant = default_card_variant;

	if (request_firmware(&fw_entry, "orinoco_ezusb_fw",
			     fw_dev_param) == 0) {
		firmware.size = fw_entry->size;
		firmware.code = fw_entry->data;
		find_fw_variant_offset(&firmware);
	}
	if (firmware.size && firmware.code) {
		ezusb_firmware_download(dev, &firmware);
	} else {
		err("No firmware to download");
		goto error;
	}

	if (ezusb_hard_reset(priv) < 0) {
		err("Cannot reset the device");
		goto error;
	}

	/* If the firmware is already downloaded orinoco.c will call
	 * ezusb_init but if the firmware is not already there, that will make
	 * the kernel very unstable, so we try initializing here and quit in
	 * case of error */
	if (ezusb_init(hw) < 0) {
		err("Couldn't initialize the device");
		err("Firmware may not be downloaded or may be wrong.");
		if (card_variant == -1) {
			err("Replug the device to pretend it is the other"
			    " variant");
			default_card_variant = !default_card_variant;
		}
		goto error;
	}
	if (card_variant == -1) {
		printk(KERN_NOTICE PFX "Your device variant seems to be "
		       "0x%02X, please send this message and the output "
		       "of 'lsusb -v' to orinoco-usb-devel@ranty.ddts.net",
		       dev->card_variant);
	}

	SET_MODULE_OWNER(net_dev);
	SET_NETDEV_DEV(net_dev, &interface->dev);
	dev->net_dev = net_dev;
	if (register_netdev(net_dev) != 0) {
		dev->net_dev = NULL;
		err("%s: register_netdev() failed", __FUNCTION__);
		goto error;
	}
	goto exit;

      error:
	ezusb_delete(dev);
	if (net_dev) {
		/* dev->net_dev was 0, so ezusb_delete() didn't free it */
		free_orinocodev(net_dev);
	}
	dev = NULL;
	retval = -EFAULT;
      exit:
	if (fw_entry) {
		firmware.code = NULL;
		firmware.size = 0;
		release_firmware(fw_entry);
	}
	usb_set_intfdata(interface, dev);
	return_probe(dev, retval);
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static void ezusb_disconnect(struct usb_device *udev, void *ptr)
{
	struct ezusb_priv *dev = ptr;
#else
static void ezusb_disconnect(struct usb_interface *intf)
{
	struct ezusb_priv *dev = usb_get_intfdata(intf);
	usb_set_intfdata(intf, NULL);
#endif
	ezusb_delete(dev);
	printk(KERN_INFO PFX "Disconnected\n");
}


/* usb specific object needed to register this driver with the usb subsystem */
static struct usb_driver orinoco_driver = {
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,19))
	.owner = THIS_MODULE,
#endif
	.name = DRIVER_NAME,
	.probe = ezusb_probe,
	.disconnect = ezusb_disconnect,
	.id_table = ezusb_table,
};

/* Can't be declared "const" or the whole __initdata section will
 * become const */
static char version[] __initdata = DRIVER_NAME " " DRIVER_VERSION
    " (Manuel Estrada Sainz <ranty@debian.org>)";

static int __init ezusb_module_init(void)
{
	int err;

	printk(KERN_DEBUG "%s\n", version);

	if (firmware.size) {
		find_fw_variant_offset(&firmware);
		register_firmware("orinoco_ezusb_fw", firmware.code,
				  firmware.size);
	}

	/* register this driver with the USB subsystem */
	err = usb_register(&orinoco_driver);
	if (err < 0) {
		printk(KERN_ERR PFX "usb_register failed, error %d\n",
		       err);
		return err;
	}

	return 0;
}

static void __exit ezusb_module_exit(void)
{
	/* deregister this driver with the USB subsystem */
	usb_deregister(&orinoco_driver);
}


module_init(ezusb_module_init);
module_exit(ezusb_module_exit);

MODULE_AUTHOR("Manuel Estrada Sainz <ranty@debian.org>");
MODULE_DESCRIPTION
    ("Driver for Orinoco wireless LAN cards using EZUSB bridge");
MODULE_LICENSE("Dual MPL/GPL");
