/*
 * USB Prism driver
 *
 * Copyright (c) 2004 Pavel Roskin <proski@gnu.org>
 *
 * Copyright notice & release notes in file orinoco.c
 */

#define DRIVER_NAME "prism_usb"
#define PFX DRIVER_NAME ": "

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/fcntl.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/smp_lock.h>
#include <linux/usb.h>
#include <linux/timer.h>
#include <asm/io.h>

#include <linux/netdevice.h>
#include <linux/if_arp.h>
#include <linux/etherdevice.h>
#include <linux/wireless.h>

#include "orinoco.h"

/* Forward compatibility macros */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 0)
# define return_probe(priv_data, retval) return retval
# define usb_get_desc(iface) (iface.desc)
#endif

#define BULK_BUF_SIZE		2048

/* Frame types sent to bulk OUT */
#define PRISMUSB_TYPE_TXFRM	0
#define PRISMUSB_TYPE_CMDREQ	1
#define PRISMUSB_TYPE_WRIDREQ	2
#define PRISMUSB_TYPE_RRIDREQ	3
#define PRISMUSB_TYPE_WMEMREQ	4
#define PRISMUSB_TYPE_RMEMREQ	5

/* Frame types received on bulk IN */
#define PRISMUSB_TYPE_INFOFRM	0x8000
#define PRISMUSB_TYPE_CMDRESP	0x8001
#define PRISMUSB_TYPE_WRIDRESP	0x8002
#define PRISMUSB_TYPE_RRIDRESP	0x8003
#define PRISMUSB_TYPE_WMEMRESP	0x8004
#define PRISMUSB_TYPE_RMEMRESP	0x8005
#define PRISMUSB_TYPE_BUFAVAIL	0x8006
#define PRISMUSB_TYPE_ERROR	0x8007

static struct usb_device_id prism_usb_table[] = {
	{USB_DEVICE(0x04bb, 0x0922)},	/* IOData AirPort WN-B11/USBS */
	{USB_DEVICE(0x07aa, 0x0012)},	/* Corega Wireless LAN USB Stick-11 */
	{USB_DEVICE(0x09aa, 0x3642)},	/* Prism2.x 11Mbps WLAN USB Adapter */
	{USB_DEVICE(0x1668, 0x0408)},	/* Actiontec Prism2.5 11Mbps WLAN USB Adapter */
	{USB_DEVICE(0x1668, 0x0421)},	/* Actiontec Prism2.5 11Mbps WLAN USB Adapter */
	{USB_DEVICE(0x066b, 0x2212)},	/* Linksys WUSB11v2.5 11Mbps WLAN USB Adapter */
	{USB_DEVICE(0x066b, 0x2213)},	/* Linksys WUSB12v1.1 11Mbps WLAN USB Adapter */
	{USB_DEVICE(0x067c, 0x1022)},	/* Siemens SpeedStream 1022 11Mbps WLAN USB Adapter */
	{USB_DEVICE(0x049f, 0x0033)},	/* Compaq/Intel W100 PRO/Wireless 11Mbps multiport WLAN Adapter */
	{USB_DEVICE(0x0411, 0x0016)},	/* Melco WLI-USB-S11 11Mbps WLAN Adapter */
	{USB_DEVICE(0x08de, 0x7a01)},	/* PRISM25 IEEE 802.11 Mini USB Adapter */
	{USB_DEVICE(0x8086, 0x1111)},	/* Intel PRO/Wireless 2011B LAN USB Adapter */
	{USB_DEVICE(0x0d8e, 0x7a01)},	/* PRISM25 IEEE 802.11 Mini USB Adapter */
	{USB_DEVICE(0x045e, 0x006e)},	/* Microsoft MN510 Wireless USB Adapter */
	{USB_DEVICE(0x0967, 0x0204)},	/* Acer Warplink USB Adapter */
	{USB_DEVICE(0x0cde, 0x0002)},	/* Z-Com 725/726 Prism2.5 USB/USB Integrated */
	{USB_DEVICE(0x413c, 0x8100)},	/* Dell TrueMobile 1180 Wireless USB Adapter */
	{USB_DEVICE(0x0b3b, 0x1601)},	/* ALLNET 0193 11Mbps WLAN USB Adapter */
	{USB_DEVICE(0x0b3b, 0x1602)},	/* ZyXEL ZyAIR B200 Wireless USB Adapter */
	{USB_DEVICE(0x0baf, 0x00eb)},	/* USRobotics USR1120 Wireless USB Adapter */
	{USB_DEVICE(0x0411, 0x0027)},	/* Melco WLI-USB-KS11G 11Mbps WLAN Adapter */
	{USB_DEVICE(0x04f1, 0x3009)},	/* JVC MP-XP7250 Builtin USB WLAN Adapter */
	{USB_DEVICE(0x0846, 0x4110)},	/* NetGear MA111 */
	{USB_DEVICE(0x03f3, 0x0020)},	/* Adaptec AWN-8020 USB WLAN Adapter */
	{USB_DEVICE(0x2821, 0x3300)},	/* ASUS-WL140 Wireless USB Adapter */
	{USB_DEVICE(0x2001, 0x3700)},	/* DWL-122 Wireless USB Adapter */
	{USB_DEVICE(0x50c2, 0x4013)},	/* Averatec USB WLAN Adapter */
	{USB_DEVICE(0x2c02, 0x14ea)},	/* Planex GW-US11H WLAN USB Adapter */
	{USB_DEVICE(0x124a, 0x168b)},	/* Airvast PRISM3 WLAN USB Adapter */
	{USB_DEVICE(0x083a, 0x3503)},	/* T-Sinus 111 USB WLAN Adapter */
	{USB_DEVICE(0x2821, 0x3300)},	/* Hawking HighDB USB Adapter */
	{USB_DEVICE(0x0411, 0x0044)},	/* Melco WLI-USB-KB11 11Mbps WLAN Adapter */
	{USB_DEVICE(0x1668, 0x6106)},	/* ROPEX FreeLan 802.11b USB Adapter */
	{ /* Terminating entry */ }
};

MODULE_DEVICE_TABLE(usb, prism_usb_table);

/* Structure to hold all of our device specific stuff */
struct prism_usb_priv {
	struct usb_device *udev;
	struct net_device *net_dev;
	struct semaphore sem;
	u16 hermes_reg_fake[0x40];
	u8 *bap_buf;
	struct urb *read_urb;
	int read_pipe;
	int write_pipe;
};

struct prism_usb_packet {
	u16 type;
	u8 data[0];
} __attribute__ ((packed));

/* State of a request */
typedef enum {
	CTX_START,
	CTX_QUEUED,
	CTX_REQ_SUBMITTED,
	CTX_REQ_COMPLETE,
	CTX_RESP_RECEIVED,
	CTX_REQ_TIMEOUT,
	CTX_REQ_FAILED,
	CTX_RESP_TIMEOUT,
	CTX_REQSUBMIT_FAIL,
	CTX_COMPLETE,
} prism_usb_state_t;

/* Data associated with one particular request */
struct request_context {
	struct list_head list;
	atomic_t refcount;
	struct completion done;	/* Signals that CTX is dead */
	int killed;
	struct urb *outurb;	/* OUT for req pkt */
	struct prism_usb_priv *dev;
	struct prism_usb_packet *buf;
	int buf_length;
	struct timer_list timer;	/* Timeout handling */
	prism_usb_state_t state;	/* Current state */
};

/* Debug functions */
static inline void hex_dump(const void *data_p, int size)
{
	int i;
	const u8 *data = data_p;

	printk(KERN_DEBUG "Dump at %p, length: %d\n", data, size);

	for (i = 0; i < size; i++) {
		if ((i & 0xf) == 0)
			printk(KERN_DEBUG "0x%04X: ", i);
		if ((i & 3) == 0)
			printk(" ");
		printk("%02X ", data[i]);
		if ((i & 0xf) == 0xf)
			printk("\n");
	}
	if ((i & 0xf) != 0)
		printk("\n");
}

static void prism_usb_req_timeout(unsigned long _ctx)
{
	printk("prism_usb_req_timeout - dummy\n");
}

struct request_context *prism_usb_alloc_ctx(struct prism_usb_priv *dev)
{
	struct request_context *ctx;

	ctx = kmalloc(sizeof(*ctx), GFP_ATOMIC);
	if (!ctx) {
		return NULL;
	}
	memset(ctx, 0, sizeof(*ctx));

	ctx->buf = kmalloc(BULK_BUF_SIZE, GFP_ATOMIC);
	if (!ctx->buf) {
		kfree(ctx);
		return NULL;
	}
	ctx->outurb = usb_alloc_urb(0, GFP_ATOMIC);
	if (!ctx->outurb) {
		kfree(ctx->buf);
		kfree(ctx);
		return NULL;
	}

	ctx->dev = dev;
	ctx->state = CTX_START;

	atomic_set(&ctx->refcount, 1);
	init_completion(&ctx->done);

	init_timer(&ctx->timer);
	ctx->timer.function = prism_usb_req_timeout;
	ctx->timer.data = (u_long) ctx;
	return ctx;
}

static void prism_usb_bulkout_cb(struct urb *urb, struct pt_regs *pt_regs)
{
	printk("prism_usb_bulkout_cb - dummy\n");
}

static int prism_usb_init(hermes_t *hw)
{
	printk("prism_usb_init - dummy\n");
	return 0;
}

static int prism_usb_docmd_wait(hermes_t *hw, u16 cmd, u16 parm0,
				struct hermes_response *resp)
{
	struct prism_usb_priv *dev = hw->priv;
	struct request_context *ctx;
	int err = 0;

	u16 data[] = {
		cpu_to_le16(PRISMUSB_TYPE_CMDREQ),
		cpu_to_le16(cmd),
		cpu_to_le16(parm0),
		cpu_to_le16(0),
		cpu_to_le16(0),
	};

	ctx = prism_usb_alloc_ctx(dev);
	if (!ctx)
		return -ENOMEM;

	memcpy(ctx->buf, data, sizeof(data));

	/* Fill the out packet */
	usb_fill_bulk_urb(ctx->outurb, dev->udev, dev->write_pipe, ctx->buf,
			  ALIGN(sizeof(data), 64), prism_usb_bulkout_cb, ctx);

	/* FIXME: submit URB here */

	if (ctx->state != CTX_COMPLETE)
		err = -EIO;

	complete(&ctx->done);
	kfree(ctx);
	return err;
}

static int prism_usb_allocate(struct hermes *hw, u16 size, u16 *fid)
{
	printk("prism_usb_allocate - dummy\n");
	return 0;
}

static int prism_usb_read_ltv(hermes_t *hw, int bap, u16 rid,
			      unsigned bufsize, u16 *length, void *buf)
{
	printk("prism_usb_read_ltv - dummy\n");
	return -EOPNOTSUPP;
}

static int prism_usb_write_ltv(hermes_t *hw, int bap, u16 rid,
			       u16 length, const void *data)
{
	printk("prism_usb_write_ltv - dummy\n");
	return -EOPNOTSUPP;
}

static int prism_usb_bap_pread(struct hermes *hw, int bap,
			       void *buf, unsigned len, u16 id, u16 offset)
{
	printk("prism_usb_bap_pread - dummy\n");
	return -EOPNOTSUPP;
}

static int prism_usb_hard_reset(struct orinoco_private *priv)
{
	printk("prism_usb_hard_reset - dummy\n");
	return 0;
}

static int prism_usb_xmit(struct sk_buff *skb, struct net_device *net_dev)
{
	printk("prism_usb_xmit - dummy\n");
	return -EOPNOTSUPP;
}

static inline void prism_usb_delete(struct prism_usb_priv *dev)
{
	struct net_device *net_dev;

	BUG_ON(in_interrupt());
	BUG_ON(!dev);

	net_dev = dev->net_dev;
	down(&dev->sem);

	dev->udev = NULL;

	up(&dev->sem);

	kfree(dev->read_urb->transfer_buffer);
	if (dev->read_urb != NULL)
		usb_free_urb(dev->read_urb);
	if (dev->net_dev) {
		unregister_netdev(dev->net_dev);
		free_orinocodev(dev->net_dev);
	}
}

static const struct hermes_ops prism_usb_ops = {
	.init = prism_usb_init,
	.docmd_wait = prism_usb_docmd_wait,
	.allocate = prism_usb_allocate,
	.read_ltv = prism_usb_read_ltv,
	.write_ltv = prism_usb_write_ltv,
	.bap_pread = prism_usb_bap_pread
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static void *prism_usb_probe(struct usb_device *udev, unsigned int ifnum,
			     const struct usb_device_id *id)
{
	struct usb_interface *interface = &udev->actconfig->interface[ifnum];
#else
static int prism_usb_probe(struct usb_interface *interface,
			   const struct usb_device_id *id)
{
	struct usb_device *udev = interface_to_usbdev(interface);
#endif
	struct orinoco_private *priv;
	struct net_device *net_dev;
	hermes_t *hw;
	struct prism_usb_priv *dev = NULL;
	struct usb_interface_descriptor *iface_desc;
	struct usb_endpoint_descriptor *ep;
	int err;

	net_dev = alloc_orinocodev(sizeof(*dev), prism_usb_hard_reset);
	if (!net_dev) {
		printk(KERN_ERR PFX "Cannot allocate network device\n");
		err = -ENOMEM;
		goto exit;
	}

	priv = netdev_priv(net_dev);
	hw = &priv->hw;

	dev = priv->card;

	init_MUTEX(&dev->sem);

	dev->udev = udev;

	hw->iobase = (unsigned long)&dev->hermes_reg_fake;
	hw->io_space = HERMES_MEM;
	hw->reg_spacing = HERMES_16BIT_REGSPACING;
	hw->priv = dev;
	hw->ops = &prism_usb_ops;
	net_dev->hard_start_xmit = prism_usb_xmit;

	priv->irq_no_disable = 1;

	/* Set up the endpoints */
	iface_desc = &usb_get_desc(interface->altsetting[0]);
	if (iface_desc->bNumEndpoints != 3) {
		printk(KERN_ERR PFX "Unexpected number of endpoints: %d\n",
		       iface_desc->bNumEndpoints);
		err = -ENODEV;
		goto error;
	}

	/* Endpoint 1 - bulk in */
	ep = &usb_get_desc(interface->altsetting[0].endpoint[0]);
	if (ep->wMaxPacketSize != 64) {
		printk(KERN_ERR PFX "Unexpected bulk IN packet size: %d\n",
		       ep->wMaxPacketSize);
		err = -ENODEV;
		goto error;
	}

	if (((ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK) != USB_DIR_IN) ||
	    ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
	     != USB_ENDPOINT_XFER_BULK)) {
		printk(KERN_ERR PFX "Endpoint 1 - not bulk IN\n");
		err = -ENODEV;
		goto error;
	}

	dev->read_urb = usb_alloc_urb(0, GFP_KERNEL);
	if (!dev->read_urb) {
		printk(KERN_ERR PFX "No free urbs available\n");
		err = -ENOMEM;
		goto error;
	}
	dev->read_pipe = usb_rcvbulkpipe(udev, ep->bEndpointAddress);
	dev->read_urb->transfer_buffer = kmalloc(BULK_BUF_SIZE, GFP_KERNEL);
	if (!dev->read_urb->transfer_buffer) {
		printk(KERN_ERR PFX "Cannot allocate IN buffer\n");
		err = -ENOMEM;
		goto error;
	}

	/* Endpoint 2 - bulk out */
	ep = &usb_get_desc(interface->altsetting[0].endpoint[1]);
	if (ep->wMaxPacketSize != 64) {
		printk(KERN_ERR PFX "Unexpected bulk OUT packet size: %d\n",
		       ep->wMaxPacketSize);
		err = -ENODEV;
		goto error;
	}

	if (((ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK) != USB_DIR_OUT) ||
	    ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) !=
	     USB_ENDPOINT_XFER_BULK)) {
		printk(PFX "Endpoint 2 - not bulk OUT\n");
		err = -EBUSY;
		goto error;
	}

	dev->write_pipe = usb_sndbulkpipe(udev, ep->bEndpointAddress);
	dev->bap_buf = kmalloc(BULK_BUF_SIZE, GFP_KERNEL);
	if (!dev->bap_buf) {
		printk(KERN_ERR PFX "Cannot allocate bulk OUT buffer\n");
		err = -ENOMEM;
		goto error;
	}

	err = prism_usb_hard_reset(priv);
	if (err) {
		printk(KERN_ERR PFX "Cannot reset the device\n");
		goto error;
	}

	err = prism_usb_init(hw);
	if (err) {
		printk(KERN_ERR PFX "Cannot initialize the device\n");
		goto error;
	}

	SET_MODULE_OWNER(net_dev);
	SET_NETDEV_DEV(net_dev, &interface->dev);
	dev->net_dev = net_dev;
	err = register_netdev(net_dev);
	if (err) {
		dev->net_dev = NULL;
		printk(KERN_ERR PFX "Cannot register network device\n");
		goto error;
	}
	goto exit;

      error:
	prism_usb_delete(dev);
	if (net_dev) {
		/* dev->net_dev was 0, so prism_usb_delete() didn't free it */
		free_orinocodev(net_dev);
	}
	dev = NULL;
      exit:
	usb_set_intfdata(interface, dev);
	return_probe(dev, err);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static void prism_usb_disconnect(struct usb_device *udev, void *ptr)
{
	struct prism_usb_priv *dev = ptr;
#else
static void prism_usb_disconnect(struct usb_interface *intf)
{
	struct prism_usb_priv *dev = usb_get_intfdata(intf);
	usb_set_intfdata(intf, NULL);
#endif
	prism_usb_delete(dev);
	printk(KERN_INFO PFX "Disconnected\n");
}

/* usb specific object needed to register this driver with the usb subsystem */
static struct usb_driver prism_usb_driver = {
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,19))
	.owner = THIS_MODULE,
#endif
	.name = DRIVER_NAME,
	.probe = prism_usb_probe,
	.disconnect = prism_usb_disconnect,
	.id_table = prism_usb_table,
};

/* Can't be declared "const" or the whole __initdata section will
 * become const */
static char version[] __initdata = DRIVER_NAME " " DRIVER_VERSION
    " (Pavel Roskin <proski@gnu.org>)";

static int __init prism_usb_module_init(void)
{
	int err;

	printk(KERN_DEBUG "%s\n", version);

	/* register this driver with the USB subsystem */
	err = usb_register(&prism_usb_driver);
	if (err < 0) {
		printk(KERN_ERR PFX "usb_register failed, error %d\n", err);
		return err;
	}

	return 0;
}

static void __exit prism_usb_module_exit(void)
{
	/* deregister this driver with the USB subsystem */
	usb_deregister(&prism_usb_driver);
}

module_init(prism_usb_module_init);
module_exit(prism_usb_module_exit);

MODULE_AUTHOR("Pavel Roskin <proski@gnu.org>");
MODULE_DESCRIPTION("Driver for Prism USB based wireless devices");
MODULE_LICENSE("Dual MPL/GPL");
