/***************************************************************************
 * Generic video post-processing functions for the W996[87]CF driver       *
 * for Linux.                                                              *
 *                                                                         *
 * Copyright (C) 2002-2005 by Luca Risolia <luca.risolia@studio.unibo.it>  *
 *                                                                         *
 * - Colorspace UYVY->RGBX conversion routine from CPiA driver by          *
 *   Peter Pregler, Scott J. Bertin and Johannes Erdfelt.                  *
 *                                                                         *
 * This program is free software; you can redistribute it and/or modify    *
 * it under the terms of the GNU General Public License as published by    *
 * the Free Software Foundation; either version 2 of the License, or       *
 * (at your option) any later version.                                     *
 *                                                                         *
 * This program is distributed in the hope that it will be useful,         *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of          *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
 * GNU General Public License for more details.                            *
 *                                                                         *
 * You should have received a copy of the GNU General Public License       *
 * along with this program; if not, write to the Free Software             *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               *
 ***************************************************************************/

#include <linux/videodev.h>
#include <linux/string.h>
#include <asm/byteorder.h>

#include "w9968cf_vppfuncs.h"


/*--------------------------------------------------------------------------
  This function changes the original UYVY-16bit component ordering of a frame
  --------------------------------------------------------------------------*/
inline void w9968cf_swap_yuvbytes(void* pIn, unsigned long length)
{
	u16* word = (u16*)pIn;

	while (length >= 2) {
		*word = swab16(*word);
		word++;
		length -= 2;
	}
}


/*-------------------------------------------------------------------------- 
  This function centers the original image in a larger one.
  You may want to write a real upscaling algorithm with integer arithmetics
  and interpolation...
  --------------------------------------------------------------------------*/
void 
w9968cf_scale_up(u8* pIn, u8* pOut, u16 w0, u16 h0, u16 bpp, u16 w1, u16 h1)
{
	u16 cs = bpp/8, a, b, r;

	if ((h1 <= h0) || (w1 <= w0)) /* just in case */
		return;

	a = (h1 - h0) / 2;
	b = (w1 - w0) / 2;

	memset(pOut, 0, a*w1*cs);
	for (r = 0; r < h0-1; r++) {
		memset(pOut+(a+r)*w1*cs, 0, b*cs);
		memcpy(pOut+(a+r)*w1*cs + b*cs, pIn+r*cs*w0, w0*cs);
		memset(pOut+(a+r)*w1*cs + (b+w0)*cs, 0, b*cs);
	}
	memset(pOut + (a+h0)*w1*cs, 0, w1*cs*(h1-h0-a));
}


/*--------------------------------------------------------------------------
  Colorspace conversion from UYVY-16 bit to RGBX.
  --------------------------------------------------------------------------*/
void w9968cf_uyvy_to_rgbx(u8* pIn, unsigned long length, u8* pOut,
                          u16 out_fmt, u8 force_rgb)
{
	int y=0, u=0, y1=0, v=0, r=0, g=0, b=0;

	#define LIMIT(x) ((( (x) > 0xffffff) ? 0xff0000 : \
	                  (( (x) <= 0xffff ) ? 0 : (x) & 0xff0000)) / 0xffff)

	while (length >= 4) {
		length -= 4;

	if (out_fmt == VIDEO_PALETTE_GREY) {
		y = *(++pIn);
		pIn++;
		y1 = *(++pIn);
		pIn++;
	} else {
		u = *pIn++ - 128;
		y = (*pIn++ - 16) * 76310;
		v = *pIn++ - 128;
		y1 = (*pIn++ - 16) * 76310;
		r = 104635 * v;
		g = -25690 * u + -53294 * v;
		b = 132278 * u;
	}

		switch(out_fmt) 
		{
		case VIDEO_PALETTE_RGB555:
			*pOut++ = ((LIMIT(g+y) & 0xf8) << 2) |
			          (LIMIT(b+y) >> 3);
			*pOut++ = ((LIMIT(r+y) & 0xf8) >> 1) |
			          (LIMIT(g+y) >> 6);
			*pOut++ = ((LIMIT(g+y1) & 0xf8) << 2) |
			          (LIMIT(b+y1) >> 3);
			*pOut++ = ((LIMIT(r+y1) & 0xf8) >> 1) | 
			          (LIMIT(g+y1) >> 6);
			break;
		case VIDEO_PALETTE_RGB565:
			*pOut++ = ((LIMIT(g+y) & 0xfc) << 3) |
			          (LIMIT(b+y) >> 3);
			*pOut++ = (LIMIT(r+y) & 0xf8) |
			          (LIMIT(g+y) >> 5);
			*pOut++ = ((LIMIT(g+y1) & 0xfc) << 3) |
			          (LIMIT(b+y1) >> 3);
			*pOut++ = (LIMIT(r+y1) & 0xf8) |
			          (LIMIT(g+y1) >> 5);
			break;
		case VIDEO_PALETTE_RGB24:
			if (!force_rgb) {
				*pOut++ = LIMIT(b+y);
				*pOut++ = LIMIT(g+y);
				*pOut++ = LIMIT(r+y);
				*pOut++ = LIMIT(b+y1);
				*pOut++ = LIMIT(g+y1);
				*pOut++ = LIMIT(r+y1);
			} else {
				*pOut++ = LIMIT(r+y);
				*pOut++ = LIMIT(g+y);
				*pOut++ = LIMIT(b+y);
				*pOut++ = LIMIT(r+y1);
				*pOut++ = LIMIT(g+y1);
				*pOut++ = LIMIT(b+y1);
			}
			break;
		case VIDEO_PALETTE_RGB32:
			if (!force_rgb) {
				*pOut++ = LIMIT(b+y);
				*pOut++ = LIMIT(g+y);
				*pOut++ = LIMIT(r+y);
				pOut++;
				*pOut++ = LIMIT(b+y1);
				*pOut++ = LIMIT(g+y1);
				*pOut++ = LIMIT(r+y1);
				pOut++;
			} else {
				*pOut++ = LIMIT(r+y);
				*pOut++ = LIMIT(g+y);
				*pOut++ = LIMIT(b+y);
				pOut++;
				*pOut++ = LIMIT(r+y1);
				*pOut++ = LIMIT(g+y1);
				*pOut++ = LIMIT(b+y1);
				pOut++;
			}
			break;
		case VIDEO_PALETTE_GREY:
			*pOut++ = y * 256 >> 8;
			*pOut++ = y1 * 256 >> 8;
			break;
		}
	}
}
