/*
 *  LIDS LOG functions 
 *
 *  Copyright (C) 2002-2003 Huagang Xie <xie@lids.org>
 *  Copyright (C) 2002 Philippe Biondi <biondi@cartel-securite.fr>
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 */

#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/lids.h>

#define STR2(x) #x
#define STR(x) STR2(x)

#define lids_print(message, args...) printk(KERN_ALERT message , ## args)

/*
 * copy from driver/tty/tty_io.c  
 *
 * This routine returns the name of tty.
 */
#define LIDS_NULL_TTY "NULL tty\0"
static char *
_lids_tty_make_name(struct tty_struct *tty, const char *name, char *buf)
{

	if (!tty)		/* Hmm.  NULL pointer.  That's fun. */
		strncpy(buf, "NULL tty\0",64);
	else
		snprintf(buf, 64, name, tty->name);

	return buf;
}

char *
lids_tty_name(struct tty_struct *tty, char *buf)
{
	return _lids_tty_make_name(tty, (tty) ? tty->name : NULL, buf);
}

/* return current dentry */
static struct dentry *
lids_current_dentry(void)
{
	struct dentry *f_dentry = NULL;
	struct vm_area_struct *vma;

	if (current->mm) {
		vma = current->mm->mmap;
		while (vma) {
			if ((vma->vm_flags & VM_EXECUTABLE) && vma->vm_file) {
				f_dentry = vma->vm_file->f_dentry;
				break;
			}
			vma = vma->vm_next;
		}
	}
	return f_dentry;
}

void
lids_log(int flood, const char *message, ...)
{
	va_list args;
	char ttyname[64];
	char progname[64];
	char proginfo[64 + 10 + 10 + 20 + 24];	/* %s+%d+%d+%ld+le texte avec un peu de marge = 128 */
	char msgstr[256];
	u32 parent_pid;

	struct dentry *f_dentry;

	/* Get args on the stack */
	va_start(args, message);

	/* Get dentry of current process, if any */
	f_dentry = lids_current_dentry();
	/* Get the tty name, if any */
	memset(ttyname, '\0', 64);
/* modules do not support the ttyname right now */
	lids_tty_name(current->signal->tty, ttyname);

	/* Make the proginfo string */
	if (f_dentry && f_dentry->d_inode) {
		strncpy(progname, f_dentry->d_iname, 63);
		snprintf(proginfo, 127, "%s (dev %d:%d inode %ld)",
			 progname,
			 MAJOR(f_dentry->d_inode->i_sb->s_dev),
			 MINOR(f_dentry->d_inode->i_sb->s_dev),
			 f_dentry->d_inode->i_ino);
	} else {
		strncpy(proginfo, "(undetermined program)\0", 63);
	}

	/* Make the message string */
	vsnprintf(msgstr, 255, message, args);

	parent_pid = current->parent->pid;

	/* Make the log string */

	lids_print("LIDS: %s pid %d ppid %d uid/gid (%d/%d) on (%s) : %s %s\n",
		   proginfo,
		   current->pid,
		   parent_pid,
		   current->uid,
		   current->gid,
		   ttyname,
		   msgstr,
		   flood ? " - logging disabled for "
		   STR(LIDS_TIMEOUT_AFTER_FLOOD) "s" : "");

	/* deal with args on the stack */
	va_end(args);

}

/* sent out message */
void
lids_alert(int type, long dst, long dst2, char *name, char *action)
{
	struct dentry *f_dentry;

	switch (type) {
	case LIDS_CAP:
		lids_security_alert("violate %s", action);
		break;
	case LIDS_SOCKET:
		lids_security_alert("attempt to %s", action);
		break;
	case LIDS_READONLY:
		/* compatible to the acl */
		type = 1;
		lids_security_alert("attempt to %s %s for reading", action,
				    name);
		break;
	case LIDS_APPEND:
		type = 3;
		lids_security_alert("attempt to %s %s for appending", action,
				    name);
		break;
	case LIDS_WRITE:
		type = 7;
		lids_security_alert("attempt to %s %s for writing", action,
				    name);
		break;
	default:
		lids_security_alert("yeee, alert type mismatch");
		break;
	}
	/* if in acl_discovery mode, print out the acl_discovery mode string */
	if (lids_acl_discovery) {
		f_dentry = lids_current_dentry();
		printk(KERN_INFO
		       "LIDS_ACL_DISCOVERY:[state %d]%ld:%d:%s:%d:0:%ld:%ld:%s:0-0\n",
		       lids_state, f_dentry->d_inode->i_ino,
		       f_dentry->d_inode->i_sb->s_dev, f_dentry->d_iname, type,
		       dst, dst2, name);
	}
}
