/*
 * Linux Intrusion Detectino System for Linux Security Modules project
 *
 * Copyright (C) 2002-2004 Huagang Xie (xie@www.lids.org) 
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *  Feb 4th, 2002, Huagang, Initial the project  
 *	
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/security.h>
#include <linux/skbuff.h>
#include <linux/netlink.h>
#include <linux/ip.h>
#include <linux/lids.h>
#include <linux/lidsext.h>
#include <linux/lidsif.h>
#include <net/sock.h>
#include <linux/in.h>
#include <linux/time.h>
#include <linux/mount.h>
#include <linux/namei.h>

struct security_operations *lids_secondary_ops;

static int
lids_ptrace(struct task_struct *parent, struct task_struct *child)
{
	if (lids_load && lids_local_load) {
		if (lids_check_capable(parent, CAP_SYS_PTRACE, 1)) {
			lids_security_alert("Attempt to trace pid %i\n",
					    child->pid);
			return -EPERM;
		}
	}
	return 0;
}

/* from security/commoncap.c  */
int lids_capget (struct task_struct *target, kernel_cap_t *effective,
		kernel_cap_t *inheritable, kernel_cap_t *permitted)
{
	/* Derived from kernel/capability.c:sys_capget. */
	*effective = cap_t (target->cap_effective);
	*inheritable = cap_t (target->cap_inheritable);
	*permitted = cap_t (target->cap_permitted);
	return 0;
}

/* from security/commoncap.c  */
void lids_capset_set (struct task_struct *target, kernel_cap_t *effective,
		     kernel_cap_t *inheritable, kernel_cap_t *permitted)
{
	target->cap_effective = *effective;
	target->cap_inheritable = *inheritable;
	target->cap_permitted = *permitted;
}

/* derived from security/commoncap.c  */
int lids_capset_check (struct task_struct *target, kernel_cap_t *effective,
		      kernel_cap_t *inheritable, kernel_cap_t *permitted)
{
	if (lids_load && lids_local_load) {
		if (lids_check_capable(target, CAP_SETPCAP, 1)) {
			lids_security_alert("Attempt to setpcap pid %i\n",
					    target->pid);
			return -EPERM;
		}
	}
	/* Derived from kernel/capability.c:sys_capset. */
	/* verify restrictions on target's new Inheritable set */
	if (!cap_issubset (*inheritable,
			   cap_combine (target->cap_inheritable,
					current->cap_permitted))) {
		return -EPERM;
	}

	/* verify restrictions on target's new Permitted set */
	if (!cap_issubset (*permitted,
			   cap_combine (target->cap_permitted,
					current->cap_permitted))) {
		return -EPERM;
	}

	/* verify the _new_Effective_ is a subset of the _new_Permitted_ */
	if (!cap_issubset (*effective, *permitted)) {
		return -EPERM;
	}

	return 0;
}

static int
lids_capable(struct task_struct *tsk, int cap)
{
/* the original one */
	if (cap_is_fs_cap(cap) ? tsk->fsuid == 0 : tsk->euid == 0) {
		if (lids_load && lids_local_load) {
			return lids_check_capable(tsk, cap, 1);
		}
		/* capability granted */
		return 0;
	}
	/* capability denied */
	return -EPERM;
}

static int
lids_bprm_alloc_security(struct linux_binprm *bprm)
{
	int rc = 0;
	if (lids_execve(bprm))
		return -1;

	if (lids_secondary_ops)
		rc = lids_secondary_ops->bprm_alloc_security(bprm);

	return rc;
}

static void
lids_bprm_free_security(struct linux_binprm *bprm)
{
	if (lids_secondary_ops)
		lids_secondary_ops->bprm_free_security(bprm);
	return;
}

static void
lids_bprm_apply_creds (struct linux_binprm *bprm, int unsafe)
{
	if (bprm->e_uid != current->uid || bprm->e_gid != current->gid) {
		current->mm->dumpable = 0;

		if ((unsafe & ~LSM_UNSAFE_PTRACE_CAP) && !capable(CAP_SETUID)) {
			bprm->e_uid = current->uid;
			bprm->e_gid = current->gid;
		}
	}

	current->suid = current->euid = current->fsuid = bprm->e_uid;
	current->sgid = current->egid = current->fsgid = bprm->e_gid;

	if (lids_secondary_ops)
		lids_secondary_ops->bprm_apply_creds(bprm,unsafe);
	return;
}

static int
lids_bprm_set_security(struct linux_binprm *bprm)
{
	int rc = 0;

	if (lids_secondary_ops)
		rc = lids_secondary_ops->bprm_set_security(bprm);

	return rc;
}
static int
lids_bprm_check_security(struct linux_binprm *bprm)
{
	lids_execve_check_envp(bprm);
	return 0;
}

#ifndef MODULE
static int __init
lids_load_setup(char *str)
{
	lids_load = simple_strtol(str, NULL, 0);
	return 1;
}
#endif

static int lids_sb_mount (char *dev_name, struct nameidata *nd, char *type,
			   unsigned long flags, void *data)
{
	struct dentry *dentry = nd->dentry;

	if(IS_ROOT(dentry)) {
		if(!lids_init_setup) {
			do_lids_setup();
		}
	}
	return 0;
}

static int lids_sb_kern_mount (struct super_block *sb, void *data)
{
	return 0;
}

static void lids_sb_post_remount (struct vfsmount *mnt, unsigned long flags,
				   void *data)
{
	struct vfsmount *parent;

//	spin_lock(&vfsmount_lock);
	parent=mnt->mnt_parent;

        if (parent == mnt || IS_ROOT(mnt->mnt_mountpoint)) {
		if(!lids_init_setup) {
			do_lids_setup();
		}
        }
//	spin_unlock(&vfsmount_lock);

	return;
}

static void
lids_sb_post_mountroot(void)
{
	if(!lids_init_setup) {
		do_lids_setup();
	}
	return;
}

static void
lids_inode_free_security(struct inode *inode)
{
#if 0
	struct ipt_mark_target_info *markinfo = inode->i_security;

	if (!markinfo)
		return;
	inode->i_security = NULL;
	kfree(markinfo);
#endif
	lids_free_inode_acl(inode->i_security);
	return;
}

static void
lids_inode_post_create(struct inode *inode, struct dentry *dentry, int mask)
{
	lids_do_inode_post_create(inode, dentry);
	return;
}

static int
lids_inode_link(struct dentry *old_dentry, struct inode *inode,
		struct dentry *new_dentry)
{
	int rc = 0;

	if (lids_secondary_ops)
		rc = lids_secondary_ops->inode_link(old_dentry, inode,
						    new_dentry);

	return rc;
}

static int
lids_inode_unlink(struct inode *inode, struct dentry *dentry)
{
	if (lids_load && lids_local_load) {
		if (lids_check_base(dentry, LIDS_WRITE)) {
			lids_alert(LIDS_WRITE, inode->i_ino, inode->i_sb->s_dev,
				   dentry->d_iname, "unlink");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_inode_symlink(struct inode *inode, struct dentry *dentry, const char *name)
{
	if (lids_load && lids_local_load) {
		if (lids_check_base(dentry, LIDS_WRITE)) {
			lids_alert(LIDS_WRITE, inode->i_ino, inode->i_sb->s_dev,
				   dentry->d_iname, "symlink");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_inode_mkdir(struct inode *inode, struct dentry *dentry, int mask)
{
	if (lids_load && lids_local_load) {
		if (lids_check_base(dentry, LIDS_WRITE)) {
			lids_alert(LIDS_WRITE, inode->i_ino, inode->i_sb->s_dev,
				   dentry->d_iname, "mkdir");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static void
lids_inode_post_mkdir(struct inode *inode, struct dentry *dentry, int mask)
{
	lids_do_inode_post_create(inode, dentry);
	return;
}

static int
lids_inode_rmdir(struct inode *inode, struct dentry *dentry)
{
	if (lids_load && lids_local_load) {
		if (lids_check_base(dentry, LIDS_WRITE)) {
			lids_alert(LIDS_WRITE, inode->i_ino, inode->i_sb->s_dev,
				   dentry->d_iname, "rmdir");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_inode_mknod(struct inode *inode, struct dentry *dentry,
		 int major, dev_t minor)
{
	if (lids_load && lids_local_load) {
		if (lids_check_base(dentry, LIDS_WRITE)) {
			lids_alert(LIDS_WRITE, inode->i_ino, inode->i_sb->s_dev,
				   dentry->d_iname, "mknod");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_inode_rename(struct inode *old_inode,
		  struct dentry *old_dentry,
		  struct inode *new_inode, struct dentry *new_dentry)
{
	if (lids_load && lids_local_load) {
		if (lids_check_base(old_dentry, LIDS_WRITE)) {
			lids_alert(LIDS_WRITE, old_inode->i_ino,
				   old_inode->i_sb->s_dev, old_dentry->d_iname,
				   "rename to");
			return LIDS_ERROR(-EPERM);
		}
		if (lids_check_base(new_dentry, LIDS_WRITE)) {
			lids_alert(LIDS_WRITE, new_inode->i_ino,
				   new_inode->i_sb->s_dev, new_dentry->d_iname,
				   "rename from");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_inode_readlink(struct dentry *dentry)
{
	if (lids_load && lids_local_load) {
		if (lids_check_base(dentry, LIDS_READONLY)) {
			lids_alert(LIDS_WRITE, dentry->d_inode->i_ino,
				   dentry->d_inode->i_sb->s_dev,
				   dentry->d_iname, "readlink");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_inode_follow_link(struct dentry *dentry, struct nameidata *nameidata)
{
	int rc = 0;

	if (lids_secondary_ops)
		rc = lids_secondary_ops->inode_follow_link(dentry, nameidata);

	if (rc)
		return rc;

	if (lids_load && lids_local_load) {
		if (lids_check_base(dentry, LIDS_READONLY)) {
			lids_alert(LIDS_WRITE, dentry->d_inode->i_ino,
				   dentry->d_inode->i_sb->s_dev,
				   dentry->d_iname, "followlink");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_inode_permission(struct inode *inode, int mask, struct nameidata *nd)
{
	struct list_head *head, *next, *tmp;
	int error = 0;
	struct dentry *d;

	if (!(lids_load && lids_local_load))
		return 0;

	/* we will not due with other type */

	spin_lock(&dcache_lock);
	head = &inode->i_dentry;
	next = inode->i_dentry.next;

	while (next != head) {
		tmp = next;
		next = tmp->next;
		d = list_entry(tmp, struct dentry, d_alias);

		spin_unlock(&dcache_lock);

		if ((mask & MAY_APPEND)) {
			error = lids_check_base(d, LIDS_APPEND);
			if (error) {
				lids_alert(LIDS_APPEND, inode->i_ino,
					   inode->i_sb->s_dev, d->d_iname,
					   "open");
				error = LIDS_ERROR(-EPERM);
			}
		} else if ((mask & MAY_WRITE)) {
			error = lids_check_base(d, LIDS_WRITE);
			if (error) {
				lids_alert(LIDS_WRITE, inode->i_ino,
					   inode->i_sb->s_dev, d->d_iname,
					   "open");
				error = LIDS_ERROR(-EPERM);
			}
		} else {
			error = lids_check_base(d, LIDS_READONLY);
			if (error) {
				lids_alert(LIDS_READONLY, inode->i_ino,
					   inode->i_sb->s_dev, d->d_iname,
					   "open");
				error = LIDS_ERROR(-ENOENT);
			}

		}
		spin_lock(&dcache_lock);
	}
	spin_unlock(&dcache_lock);
	return error;
}

static int
lids_inode_setattr(struct dentry *dentry, struct iattr *iattr)
{
	if (lids_load && lids_local_load) {
		if (lids_check_base(dentry, LIDS_WRITE)) {
			lids_alert(LIDS_WRITE, dentry->d_inode->i_ino,
				   dentry->d_inode->i_sb->s_dev,
				   dentry->d_iname, "setattr");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_inode_setxattr(struct dentry *dentry, char *name, void *value,
		    size_t size, int flags)
{
	if (!strcmp(name, XATTR_NAME_LIDS)) {
		if (lids_load && lids_local_load) {
			lids_security_alert(" setxattr denied for %s\n", name);
			return -EPERM;
		}
	}
	return 0;
}

static int
lids_inode_getxattr(struct dentry *dentry, char *name)
{
	if (!strcmp(name, XATTR_NAME_LIDS)) {
		if (lids_load && lids_local_load) {
			lids_security_alert(" getxattr denied for %s\n", name);
			return -EPERM;
		}
		/* recal the task */
	}
	return 0;
}

static int
lids_inode_removexattr(struct dentry *dentry, char *name)
{
	if (!strcmp(name, XATTR_NAME_LIDS)) {
		if (lids_load && lids_local_load) {
			lids_security_alert(" removexattr denied for %s\n",
					    name);
			return -EPERM;
		}
	}
	return 0;
}

static void
lids_d_instantiate(struct dentry *dentry, struct inode *inode)
{
	return;
}

static int
lids_file_permission(struct file *file, int mask)
{
	if (lids_load && lids_local_load) {
		struct inode *inode = file->f_dentry->d_inode;

		if (inode && S_ISBLK(inode->i_mode) &&
		    lids_check_capable(current, CAP_SYS_RAWIO, 1)) {
			if (mask & MAY_WRITE)
				lids_security_alert("CAP_SYS_RAWIO violation: "
						    "Attempt to write to raw device %d:%d",
						    MAJOR(inode->i_sb->s_dev),
						    MINOR(inode->i_sb->s_dev));
			else
				lids_security_alert("CAP_SYS_RAWIO violation: "
						    "Attempt to read from raw device %d:%d",
						    MAJOR(inode->i_sb->s_dev),
						    MINOR(inode->i_sb->s_dev));
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_task_alloc_security(struct task_struct *p)
{
	if (lids_fork_task(p))
		return -EPERM;
	return 0;
}

static void
lids_task_free_security(struct task_struct *p)
{
	struct lids_task_acl *task_acl;

	task_acl = p->security;
	task_lock(p);
	p->security = NULL;
	task_unlock(p);

	lids_free_task_acl(task_acl);
	return;
}

static int
lids_task_kill(struct task_struct *p, struct siginfo *info, int sig)
{
	if (lids_load && lids_local_load && p->security) {
		if (lids_check_task_kill(p, info, sig))
			return -EPERM;
	}
	return 0;
}

static void
lids_task_reparent_to_init(struct task_struct *p)
{
	p->euid = p->fsuid = 0;
	return;
}

#ifdef CONFIG_SECURITY_NETWORK

static int
lids_socket_create(int family, int type, int protocol,int kern)
{
	if (kern)
		return 0;
	if (lids_load && lids_local_load && family == AF_INET) {
		if (lids_ext_capable(current, LIDS_SOCKET_CREATE) < 0) {
			lids_alert(LIDS_SOCKET, -1, LIDS_SOCKET_CREATE,
				   "LIDS_SOCKET_CREATE", "create socket");
			return LIDS_ERROR(-EPERM);
		}
		if (type == SOCK_DGRAM) {
			if (lids_ext_capable(current, LIDS_SOCKET_CREATE_UDP) <
			    0) {
				lids_alert(LIDS_SOCKET, -1,
					   LIDS_SOCKET_CREATE_UDP,
					   "LIDS_SOCKET_CREATE_UDP",
					   "create udp socket");
				return LIDS_ERROR(-EPERM);
			}
		} else if (type == SOCK_STREAM) {
			if (lids_ext_capable(current, LIDS_SOCKET_CREATE_TCP) <
			    0) {
				lids_alert(LIDS_SOCKET, -1,
					   LIDS_SOCKET_CREATE_TCP,
					   "LIDS_SOCKET_CREATE_TCP",
					   "create tcp socket");
				return LIDS_ERROR(-EPERM);
			}
		}
	}
	return 0;
}

static void
lids_socket_post_create(struct socket *sock, int family, int type, int protocol, int kern)
{
#if 0
	struct lids_task_acl *task_acl;
	struct ipt_mark_target_info *markinfo;

	if (lids_load && lids_local_load && family == AF_INET) {
		if (lids_ext_capable(current, LIDS_SOCKET_NF_MARK) < 0) {
			struct inode *inode = SOCK_INODE(sock);

			if (inode) {
				task_acl = current->security;
				markinfo =
				    kmalloc(sizeof
					    (struct ipt_mark_target_info),
					    GFP_KERNEL);
				markinfo->mark = task_acl->mark;
				/* FIXME, need lock?? */
				inode->i_security = markinfo;
				LIDS_DBG("DEV: [%d %d] Mark socket as %d \n",
					 current->pid, current->parent->pid,
					 markinfo->mark);
			}
		}
	}
#endif
	return;
}

static int
lids_socket_bind(struct socket *sock, struct sockaddr *address, int addrlen)
{
	struct sockaddr_in *addr = (struct sockaddr_in *) address;
	char str[32];

	if (lids_load && lids_local_load && address
	    && address->sa_family == AF_INET) {

		if (lids_ext_capable(current, LIDS_SOCKET_BIND) < 0 || 
			lids_bind_checker(ntohs(addr->sin_port)) < 0 
			) {
			snprintf(str, 32, "bind %u.%u.%u.%u:%d",
				 NIPQUAD(addr->sin_addr),
				 ntohs(addr->sin_port));
			lids_alert(LIDS_SOCKET, -1, LIDS_SOCKET_BIND,
				   "LIDS_SOCKET_BIND", str);
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_socket_connect(struct socket *sock, struct sockaddr *address, int addrlen)
{
	struct sockaddr_in *addr = (struct sockaddr_in *) address;
	char str[32];

	if (lids_load && lids_local_load
	    && address && address->sa_family == AF_INET) {
		if (lids_ext_capable(current, LIDS_SOCKET_CONNECT) < 0) {
			snprintf(str, 32, "connect %u.%u.%u.%u:%d",
				 NIPQUAD(addr->sin_addr),
				 ntohs(addr->sin_port));
			lids_alert(LIDS_SOCKET, -1, LIDS_SOCKET_CONNECT,
				   "LIDS_SOCKET_CONNECT", str);
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_socket_listen(struct socket *sock, int backlog)
{
	struct sock *sk = sock->sk;

	if (lids_load && lids_local_load && sk->sk_family == PF_INET) {
		if (lids_ext_capable(current, LIDS_SOCKET_LISTEN) < 0) {
			lids_alert(LIDS_SOCKET, -1, LIDS_SOCKET_LISTEN,
				   "LIDS_SOCKET_LISTEN", "listen");
			return LIDS_ERROR(-EPERM);
		}
	}

	return 0;
}

static int
lids_socket_accept(struct socket *sock, struct socket *newsock)
{
	struct sock *sk = sock->sk;

	if (lids_load && lids_local_load && sk->sk_family == PF_INET) {
		if (lids_ext_capable(current, LIDS_SOCKET_ACCEPT) < 0) {
			lids_alert(LIDS_SOCKET, -1, LIDS_SOCKET_ACCEPT,
				   "LIDS_SOCKET_ACCEPT", "accept");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_socket_sendmsg(struct socket *sock, struct msghdr *msg, int size)
{
	struct sock *sk = sock->sk;

	if (lids_load && lids_local_load && sk->sk_family == PF_INET
	    && sk->sk_type == SOCK_DGRAM) {
		if (lids_ext_capable(current, LIDS_SOCKET_SENDMSG) < 0) {
			lids_alert(LIDS_SOCKET, -1, LIDS_SOCKET_SENDMSG,
				   "LIDS_SOCKET_SENDMSG", "sendmsg");
			return LIDS_ERROR(-EPERM);
		}
	}

	return 0;
}

static int
lids_socket_recvmsg(struct socket *sock, struct msghdr *msg,
		    int size, int flags)
{
	struct sock *sk = sock->sk;

	if (lids_load && lids_local_load && sk->sk_family == PF_INET
	    && sk->sk_type == SOCK_DGRAM) {
		if (lids_ext_capable(current, LIDS_SOCKET_RECVMSG) < 0) {
			lids_alert(LIDS_SOCKET, -1, LIDS_SOCKET_RECVMSG,
				   "LIDS_SOCKET_RECVMSG", "recvmsg");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_socket_getsockname(struct socket *sock)
{
	struct sock *sk = sock->sk;

	if (lids_load && lids_local_load && sk->sk_family == PF_INET) {
		if (lids_ext_capable(current, LIDS_SOCKET_GETSOCKNAME) < 0) {
			lids_alert(LIDS_SOCKET, -1, LIDS_SOCKET_GETSOCKNAME,
				   "LIDS_SOCKET_GETSOCKNAME", "getsockname");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_socket_getpeername(struct socket *sock)
{
	struct sock *sk = sock->sk;

	if (lids_load && lids_local_load && sk->sk_family == PF_INET) {
		if (lids_ext_capable(current, LIDS_SOCKET_GETPEERNAME) < 0) {
			lids_alert(LIDS_SOCKET, -1, LIDS_SOCKET_GETPEERNAME,
				   "LIDS_SOCKET_GETPEERNAME", "getpeername");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_socket_setsockopt(struct socket *sock, int level, int optname)
{
	struct sock *sk = sock->sk;

	if (lids_load && lids_local_load && sk->sk_family == PF_INET) {
		if (lids_ext_capable(current, LIDS_SOCKET_SETSOCKOPT) < 0) {
			lids_alert(LIDS_SOCKET, -1, LIDS_SOCKET_SETSOCKOPT,
				   "LIDS_SOCKET_SETSOCKOPT", "setsockopt");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_socket_getsockopt(struct socket *sock, int level, int optname)
{
	struct sock *sk = sock->sk;

	if (lids_load && lids_local_load && sk->sk_family == PF_INET) {
		if (lids_ext_capable(current, LIDS_SOCKET_GETSOCKOPT) < 0) {
			lids_alert(LIDS_SOCKET, -1, LIDS_SOCKET_GETSOCKOPT,
				   "LIDS_SOCKET_GETSOCKOPT", "getsockopt");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}

static int
lids_socket_shutdown(struct socket *sock, int how)
{
	struct sock *sk = sock->sk;

	if (lids_load && lids_local_load && sk->sk_family == PF_INET) {
		if (lids_ext_capable(current, LIDS_SOCKET_SHUTDOWN) < 0) {
			lids_alert(LIDS_SOCKET, -1, LIDS_SOCKET_SHUTDOWN,
				   "LIDS_SOCKET_SHUTDOWN", "shutdown socket");
			return LIDS_ERROR(-EPERM);
		}
	}
	return 0;
}
#endif				/* CONFIG_SECURITY_NETWORK */

static int
lids_netlink_send(struct sock *sk,struct sk_buff *skb)
{
	if (capable(CAP_NET_ADMIN))
		cap_raise (NETLINK_CB (skb).eff_cap, CAP_NET_ADMIN);
	else
		NETLINK_CB (skb).eff_cap = 0;
	return 0;
}

static int
lids_netlink_recv(struct sk_buff *skb)
{
	if (!cap_raised(NETLINK_CB(skb).eff_cap, CAP_NET_ADMIN))
		return -EPERM;
	return 0;
}

static int
lids_register(const char *name, struct security_operations *ops)
{
	if (lids_secondary_ops != NULL) {
		printk(KERN_INFO
		       "LIDS: There is already a secondary security module registered\n");
		return -EINVAL;
	}
	if (strcmp(name, "owlsm") == 0) {
		lids_secondary_ops = ops;
		printk(KERN_NOTICE "LIDS: Registering security module %s.\n",
		       name);
		return 0;
	}
	printk(KERN_NOTICE
	       "LIDS: Only openwall modules may be registered with LIDS.\n");
	return -EINVAL;
}

static int
lids_unregister(const char *name, struct security_operations *ops)
{
	if (!lids_secondary_ops) {
		printk(KERN_NOTICE "LIDS: no secondary module %s.\n", name);
		return -EINVAL;
	} else if (ops == lids_secondary_ops) {
		printk(KERN_NOTICE "LIDS: unregistering module %s.\n", name);
		return 0;
	}
	printk(KERN_INFO "LIDS: Attempt to unregister an unknown module %s. \n",
	       name);
	return -EINVAL;
}

struct security_operations lids_security_ops = {
	.ptrace = lids_ptrace,
	.capable = lids_capable,
	.capget = lids_capget,
	.capset_check = lids_capset_check,
	.capset_set = lids_capset_set,

	.bprm_alloc_security = lids_bprm_alloc_security,
	.bprm_free_security = lids_bprm_free_security,
	.bprm_apply_creds = lids_bprm_apply_creds,
	.bprm_set_security = lids_bprm_set_security,
	.bprm_check_security = lids_bprm_check_security,

	.sb_post_mountroot = lids_sb_post_mountroot,
	.sb_post_remount = lids_sb_post_remount,
	.sb_kern_mount  = lids_sb_kern_mount,
	.sb_mount	= lids_sb_mount,

	.inode_free_security = lids_inode_free_security,
	.inode_post_create = lids_inode_post_create,

	.inode_link = lids_inode_link,
	.inode_unlink = lids_inode_unlink,
	.inode_symlink = lids_inode_symlink,
	.inode_mkdir = lids_inode_mkdir,
	.inode_post_mkdir = lids_inode_post_mkdir,
	.inode_rmdir = lids_inode_rmdir,
	.inode_mknod = lids_inode_mknod,
	.inode_rename = lids_inode_rename,
	.inode_readlink = lids_inode_readlink,
	.inode_follow_link = lids_inode_follow_link,
	.inode_permission = lids_inode_permission,
	.inode_setattr = lids_inode_setattr,
	.inode_setxattr = lids_inode_setxattr,
	.inode_getxattr = lids_inode_getxattr,
	.inode_removexattr = lids_inode_removexattr,

	.d_instantiate = lids_d_instantiate,
	.file_permission = lids_file_permission,

	.netlink_send = lids_netlink_send,
	.netlink_recv = lids_netlink_recv,

	.task_alloc_security = lids_task_alloc_security,
	.task_free_security = lids_task_free_security,
	.task_kill = lids_task_kill,
	.task_reparent_to_init = lids_task_reparent_to_init,

#ifdef CONFIG_SECURITY_NETWORK
	.socket_create = lids_socket_create,
	.socket_post_create = lids_socket_post_create,
	.socket_bind = lids_socket_bind,
	.socket_connect = lids_socket_connect,
	.socket_listen = lids_socket_listen,
	.socket_accept = lids_socket_accept,
	.socket_sendmsg = lids_socket_sendmsg,
	.socket_recvmsg = lids_socket_recvmsg,
	.socket_getsockname = lids_socket_getsockname,
	.socket_getpeername = lids_socket_getpeername,
	.socket_getsockopt = lids_socket_getsockopt,
	.socket_setsockopt = lids_socket_setsockopt,
	.socket_shutdown = lids_socket_shutdown,
#endif

	.register_security = lids_register,
	.unregister_security = lids_unregister,
};

extern void setup_lids_module(void);

static void __exit
lids_lsm_exit(void)
{
	struct task_struct *p;

	if (unregister_security(&lids_security_ops)) {
		printk(KERN_INFO
		       "LIDS: Failure unregistering LIDS with the kernel\n");
		return ;
	}
	lids_load = 0;

	lids_sysctl_reset();
	read_lock(&tasklist_lock);
	for_each_process(p) {
		lids_task_free_security(p);
	}
	read_unlock(&tasklist_lock);
	printk(KERN_INFO "LIDS: Exit Successfully\n");
}

#ifdef MODULE
static int __init 
lids_modules_init(void)
{
	struct task_struct *p;

	read_lock(&tasklist_lock);
	for_each_process(p) {
		p->security = NULL;
	}
	read_unlock(&tasklist_lock);
	if (do_lids_setup()) {
		printk(KERN_NOTICE "LIDS_ERR: failed\n");
		lids_lsm_exit();
		return -1;
	}
	printk(KERN_NOTICE "LIDS: Attaching LIDS ACL to Processes\n");

	lids_setup_task_acl(LIDS_STATE_BOOT);

	printk(KERN_NOTICE "LIDS: Finished setting up.\n");

	return 0;
}
#endif

static int __init
lids_lsm_init(void)
{

	printk(KERN_NOTICE "LIDS: Initializing...\n");

	lids_init_setup = 0;
	/* register ourselves with the security framework */
	if (register_security(&lids_security_ops)) {
		printk(KERN_INFO "Failure registering LIDS with the kernel\n");
		return -EINVAL;
	}

#ifdef MODULE
	lids_modules_init();
#endif
	return 0;
}

/* for passing parameter */
__setup("lids=", lids_load_setup);
security_initcall(lids_lsm_init);
module_exit(lids_lsm_exit);

MODULE_AUTHOR("LIDS Team");
MODULE_DESCRIPTION("LIDS Module");
MODULE_LICENSE("GPL");
