/*
 *  LIDS sysctl functions 
 *
 *  Copyright (C) 2002 Huagang Xie <xie@lids.org>
 *  Copyright (C) 2002 Philippe Biondi <biondi@cartel-securite.fr>
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/security.h>
#include <linux/capability.h>
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/spinlock.h>
#include <linux/file.h>
#include <linux/sysctl.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>
#include <asm/ioctls.h>
#include <asm/scatterlist.h>
#include <linux/string.h>
#include <linux/crypto.h>
#include <linux/highmem.h>

#include <linux/lids.h>
#include <linux/lidsext.h>
#include <linux/lidsif.h>
#include <linux/lids_sysctl.h>

enum {
	LIDS_LOCKS = 1,
};
enum {
	CTL_LIDS = 11,
};

extern int lids_init(void);
static int lids_proc_locks_sysctl(ctl_table * table, int write,
				  struct file *filp, void __user *buffer,
				  size_t * lenp, loff_t *ppos);


static ctl_table lids_table[] = {
	{
		.ctl_name	= LIDS_LOCKS,
		.procname	= "locks",
		.maxlen		= sizeof (int),
		.mode		= 0600,
		.proc_handler	= &lids_proc_locks_sysctl,
	},
	{ .ctl_name = 0 }
};

static ctl_table lids_root_table[] = {
	{
		.ctl_name	= CTL_LIDS,
		.procname	= "lids",
		.mode		= 0500,
		.child 		= lids_table,
	},
	{ .ctl_name = 0 }
};

struct lids_s_inode lidsadm;
static struct timer_list fail_timer;

/***********************************************************************
 ***********************************************************************
 *
 * Now, the sysctl procedures
 *
 ***********************************************************************
 ***********************************************************************/

/***********************************************************************
 *
 * What is needed to lock init children
 *
 */

int lids_last_pid = 0;
static struct ctl_table_header *lids_root_table_header;

int
lids_sysctl_init(void)
{
	printk(KERN_NOTICE "LIDS: Initializing sysctl\n");

	lids_root_table_header = register_sysctl_table(lids_root_table, 0);

	if (lids_root_table_header)
		return 0;
	return -1;
}

void
lids_sysctl_reset(void)
{
	unregister_sysctl_table(lids_root_table_header);
	return;
}

static int
lids_load_conf(void)
{
	int old_lids_local_on;
	int old_lids_local_pid;
	int error;

	old_lids_local_on = lids_local_on;
	old_lids_local_pid = lids_local_pid;
	if (lids_load && lids_local_load) {
		LIDS_DBG
		    ("Let's give lidsadm (pid %i) the right to read the conf\n",
		     current->pid);
		lids_local_pid = current->pid;
		lids_local_on = 0;
	}

	error = lids_init();
	if (error) {
		return -1;
	}
	/* cap_bset=locks.cap_bset; */
	cap_bset = lids_cap_val;

	printk(KERN_INFO "LIDS: Attaching ACLs to Processes\n");

	lids_setup_task_acl(lids_state);

	printk(KERN_INFO "LIDS: GLOBAL and %s state Config. files loaded\n",
	       lids_state_name[lids_state - 1]);

	if (lids_state <= LIDS_STATE_SHUTDOWN) {
		printk(KERN_INFO "LIDS: Switching to %s state\n",
		       lids_state_name[lids_state - 1]);
	} else {
		printk(KERN_INFO "LIDS: Invalid State %d\n", lids_state);
	}

	lids_local_pid = old_lids_local_pid;
	lids_local_on = old_lids_local_on;

	return 0;
}

/***********************************************************************
 *
 * The one which process flags changes
 *
 * return value:
 * 	
 * 	0	successful
 * 	-1	failed
 *
 */

static int
lids_process_flags(lids_flags_t flags)
{
	int error = 0;

	lids_process_switch();

	/* if the kernel is sealed, enter "POSTBOOT" */
	if (lids_first_time) {
		lids_state = LIDS_STATE_POSTBOOT;
		error = lids_load_conf();
		if (!error) {
			lids_flag_raise(lids_flags, LIDS_FLAGS_POSTBOOT);
			lids_flag_lower(lids_flags, LIDS_FLAGS_INIT);
		}
		return error;
	}

	if (lids_acl_discovery !=
	    (lids_flag_raised(flags, LIDS_FLAGS_ACL_DISCOVERY_ON) != 0)) {
		/* if ACL_DISCOVERY mode change request */
		lids_acl_discovery =
		    (lids_flag_raised(flags, LIDS_FLAGS_ACL_DISCOVERY_ON) != 0);
		lids_security_alert("LIDS acl discovery mode %s",
				    lids_acl_discovery ? "on" : "off");
		if (lids_acl_discovery)
			lids_flag_raise(lids_flags,
					LIDS_FLAGS_ACL_DISCOVERY_ON);
		else
			lids_flag_lower(lids_flags,
					LIDS_FLAGS_ACL_DISCOVERY_ON);
	}

	/* if the flags raised as shutdown, change the state */
	if (lids_flag_raised(flags, LIDS_FLAGS_SHUTDOWN)
	    && (lids_state != LIDS_STATE_SHUTDOWN)) {
		lids_state = LIDS_STATE_SHUTDOWN;
		error = lids_load_conf();
		if (!error)
			lids_flag_raise(lids_flags, LIDS_FLAGS_SHUTDOWN);
	} else if (lids_flag_raised(flags, LIDS_FLAGS_RELOAD_CONF)) {
		/* Config file reload */
		if (lids_load && !lids_local_on && lids_local_load) {
			printk(KERN_WARNING
			       "Can't reload config files if an LFS is opened and we are not in\n");
		} else {
			error = lids_load_conf();
		}
	}

	return error;
}

/* sha 256 routine */

static int
lids_sha256(char *passwd, int len, char *result)
{
	struct scatterlist sg[8];
	struct crypto_tfm *tfm;

	tfm = crypto_alloc_tfm("sha256", 0);
	if (tfm == NULL) {
		printk("failed to load transform for sha256");
		return -1;
	}
	memset(result, 0, LIDS_PW_LEN * 2);

	sg[0].page = virt_to_page(passwd);
	sg[0].offset = offset_in_page(passwd);
	sg[0].length = len;

	crypto_digest_init(tfm);
	crypto_digest_update(tfm, sg, 1);
	crypto_digest_final(tfm, result);

	crypto_free_tfm(tfm);
	return 0;
}

/***********************************************************************
 *
 * The one which set/get security features
 *
 */

static int number_failed = 0;
static int wait_after_fail = 0;

/* called by timer */
static void
reenable_sysctl(unsigned long user_data)
{
	number_failed = 0;
	wait_after_fail = 0;
}

static int
lids_proc_locks_sysctl(ctl_table * table, int write, struct file *filp,
		       void __user *buffer, size_t * lenp, loff_t *ppos)
{
	lids_locks_t locks;
	struct dentry *dentry;
	struct tty_struct *tty = current->signal->tty;

	/* first: check the terminal and the program which access the sysctl */
	if (lids_check_tty(tty)) {
		lids_security_alert
		    ("Attempt to %s locks sysctl (unauthorized terminal)",
		     write ? "write" : "read");
		return -EPERM;
	}

	dentry = lids_get_task_dentry(current);
	if (dentry == NULL || (dentry->d_inode->i_ino != lidsadm.ino) ||
	    MAJOR(dentry->d_inode->i_sb->s_dev) != lidsadm.dev.major ||
	    MINOR(dentry->d_inode->i_sb->s_dev) != lidsadm.dev.minor) {
		lids_security_alert
		    ("Attempt to %s locks sysctl (unauthorised program)",
		     write ? "write" : "read");
		return -EPERM;
	}
	/* second: check wether it is not a timeout period after two many failed attempts */

	if (wait_after_fail) {
		lids_security_alert("Attempt to %s locks sysctl during timeout",
				    write ? "write" : "read");
		return -EPERM;
	}

	if (write) {
		/* Third : check what is submitted (size, magics, passwd) */
		if (*lenp != sizeof (lids_locks_t)) {
			lids_security_alert
			    ("Attempt to feed locks sysctl with garbage");
			return -EINVAL;
		}
		if (copy_from_user(&locks, buffer, sizeof (lids_locks_t)))
			return -EFAULT;
		if ((locks.magic1 != LIDS_MAGIC_1)
		    || (locks.magic2 != LIDS_MAGIC_2)
		    || (locks.magic3 != LIDS_MAGIC_3)
		    || (locks.magic4 != LIDS_MAGIC_4)) {
			memset((char *) locks.passwd, '\0', sizeof (passwd_t));
			lids_security_alert
			    ("Attempt to feed locks sysctl bad magic numbers");
			return -EINVAL;
		}
		lids_process_password();
	} else {
		if (copy_from_user(&locks, buffer, sizeof (lids_locks_t)))
			return -EFAULT;
		locks.cap_bset = cap_bset;
		locks.flags = lids_flags;

		LIDS_DBG("Sending caps=%#0x flags=%#0x\n", locks.cap_bset,
			 locks.flags);
		if (*lenp < sizeof (lids_locks_t))
			return -EINVAL;
		if (copy_to_user(buffer, &locks, sizeof (lids_locks_t)))
			return -EFAULT;
	}
	return 0;
}
